/*
 * MapDepthMatrix.c
 *
 *  Created on: May 21, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <fftw3.h>
//#include <mkl.h>
#include "utilities.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "Mask.h"
#include "ProblemDimension.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "fftw.h"
#include <cblas.h>

#ifndef SINGLE
#define fftwf_malloc       fftw_malloc
#define fftwf_free         fftw_free
#else
#define cblas_zcopy        cblas_ccopy
#endif

/**
 * @defgroup MapDepthMatrix MapDepthMatrix - MapDepthMatrix maintains a 5D matrix.
 * @ingroup SupportingModules
 */

/**
 * Create a new 4d matrix by allocating space for data and setting
 * internal parameters.
 * @param[in] x number of rows in xy-plane
 * @param[in] y number of columns in xy-plane
 * @param[in] z depth dimension
 * @param[in] t number of measurements (map dimension)
 * @param[in] shift boolean; if output from inverse FT needs to be shifted.
 * @return new 4d matrix
 * @ingroup MapDepthMatrix
 */
MAP_DEPTH_MATRIX newMapDepthMatrix(int x, int y, int z, int t, bool shift) {
	return newMapDepthMatrixV(x, y, z, t, 1, shift);
}

/**
 * Create a new 5d matrix by allocating space for data and setting
 * internal parameters.
 * @param[in] x number of rows in xy-plane
 * @param[in] y number of columns in xy-plane
 * @param[in] z depth dimension
 * @param[in] t number of measurements (map dimension)
 * @param[in] v dimension of vector data (usually set to 3)
 * @param[in] shift boolean; if output from inverse FT needs to be shifted.
 * @return new 5d matrix
 * @ingroup MapDepthMatrix
 */
MAP_DEPTH_MATRIX newMapDepthMatrixV(int x, int y, int z, int t, int v, bool shift) {
	MAP_DEPTH_MATRIX tdm;
	tdm.data = (FloatComplex *) td_malloc(x * y * z * t * v * sizeof(FloatComplex));
	tdm.nX = x;
	tdm.nY = y;
	tdm.nZ = z;
	tdm.nT = t;
	tdm.nV = v;
	tdm.mtype = COMPLEX;
	tdm.shift = shift;
	return tdm;
}

/**
 * Free memory used to store data in 4d matrix
 * @param[in] td A 4d matrix
 * @ingroup MapDepthMatrix
 */
void destroyMapDepthMatrix(MAP_DEPTH_MATRIX td) {
	if (td.data != NULL)
		td_free(td.data, td.nX * td.nY * td.nZ * td.nT * td.nV
				* sizeof(FloatComplex));
}

/**
 * check if MapDepthMatrix is null
 * @param[in] td A 4d matrix
 * @return True if matrix is null, False otherwise
 * @ingroup MapDepthMatrix
 */
bool isEmptyMapDepthMatrix(const MAP_DEPTH_MATRIX *td) {
	if (td->data == NULL)
		return true;
	int size;
	size = td->nX + td->nY + td->nZ + td->nT + td->nV;
	if ((size == 0) || ((td->mtype != COMPLEX) && (td->mtype != REAL)))
		return true;
	else
		return false;
}

/** Extract a 2d sub matrix from a 4d matrix.  Really, this returns a pointer
 * to a 2d matrix inside the 4d matrix.   The pointer is wrapped by a 2d matrix object.
 * @param[in] kx Row index in xy-plane
 * @param[in] ky Column index in xy-plane
 * @param[in] td 4d matrix from which to extract.
 * @return 2d matrix indexed by map (rows) and depth (columns). MATRIX(map, depth).
 * MATRIX.data is a pointer into the 4d matrix.  No data is copied, only a pointer is provided.
 * @ingroup MapDepthMatrix
 */
MATRIX extractMapDepth(int kx, int ky, MAP_DEPTH_MATRIX td) {

	MATRIX m;
	m = newComplexMatrix(td.nT, td.nZ * td.nV);
	td_free(m.data, m.nRow * m.nCol * sizeof(FloatComplex));

	int nCol1, nCol2;
	int xyidx;

	nCol1 = td.nY * td.nZ * td.nT * td.nV;
	nCol2 = td.nZ * td.nT * td.nV;
	xyidx = kx * nCol1 + ky * nCol2;
	m.data = &td.data[xyidx]; // this is a pointer to the td-plane for this x & y

	return m;
}

/** Insert a 2d sub matrix into a 4d matrix.  Here, memory is copied to keep
 * 2d sub matrix in tact, i.e. in contiguous memory.
 * @param[in] kx Row index in xy-plane
 * @param[in] ky Column index in xy-plane
 * @param[in] m  A 2d matrix indexed by map (rows) and depth (columns) to insert.
 * @param[in,out] tdptr Pointer to 4d matrix in which to copy 2d matrix data.  Need to send
 * pointer to 4d matrix because we are changing values in the 4d matrix structure.
 * @ingroup MapDepthMatrix
 */
void insertMapDepth(int kx, int ky, MATRIX m, MAP_DEPTH_MATRIX *tdptr) {

	int nCol1, nCol2;
	int xyidx;
	FloatComplex *p;
	MAP_DEPTH_MATRIX td;

	if (m.mtype != COMPLEX)
		reportError("Matrix must be complex", "insertMapDepth", FATAL);
	if ((m.nRow > 1) && (m.nCol > 1)) {
		if ((m.nRow != tdptr->nT) || (m.nCol != tdptr->nZ * tdptr->nV)) {
			printf("m.nRow =%d m.nCol = %d, tdptr.nT=%d, tdptr.nZ=%d, tdptr.nV=%d\n",
					m.nRow, m.nCol, tdptr->nT, tdptr->nZ, tdptr->nV);
			reportError(MapDepthMatrixBadInsert, "insertMapDepth", FATAL);
		}
	}
	td = *tdptr;

	nCol1 = td.nY * td.nZ * td.nT * td.nV;
	nCol2 = td.nZ * td.nT * td.nV;
	xyidx = kx * nCol1 + ky * nCol2;
	p = m.data;
	// need to copy to keep memory contiguous
	//memcpy(to, from.data, (from.n * sizeof(Float)));
	memcpy(&td.data[xyidx] , p, (m.nRow * m.nCol * sizeof(FloatComplex)));
	//td.data[xyidx] = (FloatComplex *) m.data;  // this is a pointer to the td-plane for this x & y
}

/** Extract a 2d sub matrix from a 4d matrix.  Here, data values are copied because
 * the xy-plane values are dispersed throughout the 4d matrix.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] td 4d matrix from which to extract.
 * @return 2d matrix indexed by x (rows) and y (columns). MATRIX(x,y).
 * A new matrix is created to hold the copied data and keep it contiguous.
 * @ingroup MapDepthMatrix
 */
// return MATRIX(x, y);
MATRIX extractXYPlaneFromMapDepthMatrix(int map, int depth, MAP_DEPTH_MATRIX td) {
	return extractXYPlaneFromMapDepthMatrixV(map, depth, 0, td);
}

/**
 * @ingroup MapDepthMatrix
 */
MATRIX extractXYPlaneFromMapDepthMatrixV(int map, int depth, int v, MAP_DEPTH_MATRIX td) {
	// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
	// the map-distance inversions.
	// It will be awkward to extract the xy-plane; we'll do one element at a time
	MATRIX xyPlane;
	FloatComplex *p;

	xyPlane = newComplexMatrix(td.nX, td.nY);
	p = xyPlane.data;

	FloatComplex *tp;
	int initialPosition, offset;
	initialPosition = map*(td.nZ*td.nV) + v*(td.nZ) + depth;
	// offset = M*Nz*Nv
	offset = td.nT * td.nZ * td.nV;
	tp = td.data + initialPosition;

	cblas_zcopy(xyPlane.nRow*xyPlane.nCol, tp, offset, p, 1); 
	return xyPlane;
}

/**
 * @ingroup MapDepthMatrix
 * MATRIX returned is Float
 */
/** Extract a 2d sub matrix from a linear array of Floats. Use the same data layout
 * in the Float array as in a MAP_DEPTH_MATRIX.
 *  Here, data values are copied because
 * the xy-plane values are dispersed.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] v number of scatterers (3 for vector problems, 1 otherwise)
 * @param[in] nX, nY, nZ, nT, nV  dimensions of the linear array of Floats.
 * @param[in] td linear array of Floats from which to extract.
 * @return 2d matrix indexed by x (rows) and y (columns). MATRIX(x,y).
 * A new matrix is created to hold the copied data and keep it contiguous.
 * @ingroup MapDepthMatrix
 */
// usage: xyPlane = extractXYPlaneFromDoubleArray(t, d, v, Nx, Ny, Nz, M, Nv, &(superY->yElements[superY->noiseBase]));
MATRIX extractXYPlaneFromDoubleArray(int map, int depth, int v, int nX, int nY, int nZ,
		int nT, int nV, Float *td) {
	// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
	// the map-distance inversions.
	// It will be awkward to extract the xy-plane; we'll do one element at a time
	MATRIX xyPlane;
	Float *p;
	int x;

	xyPlane = newMatrix(nX, nY);
	x = xyPlane.nRow;
	p = xyPlane.data;

	Float *tp;
	int initialPosition, offset;
	initialPosition = map*(nZ*nV) + v*(nZ) + depth;
	// offset = M*Nz*Nv
	offset = nT * nZ * nV;
	tp = td + initialPosition;
	for (x=0; x<xyPlane.nRow*xyPlane.nCol; x++) {
		*p++ = *tp;
		tp += offset;
	}
	return xyPlane;
}

/** Insert a 2d sub matrix into a 4d matrix.  Here, data values are copied because
 * the xy-plane values are dispersed throughout the 4d matrix.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] xyPlane  A 2d matrix indexed by x (rows) and y (columns) to insert.
 * @param[in,out] tdptr Pointer to 4d matrix in which to copy 2d matrix data.  Need to send
 * pointer to 4d matrix because we are changing values in the 4d matrix structure.
 * @ingroup MapDepthMatrix
 */
void insertXYPlaneIntoMapDepthMatrix(MATRIX xyPlane, int map, int depth, MAP_DEPTH_MATRIX *tdptr) {
	insertXYPlaneIntoMapDepthMatrixV(xyPlane, map, depth, 0, tdptr);
}

/**
 * @ingroup MapDepthMatrix
 */
void insertXYPlaneIntoMapDepthMatrixV(MATRIX xyPlane, int map, int depth, int v, MAP_DEPTH_MATRIX *tdptr) {
	// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
	// the map-distance inversions.
	// It will be awkward to insert the xy-plane; we'll do one element at a map
	FloatComplex *p;

	MAP_DEPTH_MATRIX td;

	if ((xyPlane.nRow != tdptr->nX) || (xyPlane.nCol != tdptr->nY))
		reportError("Incompatible sizes", "insertXYPlaneIntoMapDepthMatrixV", FATAL);

	td = *tdptr;
	p = xyPlane.data;

	FloatComplex *tp;
	int initialPosition, offset;
	initialPosition = map*(td.nZ*td.nV) + v*(td.nZ) + depth;
	// offset = M*Nz*Nv
	offset = td.nT * td.nZ * td.nV;
	tp = td.data + initialPosition;
	
	cblas_zcopy(xyPlane.nRow*xyPlane.nCol, p, 1, tp, offset); 
}

/**
 * @ingroup MapDepthMatrix
 * MATRIX is Float
 */
/** Insert a 2d map-depth matrix into a linear array of Floats.  Use the same data layout
 * in the Float array as in a MAP_DEPTH_MATRIX. Here, data values are copied because
 * the xy-plane values are dispersed.
 * @param[in] xyPlane MATRIX of Floats to insert into linear array of Floats.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] v number of scatterers (3 for vector problems, 1 otherwise)
 * @param[in] nX, nY, nZ, nT, nV  dimensions of the linear array of Floats.
 * @param[in] tdptr pointer to linear array of Floats.
 * @ingroup MapDepthMatrix
 */
// usage: insertXYPlaneIntoDoubleArray(xyPlane, t, d - 1, v, Nx, Ny, Nz, Nt, Nv, &(superY->yElements[superY->noiseBase]));

void insertXYPlaneIntoDoubleArray(MATRIX xyPlane, int map, int depth, int v, int nX, int nY,
		int nZ, int nT, int nV, Float *tdptr) {
	// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
	// the map-distance inversions.
	// It will be awkward to insert the xy-plane; we'll do one element at a map
	Float *p;
	int x;

	if ((xyPlane.nRow != nX) || (xyPlane.nCol != nY))
		reportError("Incompatible sizes", "insertXYPlaneIntoDoubleArray", FATAL);

	x = xyPlane.nRow;
	p = xyPlane.data;

	Float *tp;
	int initialPosition, offset;
	initialPosition = map*(nZ*nV) + v*(nZ) + depth;
	// offset = M*Nz*Nv
	offset = nT * nZ * nV;
	tp = tdptr + initialPosition;
	for (x=0; x<xyPlane.nRow*xyPlane.nCol; x++) {
		*tp = *p++;
		tp += offset;
	}
}


/** Extract a 2d sub matrix from a 4d matrix.  Here, data values are copied because
 * the xy-plane values are dispersed throughout the 4d matrix.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] td 4d matrix from which to extract.
 * @return 2d matrix indexed by x (rows) and y (columns). MATRIX(x,y).
 * A new matrix is created to hold the copied data and keep it contiguous.
 * @ingroup MapDepthMatrix
 */
MATRIX extractXYPlaneColMajorMapDepthMatrix(int map, int depth, const MAP_DEPTH_MATRIX *td) {
	return extractXYPlaneColMajorMapDepthMatrixV(map, depth, 0, td);
}

/**
 * @ingroup MapDepthMatrix
 */
MATRIX extractXYPlaneColMajorMapDepthMatrixV(int map, int depth, int v, const MAP_DEPTH_MATRIX *tdptr) {
	// td matrix is a 2d array of 2d arrays;
	// ColMajor ordering specific for LSQR inversions
	MATRIX xyPlane;
	FloatComplex *p;
	int xyidx;
	int sizeOfMap, sizeOfZPlane;

	xyPlane = newComplexMatrix(tdptr->nX, tdptr->nY);

	td_free(xyPlane.data, xyPlane.nRow * xyPlane.nCol * sizeof(FloatComplex));

	sizeOfMap = tdptr->nX * tdptr->nY * tdptr->nZ * tdptr->nV;
	sizeOfZPlane = tdptr->nX * tdptr->nY * tdptr->nV;
	xyidx = map * sizeOfMap + depth * sizeOfZPlane;
	p = tdptr->data;
	xyPlane.data = &(p[xyidx]); // this is a pointer to the xy-plane for this map and depth

	return xyPlane;
}

/** Insert a 2d sub matrix into a 4d matrix.  Here, data values are copied because
 * the xy-plane values are dispersed throughout the 4d matrix.
 * @param[in] map Row index in td-plane
 * @param[in] depth Column index in td-plane
 * @param[in] xyPlane  A 2d matrix indexed by x (rows) and y (columns) to insert.
 * @param[in,out] tdptr Pointer to 4d matrix in which to copy 2d matrix data.  Need to send
 * pointer to 4d matrix because we are changing values in the 4d matrix structure.
 * @ingroup MapDepthMatrix
 */
void insertXYPlaneColMajorMapDepthMatrix(MATRIX xyPlane, int map, int depth, MAP_DEPTH_MATRIX *tdptr) {
	insertXYPlaneColMajorMapDepthMatrixV(xyPlane, map, depth, 0, tdptr);
}

/**
 * @ingroup MapDepthMatrix
 */
void insertXYPlaneColMajorMapDepthMatrixV(MATRIX xyPlane, int map, int depth, int v, MAP_DEPTH_MATRIX *tdptr) {
	// td matrix is a 2d array of 2d arrays;
	// ColMajor ordering specific for LSQR inversions
	FloatComplex  *p;
	int xyidx;

	if ((xyPlane.nRow != tdptr->nX) || (xyPlane.nCol != tdptr->nY))
		reportError(MapDepthMatrixBadInsert, "insertXYPlaneColMajorMapDepthMatrixV", FATAL);


	///////////////
	// do memory copy to keep memory contiguous
	int sizeOfMap, sizeOfZPlane;

	sizeOfMap = tdptr->nX * tdptr->nY * tdptr->nZ * tdptr->nV;
	sizeOfZPlane = tdptr->nX * tdptr->nY * tdptr->nV;
	xyidx = map * sizeOfMap + depth * sizeOfZPlane;
	p = xyPlane.data;
	// need to copy to keep memory contiguous
	//memcpy(to, from.data, (from.n * sizeof(Float)));
	memcpy(&(tdptr->data[xyidx]) , p, (xyPlane.nRow * xyPlane.nCol * sizeof(FloatComplex)));
}

/**
 * @ingroup MapDepthMatrix
 */
MAP_DEPTH_MATRIX readMapDepthMatrixKernel(char *basedir, char *filenames[], int nFiles, bool useFT) {
	int vdim = 1;
	if (getProblemDimension() == 5) {
		vdim = 3;
	}
	return readMapDepthMatrixV(basedir, filenames, nFiles, vdim, useFT);
}

/**
 * @ingroup MapDepthMatrix
 */
MAP_DEPTH_MATRIX readMapDepthMatrix(char *basedir, char *filenames[],
			int nFiles, bool useFT) {
	return readMapDepthMatrixV(basedir, filenames, nFiles, 1, useFT);
}

/** Opens and reads FITS files, then stores the data in a new MAP_DEPTH_MATRIX.
* @param[in] basedir base directory for all filenames
* @param[in] filenames list of FITS files to read.
* @param[in] nFiles number of files to read
* @param[in] vdim The vector dimension; for 3d, vdim = 1; for 5d, vdim = 3.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @return new map depth matrix containing data from files.
* @ingroup MapDepthMatrix
*/
MAP_DEPTH_MATRIX readMapDepthMatrixV(char *basedir, char *filenames[],
		int nFiles, int vdim, bool useFT) {

	MultiDimFits KernelFileInfo;
	fits2d onePlane;
	MATRIX kernelk;
	MATRIX m;
	int map;
	int t, v;
	MAP_DEPTH_MATRIX tdm;
	char *fullName;
	int dirPathEnd;
	int fullNameLength;
	WhichPlane whichPlane;

	map = nFiles;
	tdm.data = NULL;

	/// HERE -- if 3dv, then read 3 x kernel file names
	fullNameLength = (longestName(filenames, nFiles*vdim) + strlen(basedir) + 3) * sizeof(char);
	fullName = (char *) td_malloc(fullNameLength);
	strcpy(fullName, basedir);
	if (!endsWith(basedir, "/") && (strlen(fullName) > 0)) {
		strcat(fullName, "/");
	}
	dirPathEnd = strlen(fullName);
	// for each kernel file
	for (t = 0; t < map; t++) {
		for (v = 0; v < vdim; v++) {
			strcpy(&fullName[dirPathEnd], filenames[t*vdim + v]);
			printf("%s\n", fullName);
			// open the kernel file and read one plane at a time
			KernelFileInfo = openMultiDimFitsRead(fullName);

			// make sure the xy-plane is square
			if (KernelFileInfo.max.x != KernelFileInfo.max.y)
				reportError("XY-plane must be square (nX = nY)", "readMapDepthMatrixV", FATAL);

			// need to know the dimensions of x,y and z before allocating the matrix
			// the dimensions are specified in the fits header
			int ydim = KernelFileInfo.max.y;
			if (useFT) { // if using Fourier transform, then the y-dimension is cut in half
				ydim = KernelFileInfo.max.y / 2 + 1;
			}
			if (tdm.data == NULL) {
				bool shift = false;

				tdm = newMapDepthMatrixV(KernelFileInfo.max.x, ydim,
						KernelFileInfo.max.z, map, vdim, shift);
			} else {
				// ydim is based on Fourier transform; if data is transformed, the dimension is decreased, otherwise
				// it is unchanged.
				if ((KernelFileInfo.max.x != tdm.nX) || ((KernelFileInfo.max.y != tdm.nY) && (!useFT)) ||
						((useFT) && ((KernelFileInfo.max.y/2+1) != tdm.nY)) ||
						(KernelFileInfo.max.z != tdm.nZ))
					reportError(MapDepthMatrixReadUnexpectedDim, filenames[t*vdim + v],
							FATAL);
			}
			int d;
			for (d = 1; d <= KernelFileInfo.max.z; d++) {

				// get next Kernel plane
				whichPlane.z = d;
				whichPlane.t = t+1;
				onePlane = readNextPlane(KernelFileInfo, whichPlane);
				if (useFT) {
					// perform fft shift before transform
					m = fftwShift(toMatrix(onePlane));
					// calculate Mij(k); Fourier transform the next plane of the Kernel file
					kernelk = fftw(MatrixToFitsObject(m));
					freeMatrix(m);
					//free(onePlane.data); don't free because it is used repeatedly by readNextPlane
				} else {
					kernelk = newComplexMatrix(onePlane.xDim, onePlane.yDim);
					FloatComplex *p;
					p = kernelk.data;
					int pi;
					for (pi = 0; pi < kernelk.nRow * kernelk.nCol; pi++)
						p[pi] = onePlane.data[pi] + 0.0 * I;
				}
				insertXYPlaneIntoMapDepthMatrixV(kernelk, t, d - 1, v, &tdm);
				if (useFT) {
					fftwf_free(kernelk.data);
				} else {
					freeMatrix(kernelk);
				}
			} // end for each plane of kernel file

			// close the kernel file
			closeMultiDimFits(KernelFileInfo);
		} // end for each vdim
	} // end for each map dim
	td_free(fullName, fullNameLength);
	return tdm;
}

/** Opens and writes MAP_DEPTH_MATRIX to FITS files. Each t-dimension is written
 * to a separate FITS file.
* @param[in] filenames list of FITS files to write; NOTE: each filename must already have
* 						directory path, i.e. the filenames are complete paths.
* @param[in] nFiles number of files to write.
* @param[in] tdm map depth matrix containing data for output.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @param[in] headerParameters  Names, values and comments of the lambda parameters for FITS header.
* @ingroup MapDepthMatrix
*/
// NOTE:  read and write are not symmetric.  Read takes basedir and constructs filename;
// write receives a list of complete paths to files
void writeMapDepthMatrix(char *filenames[], int nFiles, MAP_DEPTH_MATRIX tdm,
		bool useFT, FITSParameters headerParameters) {

	MultiDimFits outFileInfo;
	MATRIX kernelk, kernel;
	int map;
	int t, v;
	WhichPlane whichPlane;

	map = tdm.nT;

	if (nFiles != map * tdm.nV)
		reportError(MapDepthMatrixNotEnoughFilenames, "writeMapDepthMatrix", FATAL);

	// for each kernel file
	for (t = 0; t < map; t++) {

		for (v = 0; v < tdm.nV; v++) {
			// open the file and write one plane at a time
			outFileInfo.fptr = NULL;
			outFileInfo.max.x = tdm.nX;
			if (useFT) {
				// because we are inverse Fourier transforming the data (tdm.nY);
				outFileInfo.max.y = tdm.nX; // assume it is square; don't use: 2 * (tdm.nY - 1);
			} else {
				outFileInfo.max.y = tdm.nY;
			}
			outFileInfo.max.z = tdm.nZ;
			outFileInfo.dimension = 3;

			openMultiDimFitsWrite(filenames[t*tdm.nV + v], &outFileInfo, headerParameters);

			int d;
			for (d = 1; d <= tdm.nZ; d++) {
				kernelk = extractXYPlaneFromMapDepthMatrixV(t, d - 1, v, tdm);
				if (useFT) {

					// calculate Mij(k); Fourier transform the next plane of the Kernel file
					kernel = fftwReverse(kernelk, tdm.shift);
				} else {
					// just use the real part
					kernel = ConvertToReal(kernelk);
				}
				fits2d thedata = MatrixToFitsObject(kernel);
				// write next Kernel plane
				whichPlane.z = d;
				whichPlane.t = t+1;

				writeNextPlane(outFileInfo, whichPlane, thedata);
				freeMatrix(kernelk);

				if (useFT) {
					if (tdm.shift) {
						// then already called fftw_free
						freeMatrix(kernel);
					} else {
						// need to call fftw_free
						fftwf_free(kernel.data);
					}
				} else {
					freeMatrix(kernel);
				}

			} // end for each plane of kernel file

			// close the file
			closeMultiDimFits(outFileInfo);
		} // end for v
	} // end for map
}

//
// NOTE:  read and write are not symmetric.  Read takes basedir and constructs filename;
// write receives a list of complete paths to files
// write each Z dim to a separate file.
void writeDepthMapMatrix(char *filenames[], int nFiles, MAP_DEPTH_MATRIX tdm, bool useFT,
		FITSParameters headerParameters) {

	MultiDimFits outFileInfo;
	MATRIX kernelk, kernel;
	int map;
	int t;
	WhichPlane whichPlane;

	map = tdm.nT;

	if (nFiles != tdm.nZ)
		reportError(MapDepthMatrixNotEnoughFilenames, "writeDepthMapMatrix", FATAL);

	// for each depth file
	int d;
	for (d = 0; d < tdm.nZ; d++) {


		// open the kernel file and write one plane at a time
		outFileInfo.fptr = NULL;
		outFileInfo.max.x = tdm.nX;
		if (useFT) {
			outFileInfo.max.y = tdm.nX; // assume xy-plane is square; don't use: 2 * (tdm.nY - 1);
			// because we are inverse Fourier transforming the data (tdm.nY);
		} else {
			outFileInfo.max.y = tdm.nY;
		}
		outFileInfo.max.z = 1;
		outFileInfo.max.t = tdm.nT;
		outFileInfo.dimension = 4;

		printf("filename = %s\n", filenames[d]);

		openMultiDimFitsWrite(filenames[d], &outFileInfo, headerParameters);

		for (t = 0; t < map; t++) {
			kernelk = extractXYPlaneFromMapDepthMatrix(t, d, tdm);

			if (useFT) {
				// calculate Mij(k); Fourier transform the next plane of the Kernel file
				kernel = fftwReverse(kernelk, tdm.shift);
			} else {
				// just use the real part
				kernel = ConvertToReal(kernelk);

			}
			fits2d thedata = MatrixToFitsObject(kernel);
			// write next Kernel plane
			whichPlane.z = 1;
			whichPlane.t = t+1;
			writeNextPlane(outFileInfo, whichPlane, thedata);
			freeMatrix(kernelk);
			if (useFT && !tdm.shift) {
				// need to call fftw_free
				fftwf_free(kernel.data);
			} else {
				freeMatrix(kernel);
			}

		} // end for each plane of kernel file

		// close the file
		closeMultiDimFits(outFileInfo);
	}
}

/** Opens and writes MAP_DEPTH_MATRIX to a single FITS file.
* @param[in] filename name of FITS file to write; NOTE: filename must already have
* 						directory path, i.e. the filename is a complete path.
* @param[in] tdm map depth matrix containing data for output.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @param[in] headerParameters  Names, values and comments of the lambda parameters for FITS header.
* @ingroup TimeDepthMatrix
*/
void writeMapDepthMatrixToOneFile(char *filename, MAP_DEPTH_MATRIX tdm,
		bool useFT, FITSParameters headerParameters) {

	MultiDimFits outFileInfo;
	MATRIX kernelk, kernel;
	int map;
	int t, v;
	WhichPlane whichPlane;

	if (tdm.nV > 1)
		reportError("Cannot write MAP_DEPTH_MATRIX to one file because nV > 1", "writeMapDepthMatrixToOneFile", FATAL);

	map = tdm.nT;

	// open the file and write one plane at a time
	outFileInfo.fptr = NULL;
	outFileInfo.max.x = tdm.nX;
	if (useFT) {
		outFileInfo.max.y = tdm.nX; // assume xy-plane is square; don't use: 2 * (tdm.nY - 1); // because we are inverse Fourier transforming the data (tdm.nY);
	} else {
		outFileInfo.max.y = tdm.nY;
	}
	outFileInfo.max.z = tdm.nZ;
	outFileInfo.max.t = tdm.nT;
	outFileInfo.dimension = 4;

	openMultiDimFitsWrite(filename, &outFileInfo, headerParameters);

	for (t = 0; t < map; t++) {
		for (v = 0; v < tdm.nV; v++) {
			int d;
			for (d = 1; d <= tdm.nZ; d++) {
				kernelk = extractXYPlaneFromMapDepthMatrixV(t, d - 1, v, tdm);
				if (useFT) {
					// calculate Mij(k); Fourier transform the next plane of the Kernel file
					kernel = fftwReverse(kernelk, tdm.shift);
				} else {
					// just use the real part
					kernel = ConvertToReal(kernelk);
				}
				fits2d thedata = MatrixToFitsObject(kernel);
				// write next Kernel plane
				whichPlane.z = d;
				whichPlane.t = t + 1;
				writeNextPlane(outFileInfo, whichPlane, thedata);
				freeMatrix(kernelk);
				if (useFT && !tdm.shift) {
					// need to call fftw_free
					fftwf_free(kernel.data);
				} else {
					freeMatrix(kernel);
				}
			} // end for each plane of file
		} // end for v
	} // end for map
	// close the file
	closeMultiDimFits(outFileInfo);
}


/** Opens and reads 4D FITS file, then stores the data in a new MAP_DEPTH_MATRIX.
* @param[in] filename of FITS file to read; must be complete file path.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @param[in] shiftOnInput True mean shift the data before fft transform.
* @param[in] shiftOnOutput True means shift the data after inverse fft transform.
* @return new map depth matrix containing data from files.
* @ingroup TimeDepthMatrix
*/
MAP_DEPTH_MATRIX readMapDepthMatrixFromOneFileS(char *filename, bool useFT, bool shiftOnInput, bool shiftOnOutput) {

	MultiDimFits KernelFileInfo;
	fits2d onePlane;
	MATRIX kernelk;
	MATRIX m;
	int t, v;
	int d;
	MAP_DEPTH_MATRIX tdm;
	WhichPlane whichPlane;

	v = 0;

	printf("%s\n", filename);
	// open the FITS file and read one plane at a time
	KernelFileInfo = openMultiDimFitsRead(filename);

	// make sure the xy-plane is square
	if (KernelFileInfo.max.x != KernelFileInfo.max.y)
		reportError("XY-plane must be square (nX = nY)", "readMapDepthMatrixFromOneFileS", FATAL);

	// need to know the dimensions of x,y and z before allocating the matrix
	// the dimensions are specified in the fits header
	int ydim = KernelFileInfo.max.y;
	if (useFT) { // if using Fourier transform, then the y-dimension is cut in half
		ydim = KernelFileInfo.max.y / 2 + 1;
	}

	tdm = newMapDepthMatrixV(KernelFileInfo.max.x, ydim, KernelFileInfo.max.z,
			KernelFileInfo.max.t, 1, shiftOnOutput);

	for (t = 1; t <= KernelFileInfo.max.t; t++) {
		for (d = 1; d <= KernelFileInfo.max.z; d++) {

			// get next Kernel plane
			whichPlane.z = d;
			whichPlane.t = t;
			onePlane = readNextPlane(KernelFileInfo, whichPlane);
			if (useFT) {
				if (shiftOnInput) {
					// perform fft shift before transform
					m = fftwShift(toMatrix(onePlane));
				} else {
					m = toMatrix(onePlane);
				}
				// calculate Mij(k); Fourier transform the next plane of the Kernel file
				kernelk = fftw(MatrixToFitsObject(m));
				if (shiftOnInput) {
					freeMatrix(m);
				}
				//free(onePlane.data); don't free because it is used repeatedly by readNextPlane
			} else {
				kernelk = newComplexMatrix(onePlane.xDim, onePlane.yDim);
				FloatComplex *p;
				p = kernelk.data;
				int pi;
				for (pi = 0; pi < kernelk.nRow * kernelk.nCol; pi++)
					p[pi] = onePlane.data[pi] + 0.0 * I;
			}
			insertXYPlaneIntoMapDepthMatrixV(kernelk, t - 1, d - 1, v, &tdm);
			if (useFT) {
				fftwf_free(kernelk.data);
			} else {
				freeMatrix(kernelk);
			}
		} // end for each plane of FITS file
	} // end for each plane of FITS file

	// close the kernel file
	closeMultiDimFits(KernelFileInfo);
	return tdm;
}

/** Opens and reads 4D FITS file, then stores the data in a new MAP_DEPTH_MATRIX.
 * Note: the data are shifted on input (when read) and NOT shifted when written to a file
 * (on output).
* @param[in] filename of FITS file to read; must be complete file path.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @return new map depth matrix containing data from files.
* @ingroup TimeDepthMatrix
*/
MAP_DEPTH_MATRIX readMapDepthMatrixFromOneFile(char *filename, bool useFT) {
	bool shiftOnInput, shiftOnOutput;
	shiftOnInput = true;
	shiftOnOutput = false;

	return readMapDepthMatrixFromOneFileS(filename, useFT, shiftOnInput, shiftOnOutput);

}

/** Opens and reads 4D FITS file, then stores the data in a new MAP_DEPTH_MATRIX.
 * Note: the data are not shifted on input (when read) and not shifted when written to a file
 * (on output).
* @param[in] filename of FITS file to read; must be complete file path.
* @param[in] useFT Boolean; whether to Fourier transform the data before inserting into
* 				MAP_DEPTH_MATRIX.
* @return new map depth matrix containing data from files.
* @ingroup TimeDepthMatrix
*/
MAP_DEPTH_MATRIX readMapDepthMatrixForTarget(char *filename, bool useFT) {
	bool shiftOnInput, shiftOnOutput;
	shiftOnInput = false;
	shiftOnOutput = false;

	return readMapDepthMatrixFromOneFileS(filename, useFT, shiftOnInput, shiftOnOutput);

}

/**
 * Calculate the inverse of each tdm(kx,ky,:,:)
 * and store it in a new matrix.
 * @param[in] tdm Map-depth matrix
 * @return MAP_DEPTH_MATRIX containing the invervse of each map-depth submatrix.
 * If an error occurs during inversion, the returned matrix has data = NULL.
 */
MAP_DEPTH_MATRIX inverseMapDepth(MAP_DEPTH_MATRIX tdm) {
	int ii, jj;
	MATRIX m;
	MATRIX mi;
	MAP_DEPTH_MATRIX tdmI;

	tdmI = newMapDepthMatrixV(tdm.nX, tdm.nY, tdm.nZ, tdm.nT, tdm.nV, false);
	for (ii = 0; ii<tdm.nX; ii++) {
		for (jj=0; jj<tdm.nY; jj++) {
			m = extractMapDepth(ii, jj, tdm);
			mi = inverse(m);
			if (mi.data == NULL) {
				reportError("Cannot invert matrix", "inverseMapDepth", ERROR);
				tdmI.data = NULL;
			}
			insertMapDepth(ii, jj, mi, &tdmI);
			freeMatrix(mi);
		}
	}
	return tdmI;
}

/**
 * Convert the ordering of data in  mdm(kx,ky,:,:)
 * and store it in a new matrix.
 * @param[in] tdm Map-depth matrix
 * @param[in] order Enumeration: either ROW_MAJOR or COL_MAJOR
 * @return MAP_DEPTH_MATRIX containing the new ordering of each map-depth submatrix.
 */
MAP_DEPTH_MATRIX convertMapDepth(const MAP_DEPTH_MATRIX *tdm, MatrixOrder order) {
	//enum matrixOrder {ROW_MAJOR, COL_MAJOR};
	// Convert mask to column-major order
	MATRIX m;
	MATRIX mi;
	MAP_DEPTH_MATRIX tdmNew;

	reportError("convertMapDepth not implemented", "convertMapDepth", FATAL);

	tdmNew = newMapDepthMatrixV(tdm->nX, tdm->nY, tdm->nZ, tdm->nT, tdm->nV,
			false);

	int a, z, v;
	if (order == COL_MAJOR) {
		for (a = 0; a < tdm->nT; a++) {
			for (v = 0; v < tdm->nT; v++) {
				for (z = 0; z < tdm->nZ; z++) {
					// can use regular extract
					m = extractXYPlaneFromMapDepthMatrixV(a, z, v, *tdm);
					mi = toColumnMajor2(m);
					freeMatrix(m);
					// use new insert; CANNOT use regular insert
					insertXYPlaneColMajorMapDepthMatrixV(mi, a, z, v, &tdmNew);
					freeMatrix(mi);
				}
			}
		}
	} else {
		for (a = 0; a < tdm->nT; a++) {
			for (v = 0; v < tdm->nT; v++) {
				for (z = 0; z < tdm->nZ; z++) {
					// use new extract; CANNOT use regular extract because MapDepth matrix is in ColMajor order
					m = extractXYPlaneColMajorMapDepthMatrixV(a, z, v, tdm);
					mi = toRowMajor2(m);
					// can use regular insert
					insertXYPlaneIntoMapDepthMatrixV(mi, a, z, v, &tdmNew);
					freeMatrix(mi);
				}
			}
		}
	}

	return tdmNew;
}
// end convert mask to column-major

/**
 * Convert between real space and wave space
 * and store result in a new matrix.
 * @param[in] tdm Map-depth matrix
 * @param[in] toSpace Enumeration: either REAL_SPACE or WAVE_SPACE
 * @return MAP_DEPTH_MATRIX containing the new ordering of each map-depth submatrix.
 */
MAP_DEPTH_MATRIX convertMapDepthSpace(const MAP_DEPTH_MATRIX *tdm, enum DataSpace toSpace) {

	// Convert mask to column-major order
	MATRIX m;
	MAP_DEPTH_MATRIX tdmNew;
	int Nx, NyOrig, NyNew, Nz, Nv, Nt;

	Nx = tdm->nX;
	NyOrig = tdm->nY;
	if (toSpace == WAVE_SPACE) {
		NyNew = NyOrig/2 + 1;
	} else {
		NyNew = Nx;  // assume xy-plane is square; don't use this: (NyOrig - 1) * 2;
	}
	Nz = tdm->nZ;
	Nt = tdm->nT;
	Nv = tdm->nV;
	bool shift = tdm->shift;

	tdmNew = newMapDepthMatrixV(Nx, NyNew, Nz, Nt, Nv, shift);

	int a, z, v;

	for (a = 0; a < Nt; a++) {
		for (v = 0; v < Nv; v++) {
			for (z = 0; z < Nz; z++) {
				MATRIX PQ;
				PQ = extractXYPlaneFromMapDepthMatrixV(a, z, v, *tdm);
				if (toSpace == WAVE_SPACE) {
					MATRIX temp = ConvertToReal(PQ);
					freeMatrix(PQ);
					if (shift) {
						PQ = fftwShift(temp);
						freeMatrix(temp);
						m = fftw(MatrixToFitsObject(PQ)); // no new memory allocated
						freeMatrix(PQ);
					} else {
						m = fftw(MatrixToFitsObject(temp));
						freeMatrix(temp);
					}
					insertXYPlaneIntoMapDepthMatrixV(m, a, z, v, &tdmNew);
					fftwf_free(m.data);
				} else {
					m = fftwReverse(PQ, shift);
					MATRIX mc = ConvertToComplex(m);
					if (shift)
						freeMatrix(m);
					else
						fftwf_free(m.data);
					freeMatrix(PQ);
					m = mc;
					insertXYPlaneIntoMapDepthMatrixV(m, a, z, v, &tdmNew);
					freeMatrix(m);
				}


			}
		}  // end for v
	}

	return tdmNew;
}

/**
 * Multiply a 4d matrix by a scalar and take the complex conjugate of each element.
 * I combined the scalar multiplication and conjugate together in one operation
 * because it is faster to go through the 4d matrix once, rather than twice.
 * @param[in,out] tdmPtr Pointer to a MAP_DEPTH_MATRIX.  The original matrix
 * 					is changed.
 * @param[in] scalar value to multiply.
 */
void TdmMultiplyScalarTakeConjugate(MAP_DEPTH_MATRIX *tdmPtr, Float scalar) {
	MAP_DEPTH_MATRIX tdm;
	int ii;
	FloatComplex *p;

	tdm = *tdmPtr;

	p = tdm.data;
	for (ii = 0; ii<tdm.nX*tdm.nY*tdm.nZ*tdm.nT*tdm.nV; ii++) {
		*p = conj(*p * scalar);
		p++;
	}
}

/**
 * Multiply a 4d matrix by a scalar .
 * @param[in,out] tdmPtr Pointer to a MAP_DEPTH_MATRIX.  The original matrix
 * 					is changed.
 * @param[in] scalar value to multiply.
 */
void TdmMultiplyScalar(MAP_DEPTH_MATRIX *tdmPtr, Float scalar) {
	MAP_DEPTH_MATRIX tdm;
	int ii;
	FloatComplex *p;

	tdm = *tdmPtr;

	p = tdm.data;
	for (ii = 0; ii<tdm.nX*tdm.nY*tdm.nZ*tdm.nT*tdm.nV; ii++) {
		*p++ *= scalar;
	}
}


/**
 * Multiply a 4D matrix by a 2D matrix .
 * @param[in,out] tdmPtr Pointer to a MAP_DEPTH_MATRIX.  The original matrix
 * 					is changed.
 * @param[in] theta Matrix to multiply.
 */
void MdmMultiplyMatrix(MAP_DEPTH_MATRIX *tdmPtr, MATRIX theta) {
	MAP_DEPTH_MATRIX tdm;
	int ii, jj;
	MATRIX m;

	tdm = *tdmPtr;

	if ((tdm.nZ * tdm.nV) != theta.nRow)
		reportError(MatrixIncompatibleSizes, "MdmMultiplyMatrix", FATAL);

	for (ii = 0; ii<tdm.nX; ii++) {
		for (jj=0; jj<tdm.nY; jj++) {

			// need to extract submatrix because only submatrices are guaranteed to be stored contiguously.
			m = extractMapDepth(ii, jj, tdm);
			MATRIX mTheta = multiplyMM2('N', 'N', m, theta);
			insertMapDepth(ii, jj, mTheta, tdmPtr);
			freeMatrix(mTheta);
		}
	}
}

/**
 * zero inversion weights, w, for wave number, kx=0, ky=0.
 * @param[in,out] tdmPtr Pointer to a MAP_DEPTH_MATRIX containing inversion weights.
 */
void ZeroK0(MAP_DEPTH_MATRIX *tdmPtr)  {

	MAP_DEPTH_MATRIX tdm;
	int ii, jj, kk;
	MATRIX m;
	FloatComplex *p;

	tdm = *tdmPtr;

	for (ii = 0; ii<tdm.nX; ii++) {
		for (jj=0; jj<tdm.nY; jj++) {
			// need to extract submatrix because only submatrices are guaranteed to be stored contiguously.
			m = extractMapDepth(0, 0, tdm);
			p = m.data;
			// zero w(k=0) elements
			for (kk = 0; kk < m.nRow*m.nCol; kk++)
				*p++ = 0.0 + 0.0*I;
		}
	}
}

/**
 * Make a copy of a map-depth matrix. WARNING: only use when memory is guaranteed to be contiguous for
 * the data.
 * @param[in] tdm matrix to copy.
 * @return copy of matrix
 */
MAP_DEPTH_MATRIX TdmCopy(MAP_DEPTH_MATRIX tdm) {
	MAP_DEPTH_MATRIX tdm2;

	tdm2 = newMapDepthMatrixV(tdm.nX, tdm.nY, tdm.nZ, tdm.nT, tdm.nV, tdm.shift);
	int size;
	if (tdm.nV != 0) {
		size = tdm.nV;
	} else {
		size = 1;
	}
	// memcpy(to, from, size)
	memcpy(tdm2.data, tdm.data, (size * tdm.nX * tdm.nY * tdm.nZ * tdm.nT * sizeof(FloatComplex)));
	return tdm2;
}

/**
 * reconstitute weight vector, w, by adding zeros at locations specified in mask.
 * w must be in real space, NOT wave space.  Mask is always in real space.
 * @param[in] x Packed MATRIX containing weights.
 * @param[in] mask MASK matrix containing masks for all maps.
 * @param[in,out] w Weights fully expanded with zeros inserted for undefined map points
 *                which are indicated by false in the corresponding mask.
 */
void unpack(MATRIX *x, const MASK *mask, MAP_DEPTH_MATRIX *w) {
	// for each map, unpack and construct weight vector, w,
	// as a MAP_DEPTH_MATRIX

	// x is in column-major form. mask is in column-major form. ==> w is in column-major form
	// return w in row-major form

	// go through entire array for mask;
	// 	if mask(rj)== 1, then insert value from x, into w(rj) and advance pointer into x
	//  else insert 0.0 into w(rj)
	// advance pointers into mask and w
	if ((w->nZ != 1) || (w->nV != 1))
		reportError("weights can only depend on (x,y) and M=number of maps.  This means nZ and nV must = 1 for weight MAP_DEPTH_MATRIX", "unpack", ERROR);
	if ((mask->nT != w->nT) || (mask->nX != w->nX) || (mask->nY != w->nY)) // || (mask->nZ != w->nZ) || (mask->nV != w->nV))
		reportError("incompatible sizes of w and mask in unpack", "unpack", ERROR);
	if ( !((x->nCol == 1) || (x->nRow ==1)) )
		reportError("x must be a column vector", "unpack", ERROR);

	if (w->data != NULL) {
		FloatComplex *wp;
		bool *maskp;
		Float *xp;

		wp = w->data;
		maskp = mask->data;
		xp = x->data;
		int i;
		int count = mask->nX * mask->nY * mask->nT;
		for (i = 0; i < count; i++) {
			if (*maskp) {
				*wp = *xp + 0.0 * I;
				xp++;
			} else {
				*wp = 0.0;
			}
			maskp++;
			wp++;
			if (wp > (w->data + count * sizeof(FloatComplex)))
				reportError("exceeded bounds", "unpack", FATAL);
		}

	}
}

/**
 * condense weight vector, w, by removing zero values specified in mask.
 * send already allocated dvec, x
 * matrix w_n, is inserted into dvec *x.
 * @param[in] w_n weights only depend on r_i = (x,y) and M = # of maps
 * @param[in,out] x weights packed into a column vector with this ordering:
 * w_a(r_1), w_a(r_2) ... w_b(r_1), w_b(r_2) where r_1=(x,y)=(0,0) and r_2=(0,1), i.e. row-major order
 * @param[in] mask
 */
// because we don't  know the offset into the dvec for the map, we need to construct the full MapDepthMatrix,
// then pack it into a vector.
void pack(const MAP_DEPTH_MATRIX *w_n, MATRIX *x, const MASK *mask) {

	MAP_DEPTH_MATRIX w = *w_n;

	if ((w.nZ != 1) || (w.nV != 1))
		reportError("weights can only depend on (x,y) and M=number of maps.  This means nZ and nV must = 1 for weight MAP_DEPTH_MATRIX", "pack", ERROR);
	if ((mask->nT != w.nT) || (mask->nX != w.nX) || (mask->nY != w.nY)) //  || (mask->nZ != w.nZ) || (mask->nV != w.nV))
		reportError("incompatible sizes of w and mask in pack", "pack", ERROR);
	if ( !((x->nCol == 1) || (x->nRow ==1)) )
		reportError("x must be a column vector", "pack", ERROR);
	// we cannot check the number of rows in x, because it depends on the number of 1's in mask

	// go through entire array for mask;
	// 	if mask(rj)== 1, then insert value from w(rj), into x and advance pointer into x
	// advance pointers into mask and w

	FloatComplex *wp;
	bool *maskp;
	Float *xp;
	int i;

	wp = w.data;
	maskp = mask->data;
	xp = x->data;
	int count = mask->nX * mask->nY * mask->nT;
	for (i=0; i< count; i++) {
		if (*maskp) { // if (creal(*maskp) > 0.0) {
			*xp = *wp;
			xp++;
		}
		maskp++;
		wp++;
	}
}

/**
 * Flatten a map-depth matrix.
 * | xy_0,0 xy_0,1 ...   xy_0,mi  |
 * | xy_1,0 xy_1,1 ...   xy_1,mi  |
 * | ...                          |
 * | xy_mj,0 xy_mj,1 ... xy_mj,mi |
 *
 * Used to flatten lambda, e.g.
 * | L_aa L_ab |
 * | L_ba L_bb |
 *
 * @param[in] tdm matrix to flatten.
 * @return 2D matrix
 */

MATRIX TdmFlatten(MAP_DEPTH_MATRIX tdm) {
	MATRIX flatM;

	int nMaps = tdm.nT;
	flatM = newComplexMatrix(nMaps * tdm.nX, nMaps * tdm.nY);
	FloatComplex *tdm2Data;
	tdm2Data = flatM.data;

	int a, b;
	for (a = 0; a < tdm.nT; a++) {
		for (b = 0; b < tdm.nZ; b++) {
			MATRIX lambdaAB;
			lambdaAB = extractXYPlaneFromMapDepthMatrix(a, b, tdm);
			FloatComplex *lambdaABData;
			lambdaABData = lambdaAB.data;
			int submatrix;
			submatrix = a * ((lambdaAB.nRow*lambdaAB.nCol) * tdm.nZ) + b * lambdaAB.nCol;
			int r, rowStart;
			// for each row in lambdaAB
			for (r=0; r<lambdaAB.nRow; r++) {
				rowStart = r*lambdaAB.nCol;
				int idx;
				idx = submatrix + r * (tdm.nZ* lambdaAB.nCol);
				// memcpy(to, from, size)
				memcpy(&tdm2Data[idx], &lambdaABData[rowStart], (lambdaAB.nCol* sizeof(FloatComplex)));
			}
			freeMatrix(lambdaAB);
		}
	}
	return flatM;
}


void TdmRebuild(MATRIX m, MAP_DEPTH_MATRIX *tdmPtr) {

	MAP_DEPTH_MATRIX tdm;
	tdm = *tdmPtr;

	FloatComplex *tdm2Data;
	tdm2Data = m.data;

	MATRIX lambdaAB;
	lambdaAB = newComplexMatrix(tdm.nX, tdm.nY);
	int a, b;
	for (a = 0; a < tdm.nT; a++) {
		for (b = 0; b < tdm.nZ; b++) {
			FloatComplex *lambdaABData;
			lambdaABData = lambdaAB.data;
			int submatrix;
			submatrix = a * ((lambdaAB.nRow*lambdaAB.nCol) * tdm.nZ) + b * lambdaAB.nCol;
			int r, rowStart;
			// for each row in lambdaAB
			for (r=0; r<lambdaAB.nRow; r++) {
				rowStart = r*lambdaAB.nCol;
				int idx;
				idx = submatrix + r * (tdm.nZ* lambdaAB.nCol);
				// memcpy(to, from, size)
				memcpy(&lambdaABData[rowStart], &tdm2Data[idx], (lambdaAB.nCol* sizeof(FloatComplex)));
			}
			insertXYPlaneIntoMapDepthMatrix(lambdaAB, a, b, tdmPtr);
		}
	}
}

MAP_DEPTH_MATRIX XYTransposeMapDepthMatrix(const MAP_DEPTH_MATRIX *K) {
	MAP_DEPTH_MATRIX KT = newMapDepthMatrixV(K->nX, K->nY, K->nZ, K->nT, K->nV, K->shift);

	int m, z, v;
	for (m=0; m<K->nT; m++) {
		for (v=0; v<K->nV; v++) {
			for (z=0; z<K->nZ; z++) {
				MATRIX xyPlane = extractXYPlaneFromMapDepthMatrixV(m, z, v, *K);
				MATRIX xyPlaneT = transpose(xyPlane);
				insertXYPlaneIntoMapDepthMatrixV(xyPlaneT, m, z, v, &KT);
				freeMatrix(xyPlane);
				freeMatrix(xyPlaneT);
			}
		}
	}

	return KT;

}
