/*
 * MapDepthMatrix.h
 *
 *  Created on: May 21, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#ifndef MAPDEPTHMATRIX_H_
#define MAPDEPTHMATRIX_H_

#include <complex.h>
#include "objectfits.h"
#include "matrix.h"
#include "Mask.h"

/**
 * @ingroup MapDepthMatrix
 */

/**
 * Four dimensional matrix that keeps complete input and output data.
 * This is a main datatype for the 3d problem.  The input kernel data,
 * input map data, solution data, averaging kernel data and covariance
 * data are all kept in a matrix4d object.  During the inversion,
 * pointers to 2d sub-matrices are passed around and
 * inserted to accumulate full solutions.  Map-depth data is kept
 * contiguous to make the inversion run faster.  This is done at the
 * expense of data in the xy-plane.  Since input and output to FITS
 * files uses the xy-plane, the expense is incurred during I/O which
 * is inherently slow anyway.
 * @ingroup MapDepthMatrix
 */
typedef struct {
	FloatComplex *data;  	///< all data values are complex
	int nX;					///< number of rows in xy-plane
	int nY;					///< number of columns in xy-plane
	int nZ;  				///< depth
	int nT;					///< number of measurements
	int nV;					///< if each data point is now a vector, then nV is the dimension of the vector usually (x,y,z)
	MatrixType mtype;		///< type of matrix: always complex
	bool shift; 			///< boolean: whether to shift data after inverse Fourier Transform
	int nYOriginal;         ///< original dimension of y, because sometimes the xy-plane has been Fourier transformed

} MAP_DEPTH_MATRIX;

/**
 * DataSpace
 * @ingroup MapDepthMatrix
 */
enum DataSpace {REAL_SPACE, WAVE_SPACE};

#endif /* MAPDEPTHMATRIX_H_ */

MAP_DEPTH_MATRIX newMapDepthMatrix(int x, int y, int z, int t, bool shift);
MAP_DEPTH_MATRIX newMapDepthMatrixV(int x, int y, int z, int t, int v, bool shift);

void destroyMapDepthMatrix(MAP_DEPTH_MATRIX td);

bool isEmptyMapDepthMatrix(const MAP_DEPTH_MATRIX *td);

MATRIX extractMapDepth(int kx, int ky, MAP_DEPTH_MATRIX td);

void insertMapDepth(int kx, int ky, MATRIX m, MAP_DEPTH_MATRIX *td);

MATRIX extractXYPlaneFromMapDepthMatrix(int map, int depth, MAP_DEPTH_MATRIX td);
MATRIX extractXYPlaneFromMapDepthMatrixV(int map, int depth, int v, MAP_DEPTH_MATRIX td);
MATRIX extractXYPlaneFromDoubleArray(int map, int depth, int v, int nX, int nY, int nZ, int nT, int nV, Float *td);

void insertXYPlaneIntoMapDepthMatrix(MATRIX xyPlane, int map, int depth, MAP_DEPTH_MATRIX *td);
void insertXYPlaneIntoMapDepthMatrixV(MATRIX xyPlane, int map, int depth, int v, MAP_DEPTH_MATRIX *td);
void insertXYPlaneIntoDoubleArray(MATRIX xyPlane, int map, int depth, int v, int nX, int nY,
		int nZ, int nT, int nV, Float *tdptr);

MATRIX extractXYPlaneColMajorMapDepthMatrix(int map, int depth, const MAP_DEPTH_MATRIX *td);
MATRIX extractXYPlaneColMajorMapDepthMatrixV(int map, int depth, int v, const MAP_DEPTH_MATRIX *tdptr);
void insertXYPlaneColMajorMapDepthMatrix(MATRIX xyPlane, int map, int depth, MAP_DEPTH_MATRIX *tdptr);
void insertXYPlaneColMajorMapDepthMatrixV(MATRIX xyPlane, int map, int depth, int v, MAP_DEPTH_MATRIX *tdptr);

MAP_DEPTH_MATRIX readMapDepthMatrix(char *basedir, char *filenames[], int nFiles, bool useFT);
MAP_DEPTH_MATRIX readMapDepthMatrixKernel(char *basedir, char *filenames[], int nFiles, bool useFT);
MAP_DEPTH_MATRIX readMapDepthMatrixV(char *basedir, char *filenames[], int nFiles, int vdim, bool useFT);
MAP_DEPTH_MATRIX readMapDepthMatrixFromOneFile(char *filename, bool useFT);
MAP_DEPTH_MATRIX readMapDepthMatrixFromOneFileS(char *filename, bool useFT, bool shiftOnInput, bool shiftOnOutput);
MAP_DEPTH_MATRIX readMapDepthMatrixForTarget(char *filename, bool useFT);

void writeMapDepthMatrix(char *filenames[], int nFiles, MAP_DEPTH_MATRIX tdm,
		bool useFT, FITSParameters headerParameters);

void writeDepthMapMatrix(char *filenames[], int nFiles, MAP_DEPTH_MATRIX tdm, bool useFT, FITSParameters headerParameters);


void writeAvgKernelMatrix(char *filenames[], int nFiles, MAP_DEPTH_MATRIX tdm, bool useFT, Float lambda);

void writeMapDepthMatrixToOneFile(char *filename, MAP_DEPTH_MATRIX tdm,
		bool useFT, FITSParameters headerParameters);

MAP_DEPTH_MATRIX inverseMapDepth(MAP_DEPTH_MATRIX tdm);

MAP_DEPTH_MATRIX convertMapDepth(const MAP_DEPTH_MATRIX *tdm, MatrixOrder order);
MAP_DEPTH_MATRIX convertMapDepthSpace(const MAP_DEPTH_MATRIX *tdm, enum DataSpace toSpace);

void TdmMultiplyScalarTakeConjugate(MAP_DEPTH_MATRIX *tdmPtr, Float scalar);
void TdmMultiplyScalar(MAP_DEPTH_MATRIX *tdmPtr, Float scalar);

void MdmMultiplyMatrix(MAP_DEPTH_MATRIX *tdmPtr, MATRIX theta);

void ZeroK0(MAP_DEPTH_MATRIX *tdmPtr);

MAP_DEPTH_MATRIX TdmCopy(MAP_DEPTH_MATRIX tdm);

void pack(const MAP_DEPTH_MATRIX *w_n, MATRIX *x, const MASK *mask);
void unpack(MATRIX *x, const MASK *mask, MAP_DEPTH_MATRIX *w);

void pack2(const MAP_DEPTH_MATRIX *w_n, MATRIX *x, const MAP_DEPTH_MATRIX *mask);
void unpack2(const MAP_DEPTH_MATRIX *in, const MAP_DEPTH_MATRIX *mask, MAP_DEPTH_MATRIX *out);
void unpack3(const MAP_DEPTH_MATRIX *in, const MAP_DEPTH_MATRIX *mask, MAP_DEPTH_MATRIX *out);

MATRIX TdmFlatten(MAP_DEPTH_MATRIX tdm);
void TdmRebuild(MATRIX m, MAP_DEPTH_MATRIX *tdm);

MAP_DEPTH_MATRIX XYTransposeMapDepthMatrix(const MAP_DEPTH_MATRIX *K);
