/*
 * Mask.c
 *
 *  Created on: Jan 7, 2013
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include "Mask.h"
#include "matrix.h"
#include "MatrixFunctionErrorMessages.h"
#include "errorHandler.h"

/** Creates a new 3D matrix and allocates space for the data of type boolean
 * @param[in] r number of rows
 * @param[in] c number of columns
 * @param[in] m number of masks; corresponds to the number of maps
 * @return A new matrix
 * @ingroup Matrix
 */
MASK newMask(int r, int c, int m) {
   MASK mask;

   int maskLength = r * c * m;
   mask.data = (bool *) td_malloc(sizeof(bool) * maskLength);
   mask.nX = r;
   mask.nY = c;
   mask.nT = m;
   return mask;
}

/**
 * Free memory used to store data in mask
 * @param[in] m A mask
 * @ingroup Mask
 */
void destroyMask(MASK *m) {
	if (m->data != NULL)
		td_free(m->data, m->nX * m->nY * m->nT * sizeof(bool));
}
/**
 * Test if Mask is empty.
 * @param[in] m Pointer to a 3D MASK matrix.
 * @return true if empty; false otherwise.
 */
bool isEmptyMask(const MASK *m) {
	return(m->data == NULL);
}

/** Opens and reads FITS files, then stores the data in a new MASK.
* @param[in] basedir base directory for all filenames
* @param[in] filenames list of FITS files to read.
* @param[in] nFiles number of files to read
* @return new MASK containing data from files.
* @ingroup MapDepthMatrix
*/
MASK readMasks(char *basedir, char *filenames[], int nFiles) {

	MultiDimFits fileInfo;
	fits2d onePlane;
	MATRIX xyMask;
	int map;
	int t, v;
	MASK xymMask;
	char *fullName;
	int dirPathEnd;
	int fullNameLength;
	WhichPlane whichPlane;
    int vdim;

	map = nFiles;
	xymMask.data = NULL;
	vdim = 1;

	fullNameLength = (longestName(filenames, nFiles*vdim) + strlen(basedir) + 2) * sizeof(char);
	fullName = (char *) td_malloc(fullNameLength);
	strcpy(fullName, basedir);
	if (!endsWith(basedir, "/") && (strlen(fullName) > 0)) {
		strcat(fullName, "/");
	}
	dirPathEnd = strlen(fullName);
	// for each mask file
	for (t = 0; t < map; t++) {
		for (v = 0; v < vdim; v++) {
			strcpy(&fullName[dirPathEnd], filenames[t*vdim + v]);
			printf("%s\n", fullName);
			// open the mask file and read one plane
			fileInfo = openMultiDimFitsRead(fullName);

			// need to know the dimensions of x,y and z before allocating the matrix
			// the dimensions are specified in the fits header
			int ydim = fileInfo.max.y;

			if (xymMask.data == NULL) {
				// first time through the loop, allocate space based on dimensions of
				// mask file

				xymMask = newMask(fileInfo.max.x, ydim, map);
			} else {
				// check for compatible dimensions
				if ((fileInfo.max.x != xymMask.nX) || (fileInfo.max.y != xymMask.nY))
					reportError(MaskReadUnexpectedDim, filenames[t*vdim + v],
							FATAL);
			}
			int d;
			for (d = 1; d <= fileInfo.max.z; d++) {

				// get next mask plane
				whichPlane.z = d;
				whichPlane.t = t+1;
				onePlane = readNextPlane(fileInfo, whichPlane);

				xyMask = newMatrixG(onePlane.xDim, onePlane.yDim, BOOL);
				bool *p;
				Float *q;
				p = xyMask.data;
				q = onePlane.data;
				int pi;
				for (pi = 0; pi < xyMask.nRow * xyMask.nCol; pi++)
					*p++ = *q++ != 0.0;  // assign true or false

				insertXYPlaneIntoMask(xyMask, t, &xymMask);
				// It would be nice to combine this operation with the above for loop
				// but, I'm going to choose easy of maintenance over speed and keep
				// the operation separate, and similar to the operation on MAP_DEPTH_MATRIX.

				freeMatrix(xyMask);
			} // end for each plane of kernel file

			// close the FITS file
			closeMultiDimFits(fileInfo);
		} // end for each vdim
	} // end for each map dim
	td_free(fullName, fullNameLength);
	return xymMask;
}

/**
 * Insert 2D boolean matrix into 3D mask matrix.  The 2D matrix is the xy-mask for
 * a map.
 * @param[in] xyPlane A single mask which is a 2D MATRIX of boolean for a single map.
 * @param[in] map The mask corresponds to this map number.
 * @param[in,out] tdptr A pointer to a 3D MASK matrix. The single mask is inserted into
 *                this matrix.
 * @ingroup Mask
 */
void insertXYPlaneIntoMask(MATRIX xyPlane, int map, MASK *tdptr) {
	// matrix is an array (maps) of 2D arrays (x and y); we want map-distance plane easily extracted, for
	// the time-distance inversions.
	// It will be awkward to insert the xy-plane; we'll do one element at a map
	bool *p;

	MASK td;

	int x;

	if ((xyPlane.nRow != tdptr->nX) || (xyPlane.nCol != tdptr->nY))
		reportError("Incompatible sizes", "insertXYPlaneIntoMapDepthMatrixV",
				FATAL);

	td = *tdptr;
	x = xyPlane.nRow;
	p = xyPlane.data;

	bool *tp;
	int initialPosition, offset;
	initialPosition = map;
	// depth should be zero; nZ = 1, nV = 1, v = 0
	//initialPosition = map * (td.nZ * td.nV) + v * (td.nZ) + depth;
	// offset = M*Nz*Nv
	offset = td.nT;
	tp = td.data + initialPosition;
	for (x = 0; x < xyPlane.nRow * xyPlane.nCol; x++) {
		*tp = *p++;
		tp += offset;
	}
}

/**
 * Extract the mask associated with a map.
 * @param[in] map Which map number.
 * @param[in] masks The 3D matrix containing masks for all maps.
 * @return The mask as a 2D MATRIX of boolean.
 * @ingroup Mask
 *
 */
MATRIX extractXYPlaneFromMask(int map, MASK masks) {
	// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
	// the map-distance inversions.
	// depth is always one, because the masks have nZ = 1
	// It will be awkward to extract the xy-plane; we'll do one element at a time
	MATRIX xyPlane;
	bool *p;
	int x;

	xyPlane = newMatrixG(masks.nX, masks.nY, BOOL);
	x = xyPlane.nRow;
	p = xyPlane.data;

	bool *tp;
	int initialPosition, offset;
	//initialPosition = map*(masks.nZ*masks.nV) + v*(masks.nZ) + depth;
	// depth should be zero; nZ = 1, nV = 1; v = 0
	initialPosition = map;
	// offset = M*Nz*Nv
	offset = masks.nT;
	tp = masks.data + initialPosition;
	for (x=0; x<xyPlane.nRow*xyPlane.nCol; x++) {
		*p++ = *tp;
		tp += offset;
	}
	return xyPlane;
}

/** Extract a 2d sub matrix from a 4d matrix.  Really, this returns a pointer
 * to a 2d matrix inside the 4d matrix.   The pointer is wrapped by a 2d matrix object.
 * @param[in] kx Row index in xy-plane
 * @param[in] ky Column index in xy-plane
 * @param[in] masks 3D matrix from which to extract.
 * @return 2D matrix indexed by map (rows) and depth (columns). MATRIX(map, depth).
 * MATRIX.data is a pointer into the 3D matrix.  No data is copied, only a pointer is provided.
 * @ingroup MapDepthMatrix
 */
MATRIX extractMapZ1FromMask(int kx, int ky, const MASK *masks) {

	MATRIX m;

	int nCol1, nCol2;
	int xyidx;
	bool *dataPtr;

	nCol1 = masks->nY * masks->nT;
	nCol2 = masks->nT;
	xyidx = kx * nCol1 + ky * nCol2;
	dataPtr = masks->data + xyidx;
	m = newMatrixGI(masks->nT, 1, dataPtr, BOOL); // construct new matrix and initialize it with data
	return m;
}

/**
 * Print the masks.
 * @param[in] mask A pointer to a MASK matrix.
 */
void printMask(const MASK *mask) {

   int r,c,m;

   bool *maskDataPtr;
   maskDataPtr = mask->data;
   for (m=0; m<mask->nT; m++) {
	   printf("mask for map %d\n", m);
	   for (r=0; r<mask->nX; r++) {
		   for (c=0; c<mask->nY; c++) {
				if (*maskDataPtr++) {
					//include point
					printf("1");
				} else {
					printf("0");
				}
		   }
		   printf("\n");
	   }
	   printf("\n");
   }
}


