/*
 * SuperY.c
 *
 *  Created on: Jun 16, 2011
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include <stdbool.h>
#include <complex.h>
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "utilities.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "SuperY.h"
#include "fftw.h"

/** Initialize super vector Y.
 * @param[in,out] y Pointer to SuperVectorY with allocated space; return initialized vector.
 * @param[in] m Defining kernel matrix in wave space
 */
void InitSuperVectorY(SuperVectorY *y, MAP_DEPTH_MATRIX m) {
	y->Nx = m.nX;
	y->Ny = m.nX;  // (m.nY - 1) * 2;
	y->Nz = m.nZ;
	y->Nm = 1; // because Y contains averaging kernels and they are independent of maps
	y->Nv = m.nV;
	y->noiseBase = y->Nx * y->Ny * y->Nz * y->Nv;
	y->RC = y->Nx * y->Ny;
}



/** Convert super vector Y (which is Float) to FloatComplex.
 * @param[in] rData Pointer to Float array for real data; must be allocated.
 * @param[in] nelements Number of elements in rData.
 * @param[in,out] cData Pointer to FloatComplex data; must be allocated.  Complex
 * form of real data returned in this array.
 */
void RealToDoubleComplex(Float *rData, int nelements, FloatComplex *cData) {

	if (rData == NULL)
		reportError(SuperYRDataNull, "RealToDoubleComplex", FATAL);
	if (cData == NULL)
		reportError(SuperYCDataNull, "RealToDoubleComplex", FATAL);
	int i;
	Float *dp;
	dp = rData;
	FloatComplex *dcp;
	dcp = cData;
	for (i = 0; i < nelements; i++) {
		*dcp++ = *dp++ + 0.0 * I;
	}
}

/** Convert from FloatComplex to Float
 * @param[in] cData Pointer to array of FloatComplex data; cannot be empty.
 * @param[in] nelements Number of elements in cData.
 * @param[in,out] rData Pointer to array of Floats; must be allocated.  Real
 * form of complex data returned in this array.
 * @return FALSE if error; TRUE if successful
 */
bool DoubleComplexToReal(FloatComplex *cData, int nelements, Float *rData) {

	bool conversionOk = TRUE;
	if (rData == NULL)
		reportError(SuperYRDataNull, "DoubleComplexToReal", FATAL);
	if (cData == NULL)
		reportError(SuperYCDataNull, "DoubleComplexToReal", FATAL);
	int i;
	Float *dp;
	dp = rData;
	FloatComplex *dcp;
	dcp = cData;
	for (i = 0; i < nelements; i++) {
		if (cimag(*dcp) != 0.0) {
			if (conversionOk) {
				char msg[MAX_ERROR_MSG_LENGTH];
				snprintf(msg, MAX_ERROR_MSG_LENGTH, SuperYCannotConvertCToI, i, creal(*dcp), cimag(*dcp));
				reportError(msg, "DoubleComplexToReal", ERROR);
				conversionOk = FALSE;
			}
		}
		*dp++ = creal(*dcp++); // ignoring imaginary part
	}
	return conversionOk;
}

void print_superY(SuperVectorY *superY) {

	int i;
	printf("\nAvgKern = [");
	for (i=0; i<superY->noiseBase; i++)
		printf("%le ", superY->yElements[i]);
	printf("];\nNoise = [");
	for (i=superY->noiseBase; i<superY->yLength; i++)
		printf("%le ", superY->yElements[i]);
	printf("];\n");
}

/**
 * Reconstitute averaging kernel and noise matrices from packed super vector Y.
 * @param[in] superY is in column-major order.
 * y must be in real space, NOT wave space.
 * @param[in,out] AvgKern is in real space. Data is overwritten. Send already allocated matrix.
 * @param[in,out] noise is in real space. Data is overwritten. Send already allocated matrix.
 * @param[in] waveSpace If true, then averaging kernel and noise are in wave space; NOT IMPLEMENTED!
 * otherwise, they are in real space.
 */
void unpackSuperYOld(const SuperVectorY *superY, MAP_DEPTH_MATRIX *AvgKern, MAP_DEPTH_MATRIX *noise, bool waveSpace) {

	int Nx, NyOrig, Nz, Nv, Nt;

	Nx = AvgKern->nX;
	NyOrig = AvgKern->nY;
	Nz = AvgKern->nZ;
	Nt = AvgKern->nT;
	Nv = AvgKern->nV;

	// unpack averaging kernel first
	// pull from superY; superY is always in real space and Float
	int nelements;
	nelements = Nx*NyOrig*Nz*Nt*Nv;
	RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data);

	// pull noise from superY next

	// convert from complex to real
	nelements = noise->nX * noise->nY * noise->nZ * noise->nT * noise->nV;
	RealToDoubleComplex(&(superY->yElements[superY->noiseBase]), nelements, noise->data);
}

/**
 * Condense averaging kernel and noise matrices into packed super vector Y.
 * [AvgKern; noise] averaging kernel is at the top of Y, noise is in the bottom.
 * @param[in,out] AvgKern is in real space. Data is overwritten.
 * @param[in,out] noise is in real space. Data is overwritten.
 * @param[in] superY is in column-major order. ??  Send already allocated matrix.
 * y must be in real space, NOT wave space.
 * @param[in] realSpace If true, then averaging kernel and noise are in real space; NOT IMPLEMENTED!
 * otherwise, they are in wave space.
 * @return false if packing failed; true if successful.
 */
bool packSuperYOld(const MAP_DEPTH_MATRIX *AvgKern, const MAP_DEPTH_MATRIX *noise, SuperVectorY *superY, bool realSpace) {

	bool packOk = TRUE;
	// pack averaging kernel first

	// convert from complex to real
	int nelementsAK;
	nelementsAK = AvgKern->nX * AvgKern->nY * AvgKern->nZ * AvgKern->nT * AvgKern->nV;

	int nelementsNoise;
	nelementsNoise = noise->nX * noise->nY * noise->nZ * noise->nT * noise->nV;

	if (nelementsAK + nelementsNoise > superY->yLength) {
		reportError(SuperYAvgKNoiseExceed, "packSuperY", ERROR);
		packOk = FALSE;
	} else if (!DoubleComplexToReal(AvgKern->data, nelementsAK, superY->yElements)) {
		reportError(SuperYAvgKIToReal,"packSuperY", ERROR);
		packOk = FALSE;
	} else {

		// pack noise second

		// convert from complex to real

		int noiseBaseIdx;
		noiseBaseIdx = superY->noiseBase;
		if (!DoubleComplexToReal(noise->data, nelementsNoise, &(superY->yElements[noiseBaseIdx]))) {
			reportError(SuperYNoiseIToReal,"packSuperY", ERROR);
			packOk = FALSE;
		}
	}
	return packOk;

}

/**
 * Rebuild averaging kernel and noise matrices from packed super vector Y.
 * @param[in] superY is in row-major order for each map-depth matrix (for each kx, ky).
 * y must be in real space, NOT wave space.
 * @param[in,out] AvgKern is returned in wave space. Data is overwritten. Send already allocated matrix.
 * @param[in,out] noise is returned in wave space. Data is overwritten. Send already allocated matrix.
 * @param[in] outSpace Enumeration: AvgKern and noise are returned in either REAL_SPACE or WAVE_SPACE
 */
// unpack SuperY and move to wave space.  Return AvgKern and noise in wave space.
// Assume superY is contiguous for each kxy.
// Currently, superY is packed contiguous for each map-depth matrix.
void unpackSuperY(const SuperVectorY *superY, MAP_DEPTH_MATRIX *AvgKern, MAP_DEPTH_MATRIX *noise,  enum DataSpace outSpace) {

	int Nx, NyOrig, Nz, Nv, Nt;
	bool shift;

	if (outSpace == WAVE_SPACE) {

		Nx = AvgKern->nX;
		NyOrig = AvgKern->nX; // (AvgKern->nY -1) * 2;
		Nz = AvgKern->nZ;
		Nt = AvgKern->nT;
		Nv = AvgKern->nV;
		shift = AvgKern->shift;

		// unpack averaging kernel first
		// pull from superY; superY is always in real space and Float
		//RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		//--------------
		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which stores map-depth matrix contiguous).
		// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
		// the map-distance inversions.
		// It will be awkward to extract the xy-plane; we'll do one element at a time
		int t, v, d;
		for (t=0; t<Nt; t++) {  // for each map
			for (v=0; v<Nv; v++) {  // for each scatterer
				for (d=0; d<Nz; d++) {  // for each depth, z
					MATRIX xyPlane, xyPlaneK;

					xyPlane = extractXYPlaneFromDoubleArray(t, d, v, Nx, NyOrig, Nz, Nt, Nv, superY->yElements);

									if (shift) {
										MATRIX temp = fftwShift(xyPlane);
										freeMatrix(xyPlane);
										xyPlaneK = fftw(MatrixToFitsObject(temp)); // no new memory allocated
										freeMatrix(temp);

									} else {
					xyPlaneK = fftw(MatrixToFitsObject(xyPlane));
					freeMatrix(xyPlane);
									}
					//print_matrix2("unpack AvgKern Wave", xyPlaneK, ROW_MAJOR);
					//insert xyPlane into AvgKern MAP_DEPTH_MATRIX
					// this should be complex MATRIX inserted into complex MAP_DEPTH
					insertXYPlaneIntoMapDepthMatrixV(xyPlaneK, t, d, v, AvgKern);
					fftw_free(xyPlaneK.data);
				}
			}
		}

		// pull noise from superY next
		//RealToDoubleComplex(&(superY->yElements[superY->noiseBase]), nelements, noise->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which is stores map-depth matrixes contiguously).
		// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
		// the map-distance inversions.
		// It will be awkward to extract the xy-plane; we'll do one element at a time

		Nx = noise->nX;
		NyOrig = noise->nX;
		Nz = noise->nZ;
		Nt = noise->nT;
		Nv = noise->nV;
		shift = noise->shift;

		// unpack averaging kernel first
		// pull from superY; superY is always in real space and Float
		//RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		//--------------
		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which stores map-depth matrix contiguous).
		// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
		// the map-distance inversions.
		// It will be awkward to extract the xy-plane; we'll do one element at a time
		for (t=0; t<Nt; t++) {  // for each map
			for (v=0; v<Nv; v++) {  // for each scatterer
				for (d=0; d<Nz; d++) {  // for each depth, z
					MATRIX xyPlane, xyPlaneK;

					xyPlane = extractXYPlaneFromDoubleArray(t, d, v, Nx, NyOrig, Nz, Nt, Nv, &(superY->yElements[superY->noiseBase]));

					if (shift) {
						MATRIX temp = fftwShift(xyPlane);
						freeMatrix(xyPlane);
						xyPlaneK = fftw(MatrixToFitsObject(temp)); // no new memory allocated
						freeMatrix(temp);
					} else {
						xyPlaneK = fftw(MatrixToFitsObject(xyPlane));  // no new memory allocated
						freeMatrix(xyPlane);
					}
					//insert xyPlane into AvgKern MAP_DEPTH_MATRIX
					insertXYPlaneIntoMapDepthMatrixV(xyPlaneK, t, d, v, noise);
					fftw_free(xyPlaneK.data);
				}
			}
		}
	} else {
		unpackSuperYOld(superY, AvgKern, noise, FALSE);
	}
}

/**
 * Condense averaging kernel and noise matrices into packed super vector Y.
 * [AvgKern; noise] averaging kernel at the top of Y, noise at the bottom.
 * @param[in] AvgKern is in wave space.
 * @param[in] noise is in wave space.
 * @param[in, out] superY is in row-major order, with contiguous map-depth matrixes, which means
 * the kx,ky pixels are all stored contiguously in row-major order.  Send already allocated matrix.
 * y must be in real space, NOT wave space.
 * @param[in] inSpace Enumeration: AvgKern and noise are in either REAL_SPACE or WAVE_SPACE
 * @return false if packing failed; true if successful.
 */
/*
// aka pack SuperY and move to real space.  Receive AvgKern and noise in wave space.
// Return superY contiguous for each kxy, in real space.
// Currently, superY is packed contiguous for each  map-depth matrix.
 *
 */
bool packSuperY(const MAP_DEPTH_MATRIX *AvgKern, const MAP_DEPTH_MATRIX *noise, SuperVectorY *superY,
		enum DataSpace inSpace) {

	if (inSpace == WAVE_SPACE) {

		int Nx, NyOrig, Nz, Nv, Nt;

		Nx = AvgKern->nX;
		NyOrig = AvgKern->nX;
		Nz = AvgKern->nZ;
		Nt = AvgKern->nT;
		Nv = AvgKern->nV;

		// unpack averaging kernel first
		// pull from superY; superY is always in real space and Float
		// RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		//--------------
		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which stores map-depth matrix contiguous).
		// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
		// the map-distance inversions.
		// It will be awkward to extract the xy-plane; we'll do one element at a time
		int t, v, d;
		for (t=0; t<Nt; t++) {  // for each map
			for (v=0; v<Nv; v++) {  // for each scatterer
				for (d=0; d<Nz; d++) {  // for each depth, z
					MATRIX xyPlane, xyPlaneK;

					xyPlaneK = extractXYPlaneFromMapDepthMatrixV(t, d, v, *AvgKern);
					xyPlane = fftwReverse(xyPlaneK, AvgKern->shift);
					freeMatrix(xyPlaneK);
					//insert xyPlane into AvgKern MAP_DEPTH_MATRIX
					insertXYPlaneIntoDoubleArray(xyPlane, t, d, v, Nx, NyOrig, Nz, Nt, Nv, superY->yElements);
					freeMatrix(xyPlane);
				}
			}
		}

		// next, pull noise from superY

		//RealToDoubleComplex(&(superY->yElements[superY->noiseBase]), nelements, noise->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which is stores map-depth matrixes contiguously).
		// td matrix is a 2d array of 2d arrays; we want map-distance plane easily extracted, for
		// the map-distance inversions.
		// It will be awkward to extract the xy-plane; we'll do one element at a time

		Nx = noise->nX;
		NyOrig = noise->nX;
		Nz = noise->nZ;
		Nt = noise->nT;
		Nv = noise->nV;

		// unpack averaging kernel first
		// pull from superY; superY is always in real space and Float
		//RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data); replaced by the following
		// code which comes from extractXYPlaneFromMapDepthMatrix

		// for each xyplane, extractXYPlaneFromMapDepthMatrix (which stores map-depth matrix contiguous).
		// td matrix is a 2d array of 2d arrays
		// extract the xy-plane one element at a time
		for (t=0; t<Nt; t++) {  // for each map
			for (v=0; v<Nv; v++) {  // for each scatterer
				for (d=0; d<Nz; d++) {  // for each depth, z
					MATRIX xyPlane, xyPlaneK;

					xyPlaneK = extractXYPlaneFromMapDepthMatrixV(t, d, v, *noise);
					xyPlane = fftwReverse(xyPlaneK, noise->shift);
					freeMatrix(xyPlaneK);

					//insert xyPlane into superY
					insertXYPlaneIntoDoubleArray(xyPlane, t, d, v, Nx, NyOrig, Nz, Nt, Nv, &(superY->yElements[superY->noiseBase]));
					freeMatrix(xyPlane);
				}
			}
		}

	} else {
		// AvgKern and noise are in real space.
		// just convert from FloatComplex to Float and insert the pointers
		bool value =  packSuperYOld(AvgKern, noise, superY, TRUE);
		//print_superY(superY);
		return value;

	}
	//print_superY(superY);
	return TRUE;
}



