/*
 * fftw.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include <stdio.h>
#include <math.h>
// include complex.h and fftw will use the native c complex datatype (FloatComplex);
// then fftwf_complex is defined to be the native complex type
#include <complex.h>
#include <fftw3.h>
#include "utilities.h"
#include "objectfits.h"
#include "matrix.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"

#include <time.h>

#include "fftw.h"

#ifndef SINGLE
#define fftwf_complex      fftw_complex
#define fftwf_plan         fftw_plan
#define fftwf_malloc       fftw_malloc
#define fftwf_free       fftw_free
#define fftwf_destroy_plan fftw_destroy_plan
#define fftwf_import_system_wisdom fftw_import_system_wisdom
#define fftwf_import_wisdom_from_file fftw_import_wisdom_from_file
#define fftwf_plan_dft_r2c_2d fftw_plan_dft_r2c_2d
#define fftwf_plan_dft_c2r_2d fftw_plan_dft_c2r_2d
#define fftwf_execute_dft_c2r fftw_execute_dft_c2r
#define fftwf_execute_dft_r2c fftw_execute_dft_r2c

#endif
     



	bool reversePlanDone = false;	///< fftw plans are reused as long as the size of input and output data remain the same.
	fftwf_plan fplaninverse = NULL;	///< fftw reverse plan
	int RinX = 0; 					///< x dimension of input data for reverse plan
	int RinY = 0;					///< y dimension of input data for reverse plan

MATRIX fftwShift(MATRIX m);
MATRIX ifftwShift(MATRIX m);

/** fftw Performs a forward Fourier transform using the Fasted Fourier Transform in
 * the West (FFTW) library.
 * @param[in] function A fits2d object which contains data values for a function to transform.
 * @return A MATRIX object containing the transformed data.
 * @ingroup Matrix
 */
// use fftwf_malloc to create input function data
MATRIX fftw(fits2d function) {
	Float *in;
	fftwf_complex *out;
	static bool forwardPlanDone = false;
	static fftwf_plan fplanforward = NULL;
	static int inX = 0, inY = 0;
	Float norm;
	int x;
	
	// transform the data

	in = function.data;

	out = (fftwf_complex *) fftwf_malloc(function.xDim * (function.yDim / 2 + 1)
			* sizeof(fftwf_complex));
	if (out == NULL)
		reportError(OutOfMemory, "fftw", FATAL);

	// make a new plan, if one doesn't exist or the dimensions have changed
	if ((!forwardPlanDone) || (inX != function.xDim) || (inY != function.yDim)) {
		if (fplanforward != NULL)
			fftwf_destroy_plan(fplanforward);
		
#ifdef USE_WISDOM
			if (fftwf_import_system_wisdom() == 0)
				reportError(FftwSystemWisdomImport, "fftw", ERROR);
			FILE *wisdomFP;
			wisdomFP = fopen("mywisdom", "r");
			if (wisdomFP == 0) {
				reportError(FftwWisdomImport, "fftw", ERROR);
			} else {
				printf("opened wisdom file\n");
				if (fftwf_import_wisdom_from_file(wisdomFP) == 0) {
					reportError(FftwWisdomRead, "fftw", ERROR);
				} else {
					printf("imported wisdom\n");
				}
				fclose(wisdomFP);
			}
			fplanforward = fftwf_plan_dft_r2c_2d(function.xDim, function.yDim, in,
					out,
					FFTW_MEASURE | FFTW_WISDOM_ONLY);
			if (fplanforward == NULL) {
				reportError(FftwForwardPlanNull, "fftw", INFO);
				fplanforward = fftwf_plan_dft_r2c_2d(function.xDim, function.yDim, in, out, FFTW_MEASURE);
			} else {
				printf("using wisdom for forward plan\n");
			}
			forwardPlanDone = true;
			inX = function.xDim;
			inY = function.yDim;

			fplaninverse = fftwf_plan_dft_c2r_2d(function.xDim, function.yDim, out, in,
					FFTW_MEASURE | FFTW_WISDOM_ONLY);
			if (fplaninverse == NULL) {
				reportError(FftwReversePlanNull, "fftw", INFO);
				fplaninverse = fftwf_plan_dft_c2r_2d(function.xDim, function.yDim, out, in,
						 FFTW_MEASURE);

			} else {
				printf("using wisdom for reverse plan\n");
			}
			reversePlanDone = true;
			RinX = function.xDim;
			RinY = function.yDim;

			// The plan is associated with the out pointer;
			// if we change the pointer, then we may need to make a new plan or call fftwf_execute_xxx instead
			// of fftwf_execute();

			wisdomFP = fopen("mywisdom", "a");
			if (wisdomFP == 0) {
				reportError(FftwWisdomWrite, "fftw", ERROR);
			} else {
				fftwf_export_wisdom_to_file(wisdomFP);
			}
			fclose(wisdomFP);
#else
			fplanforward = fftwf_plan_dft_r2c_2d(function.xDim, function.yDim, in,
					out,
					FFTW_ESTIMATE);
			forwardPlanDone = true;
			inX = function.xDim;
			inY = function.yDim;

			fplaninverse = fftwf_plan_dft_c2r_2d(function.xDim, function.yDim, out, in,
					FFTW_ESTIMATE);
			reversePlanDone = true;
			RinX = function.xDim;
			RinY = function.yDim;

			// The plan is associated with the out pointer;
			// if we change the pointer, then we may need to make a new plan or call fftwf_execute_xxx instead
			// of fftwf_execute();
#endif
	}

	// reuse the plan
	fftwf_execute_dft_r2c(fplanforward, in, out);

	norm = 1.0 / ((2.0 * M_PI) * (2.0 * M_PI));

	//scale the data
	for (x=0; x<function.xDim * (function.yDim / 2 + 1); x++)
		out[x] = out[x] * norm;

	// convert out matrix to a MATRIX datatype
	MATRIX result;

	result = newComplexMatrix(function.xDim, (function.yDim / 2 + 1));
	td_free(result.data, function.xDim * (function.yDim / 2 + 1) * sizeof(FloatComplex));
	result.data = out;
	
	return result;

}

/** fftwReverse Performs an inverse Fourier transform using the Fasted Fourier Transform in
 * the West (FFTW) library.
 * @param[in] function A MATRIX object which contains data values for a function to transform.
 * @param[in] shift A boolean value whether to shift the data values.  true means shift the values.
 * false means no shift.
 * @return A MATRIX object containing the transformed data.
 * @ingroup Matrix
 */
MATRIX fftwReverse(MATRIX function, bool shift) {
	Float *in;
	FloatComplex *out;
	static
	FloatComplex *outcopy = NULL;
	int fullSize;

	int x;

	fullSize = function.nRow; // assume the matrix is square; don't use 2*(function.nCol-1);
	// transform the data

	// allocate Float space for output of ifftw
	int inSize = function.nRow * fullSize * sizeof(Float);
	in = (Float*) fftwf_malloc(inSize);
	if (in == NULL)
		reportError(OutOfMemory, "fftwReverse", FATAL);

	out = function.data;

	// get memory corruption here !!!!!!!! for fftwf_plan_dft_c2r_2d
	// we are doing an out-of-place transform.
	// FFTW_PRESERVE_INPUT specifies that an out-of-place transform must not change its input array.
	// This is ordinarily the default, except for c2r and hc2r (i.e. complex-to-real) transforms for which
	// FFTW_DESTROY_INPUT is the default. In the latter cases, passing FFTW_PRESERVE_INPUT will attempt to
	// use algorithms that do not destroy the input, at the expense of worse performance;
	// for multi-dimensional c2r transforms, however, no input-preserving algorithms are implemented and
	// the planner will return NULL if one is requested.

	// make a new plan, if one doesn't exist or the dimensions have changed
	if ((!reversePlanDone) || (RinX != function.nRow) || (RinY != fullSize)) {
		if (fplaninverse != NULL)
			fftwf_destroy_plan(fplaninverse);
#ifdef USE_WISDOM
			fplaninverse = fftwf_plan_dft_c2r_2d(function.nRow, fullSize, outcopy, in,
					 FFTW_WISDOM_ONLY);
			if (fplaninverse == NULL) {
				fplaninverse = fftwf_plan_dft_c2r_2d(function.nRow, fullSize, outcopy, in,
					 FFTW_ESTIMATE);
			} else {
				printf("using wisdom\n");
			}
#else
			fplaninverse = fftwf_plan_dft_c2r_2d(function.nRow, fullSize, outcopy, in,
					 FFTW_ESTIMATE);
#endif

		reversePlanDone = true;
		RinX = function.nRow;
		RinY = function.nCol;
	}

	// reuse the plan
	fftwf_execute_dft_c2r(fplaninverse, out, in);

	// convert out matrix to a MATRIX datatype
	MATRIX result;
	Float *p;

	result = newMatrix(function.nRow, function.nRow); // assume matrix is square; don't use 2*(function.nCol-1));
	td_free(result.data, result.nCol * result.nRow * sizeof(Float));
	result.data = in;

	Float norm = ((2.0 * M_PI) * (2.0 * M_PI));
	p = result.data;
	// scale the data
	for (x=0; x<result.nRow*result.nCol; x++) {
		p[x] =  p[x] / (result.nRow * result.nCol) * norm;
	}

	if (shift) {
		MATRIX shiftedResult;
		if ((result.nRow % 2) == 0) { // even
			shiftedResult = fftwShift(result);
		} else { // odd
			shiftedResult = ifftwShift(result);
		}
		fftwf_free(result.data);
		return shiftedResult;
	} else {
		return result;
	}
}

// shift data from this format
// 2 1
// 4 3
//
// to this format
// 3 4
// 1 2
//
/** fftwShift shifts data values after a reverse transform.  Because of the way FFTW
 * performs the Fourier transform, the data sometimes needs to be shifted.
 * @param[in] m A MATRIX containing data to shift.
 * @return A MATRIX containing shifted data.
 * @ingroup Matrix
 */
MATRIX fftwShift(MATRIX m) {

	MATRIX result;

	if (m.mtype != REAL) {
		reportError(FftwShiftNoComplexData, "", FATAL);
	} else {
		int i;
		int toR, toC, fromR, fromC;
		int halfRow, halfCol;
		Float *p, *q;

		result = newMatrix(m.nRow, m.nCol);
		p = m.data;
		q = result.data;

		if ((m.nRow % 2) == 0) { // even
		halfRow = m.nRow/2;
		halfCol = m.nCol/2;
		for (i=0; i<m.nRow/2; i++) {
			// memcpy(to, from, size)
			// copy quadrant 1
			toR = halfRow + i;
			toC = 0;
			fromR = 0 + i;
			fromC = halfCol;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(Float));
			// copy quadrant 2
			toR = halfRow + i;
			toC = halfCol;
			fromR = 0 + i;
			fromC = 0;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(Float));
			// copy quadrant 3
			toR = 0 + i;
			toC = 0;
			fromR = halfRow + i;
			fromC = halfCol;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(Float));
			// copy quadrant 4
			toR = 0 + i;
			toC = halfCol;
			fromR = halfRow + i;
			fromC = 0;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(Float));
		}

		} else { // odd

			int ii;
		    //  quadrant 1 has the larger square to move
			halfRow = m.nRow/2;
			halfCol = m.nCol/2;
			Float *from, *to;
			Float *pp, *qq;

				// memcpy(to, from, size)
				// copy quadrant 4
				from = m.data;
				to = result.data;
				q = to;
				printf("adding %d to p\n", m.nCol * (m.nCol/2) + (m.nCol/2));
				p = from + m.nCol * (m.nCol/2) + (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2 + 1; ii++) {
					// memcpy(to, from, size)
					memcpy(qq, pp, sizeof(Float) * ((m.nCol/2) + 1));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 3
				q += (m.nCol/2) + 1;
				p -= (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2 + 1; ii++) {
					// memcpy(to, from, size)
					memcpy(qq, pp, sizeof(Float) * (m.nCol/2));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 2
				q += ((m.nCol/2) * m.nCol) + 1;
				p = from + (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2; ii++) {
					// memcpy(to, from, size)
					memcpy(qq, pp, sizeof(Float) * ((m.nCol/2) + 1));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 1
				q += (m.nCol/2) + 1;
				p = from;
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2; ii++) {
					// memcpy(to, from, size)
					memcpy(qq, pp, sizeof(Float) * (m.nCol/2));
					qq += m.nCol;
					pp += m.nCol;
				}

		}
	}

	return result;
}


/** ifftwShift (use for odd size matrix) shifts data values after a reverse transform.
 * Because of the way FFTW
 * performs the Fourier transform, the data sometimes needs to be shifted.
 * @param[in] m A MATRIX containing data to shift.
 * @return A MATRIX containing shifted data.
 * @ingroup Matrix
 */
MATRIX ifftwShift(MATRIX m) {

	MATRIX result;

	if (m.mtype != REAL) {
		reportError(FftwShiftNoComplexData, "", FATAL);
	} else {
		Float *p, *q;

		result = newMatrix(m.nRow, m.nCol);
		p = m.data;
		q = result.data;

		if ((m.nRow % 2) == 0) { // even
			reportError("Don't use ifftwShift for even matrix", "ifftwShift", FATAL);
		} else { // odd
			int ii;
		    //  quadrant 1 has the larger square to move
			Float *from, *to;
			Float *pp, *qq;
				// memcpy(to, from, size)
				// copy quadrant 1
				from = m.data;
				to = result.data;
				q = from;
				printf("adding %d to p\n", m.nCol * (m.nCol/2) + (m.nCol/2));
				p = to + m.nCol * (m.nCol/2) + (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2 + 1; ii++) {
					// memcpy(to, from, size)
					memcpy(pp, qq, sizeof(Float) * ((m.nCol/2) + 1));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 2
				q += (m.nCol/2) + 1;
				p -= (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2 + 1; ii++) {
					// memcpy(to, from, size)
					memcpy(pp, qq, sizeof(Float) * (m.nCol/2));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 3
				q += ((m.nCol/2) * m.nCol) + 1;
				p = to + (m.nCol/2);
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2; ii++) {
					// memcpy(to, from, size)
					memcpy(pp, qq, sizeof(Float) * ((m.nCol/2) + 1));
					qq += m.nCol;
					pp += m.nCol;
				}
				// copy quadrant 4
				q += (m.nCol/2) + 1;
				p = from;
				pp = p;
				qq = q;
				for (ii=0; ii<m.nCol/2; ii++) {
					// memcpy(to, from, size)
					memcpy(pp, qq, sizeof(Float) * (m.nCol/2));
					qq += m.nCol;
					pp += m.nCol;
				}
		}
	}

	return result;
}

