/*
 * linearEqSolve.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <complex.h>

#include "InvTypes.h"

#ifdef MKL
#define MKL_Complex16 FloatComplex
#define MKL_INT int
#include "mkl_lapacke.h"
#include "cblas.h"
#define CblasRowMajor LAPACK_ROW_MAJOR

#else
#include <clapack.h>
#endif
#include "utilities.h"
#include "matrix.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "ProblemDimension.h"

#ifdef SINGLE
#define clapack_zgesv clapack_cgesv
#define clapack_zgetrs clapack_cgetrs
#define clapack_zpotrs clapack_cpotrs
#define clapack_dgesv clapack_sgesv
#define clapack_dgetrs clapack_sgetrs
#define clapack_dpotrs clapack_spotrs

#define LAPACKE_dgesv LAPACKE_sgesv
#define LAPACKE_zgesv LAPACKE_cgesv

#define zgelss_ cgelss_
#endif

#ifndef MKL
// external LAPACK function declarations
extern void zgelss_( const int *M,
		const int *N,
		const int *NRHS,
		FloatComplex *A,
		const int *LDA,
		FloatComplex *B,
		const int *LDB,
		Float *S,
		Float *RCOND,
		const int *RANK,
		FloatComplex *WORK,
		const int *LWORK,
		Float *RWORK,
		int *INFO
	);
#endif

/* Keep the LUDecomp and pivot values of the last call to linearEqSolve
 * because it can be reused when calculating the
 * Averaging Kernel.
 *
 */
static MATRIX LUDecomp = {NULL, -1, -1, REAL, ROW_MAJOR};	///< Keep the LUDecomp values of the last call to linearEqSolve because it can be reused when calculating the Averaging Kernel.
static int *IPiv = NULL;		///< Keep the pivot values of the last call to linearEqSolve because it can be reused when calculating the Averaging Kernel.



/**
 * Solve Ax=B for x.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
linearEqSolve(MATRIX A, MATRIX B)
{
    MATRIX result;

    int N; 	// The number of linear equations
    int nrhs;
    int lda;
    int* ipiv;
    int ldb;
    int info;

    // leading dimension = # of columns

    N = B.nRow;
    nrhs = B.nCol;
    lda = A.nCol;

    ipiv = (int *) td_malloc(N * sizeof(int));
	if (ipiv == NULL)
		reportError(OutOfMemory, "linearEqSolve.ipiv", FATAL);
#ifdef MKL
    ldb = B.nCol; // B.nRow; <<<<<<<<<<<<< VERY IMPORTANT!!!!!!!!!
#else
    ldb = B.nRow;
#endif

    if (LUDecomp.data != NULL) {
    	freeMatrix(LUDecomp);
    }

    if (IPiv != NULL) {
    	td_free(IPiv, LUDecomp.nRow * sizeof(int));
    }

	if ((A.mtype == REAL) && (B.mtype == REAL)) {
		result = matrix_copy(B);
		LUDecomp = matrix_copy(A);
#ifdef MKL
		info = LAPACKE_dgesv(LAPACK_ROW_MAJOR, N, nrhs, LUDecomp.data, lda, ipiv, result.data, ldb);
#else
		info = clapack_dgesv(CblasRowMajor, N, nrhs, LUDecomp.data, lda, ipiv, result.data, ldb);
#endif
	} else {
		// complex data
		// switch LUDecomp and VCopy to row major order
		LUDecomp = matrix_copy(A);

#ifdef MKL
		result = matrix_copy(B);
		info = LAPACKE_zgesv(LAPACK_ROW_MAJOR, N, nrhs, LUDecomp.data, lda, ipiv, result.data, ldb);
#else
		MATRIX resultColM = toColumnMajor2(B);
		info = clapack_zgesv(CblasRowMajor, N, nrhs, LUDecomp.data, lda, ipiv, resultColM.data, ldb);
		result = toRowMajor2(resultColM);
		freeMatrix(resultColM);
#endif
		// when reusing the LUDecomp do NOT transpose the LUDecomp matrix
	}

    // save intermediate values
    IPiv = ipiv;

    if(info == 0) { // succeed
    	;
    }
    else {
    	char msg[MAX_ERROR_MSG_LENGTH];
    	if (info < 0) {
    		// if INFO = -i, the i-th argument had an illegal value
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
    	} else {
    		//from description of dgesv.f
    		//  if INFO = i, U(i,i) is exactly zero.  The factorization
    		//                has been completed, but the factor U is exactly
    		//                singular, so the solution could not be computed.
    		char msg1[] = "U(i,i) is exactly zero.  The factorization has been completed, but the factor U is exactly singular, so the solution could not be computed.";
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
    	}
    	reportError(DGESVFailed, msg, WARNING);
    	// recover by returning a NaN vector or matrix
    	setToNaNMatrix(result);
    }

	return result;
}

/** Using this function requires first calling linearEqSolve to
 * calculate the LUDecomp of Matrix A.  Solves set of linear equations
 * with multiple right-hand sides, i.e. the number of columns of matrix B.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
LinearEqSolveNRHS(MATRIX A, MATRIX B)
{
    MATRIX result;

    int N; 	// The number of linear equations
    int nrhs;
    int lda;
    int ldb;
    int info;
	MATRIX resultColM;

    // use LUDecomp for matrix A
    if (LUDecomp.data == NULL) {
    	 reportError(LinearEqSolveFirst, "linearEqSolve.LinearEqSolveNRHS", ERROR);
    }

    N = A.nRow;
    nrhs = B.nCol;
    lda = A.nRow;
    ldb = B.nRow;



    //*********** crucial info ********
    // LU comes back in column major order, but it doesn't matter; leave it alone!
    // send the right-hand side in column major order
    // ldb is the number of rows in the right hand side
    // N is the number of rows in A (which becomes LUDecomp)
    // ipiv[N]
    // lda is the number of rows or columns in A -- A must be square
    // ******************

	/*
	 * What's the deal with the RHS in the row-major factorization/solves?
		Most users are confused by the row major factorization and related solves.
		The right-hand side vectors are probably the biggest source of confusion.
		The RHS array does not represent a matrix in the mathematical sense, it is
		instead a pasting together of the various RHS into one array for calling convenience.
		As such, RHS vectors are always stored contiguously, regardless of the row/col major that is chosen.
		This means that ldb/ldx is always independent of NRHS, and dependant on N,
		regardless of the row/col major setting.
	 */
	resultColM = toColumnMajor2(B);

		if (getProblemDimension() == 1) {
			// take advantage of intermediate value LUDecomp
#ifdef MKL
			info = LAPACKE_dgetrs(CblasRowMajor, CblasNoTrans, N, nrhs, LUDecomp.data, lda, IPiv, resultColM.data, ldb);
#else
			info = clapack_dgetrs(CblasRowMajor, CblasNoTrans, N, nrhs, LUDecomp.data, lda, IPiv, resultColM.data, ldb);
#endif
	
		} else {
			// 3d problem; complex data
			// remember, no need to transpose LUDecomp matrix, it is already in column major order
#ifdef MKL
			info = LAPACKE_zgetrs(CblasRowMajor, CblasNoTrans, N, nrhs, LUDecomp.data, lda, IPiv, resultColM.data, ldb);
#else
			info = clapack_zgetrs(CblasRowMajor, CblasNoTrans, N, nrhs, LUDecomp.data, lda, IPiv, resultColM.data, ldb);
#endif
			
		}
	result = toRowMajor2(resultColM);
	// free some memory
	freeMatrix(resultColM);

    if(info == 0) { /* succeed */
    	;
    }
    else {
    	char msg[MAX_ERROR_MSG_LENGTH];
    	if (info < 0) {
    		// if INFO = -i, the i-th argument had an illegal value
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
    	} else {
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Unknown error %d", info);
    	}
    	reportError(DGETRSFailed, msg, WARNING);
    	// this seems a bit strange to free 'result', then set its values and return it, but
    	// really, we are only freeing the memory pointed to by result.data, nothing else.
    	//freeMatrix(result);
    	result.data = NULL;
    	result.nRow = -1;
    	result.nCol = -1;
    }
	return result;
}

/**
 * Free memory allocated for LUDecomp and IPiv intermediate values.
 * @ingroup Matrix
 */
void linearEqSolveCleanup() {
    if (LUDecomp.data != NULL) {
    	freeMatrix(LUDecomp);
    }

    if (IPiv != NULL) {
    	td_free(IPiv, LUDecomp.nRow * sizeof(int));
    }
}





/**
 * Solve Ax=B for x.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
linearEqSolveCholesky(MATRIX A, MATRIX B)
{
	MATRIX result;

	int N; 	// The number of linear equations
	int nrhs;
	int lda;
	int* ipiv = NULL;
	int ldb;
	int info;

	// leading dimension = # of rows

	N = B.nRow;
	nrhs = B.nCol;
	lda = A.nRow;

	ldb = B.nRow;


	// We want to use DPOTRS( UPLO, N, NRHS, A, LDA, B, LDB, INFO )
	if ((A.mtype == REAL) && (B.mtype == REAL)) {
		reportError("not implemented for real numbers", "linearEqSolveCholesky", FATAL);
		result = matrix_copy(B);
		LUDecomp = matrix_copy(A);
#ifdef MKL
		info = LAPACKE_dgesv(CblasRowMajor, N, nrhs, LUDecomp.data, lda, ipiv, result.data, ldb);
#else
		info = clapack_dgesv(CblasRowMajor, N, nrhs, LUDecomp.data, lda, ipiv, result.data, ldb);
#endif
	} else {
		// complex data
		MATRIX L;
		L = Cholesky(A);
		result = matrix_copy(B);
#ifdef MKL
		info = LAPACKE_zpotrs(CblasRowMajor, CblasUpper, L.nRow, nrhs, L.data, L.nCol, result.data, ldb);
#else
		info = clapack_zpotrs(CblasRowMajor, CblasUpper, L.nRow, nrhs, L.data, L.nCol, result.data, ldb);
#endif
	}

	if(info == 0) { /* succeed */
		;
	}
	else {
		char msg[MAX_ERROR_MSG_LENGTH];
		if (info < 0) {
			// if INFO = -i, the i-th argument had an illegal value
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
		} else {
			char msg1[] = " Returned with error value of ";
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "%s  %d", msg1, info);
		}
		reportError("lapack_zpotrs failed", msg, WARNING);
		// recover by returning a NaN vector or matrix
		setToNaNMatrix(result);
	}
	return result;
}



/**
 * Solve Ax=B for x, using SVD.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
linearEqSolveSvd(MATRIX A, MATRIX B)
{
	MATRIX result;

	int n; 	// The number of linear equations
	int m;  // The number of rows in matrix A
	int nrhs;
	int lda;
	int info;

	// leading dimension = # of rows
    m = A.nRow;
	n = B.nRow;
	nrhs = B.nCol;
	lda = A.nRow;

	if ((A.mtype == REAL) && (B.mtype == REAL)) {
		reportError("Not Implemented for REAL matrix", "linearEqSolveSvd",
				FATAL);
	} else {
		// complex data
		// switch A and B to column-major order

		// array 's' is still Float precision
		Float *s;
		s = (Float *) td_malloc(m * sizeof(Float));

		int lwork;
		lwork = 3 * min(m,n) + max(max(2*min(m,n), max(m,n)), nrhs);
		FloatComplex *worki;
		Float *rwork;
		worki = (FloatComplex *) td_malloc(lwork * sizeof(FloatComplex));
		rwork = (Float *) td_malloc(5 * max(m,n) * sizeof(Float));

		/*
		 * RCOND is used to determine the effective rank of A.
		 * Singular values S(i) <= RCOND*S(1) are treated as zero.
		 * If RCOND < 0, machine precision is used instead.
		 */
		Float rcond = -1.0;
		int rank;

		MATRIX ACopyCM;
		ACopyCM = toColumnMajor2(A);  // produces a copy of the matrix
		// B should just be a vector, so no need to move to column-major order
		// Ewh, dangerous ... B could have multiple right-hand sides!!!
		// Ok, I can check to see if a vector or nrhs - or-
		// I could always convert from row to column

		MATRIX resultColM = toColumnMajor2(B);

		int ldb;
		ldb = max(m,n);

		zgelss_(&m, &n, &nrhs, ACopyCM.data, &lda, resultColM.data, &ldb, s, &rcond,
				&rank, worki, &lwork, rwork, &info);

		//info = LAPACKE_zgesvd( LAPACK_ROW_MAJOR, 'A', 'A', m, n, a, lda, s,
		// u, ldu, vt, ldvt, superb );

		result = toRowMajor2(resultColM);
		freeMatrix(resultColM);
		freeMatrix(ACopyCM);
		td_free(worki, lwork * sizeof(FloatComplex));
		td_free(rwork, 5 * max(m,n) * sizeof(Float));
		td_free(s, m * sizeof(Float));

		// Check for convergence
		if (info > 0) {
			reportError("The algorithm computing SVD failed to converge", "linearEqSolveSvd", FATAL);
		}

	}

	if (info == 0) { // succeed
		;
	} else {
		char msg[MAX_ERROR_MSG_LENGTH];
		if (info < 0) {
			// if INFO = -i, the i-th argument had an illegal value
			snprintf(msg, MAX_ERROR_MSG_LENGTH,
					"Argument %d had an illegal value", -info);
		} else {
			//from description of dgesv.f
			//  if INFO = i, U(i,i) is exactly zero.  The factorization
			//                has been completed, but the factor U is exactly
			//                singular, so the solution could not be computed.
			char msg1[] =
					"U(i,i) is exactly zero.  The factorization has been completed, but the factor U is exactly singular, so the solution could not be computed.";
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
		}
		reportError("ZGELSSFailed", msg, WARNING);
		// recover by returning a NaN vector or matrix
		setToNaNMatrix(result);
	}
	return result;
}
