/*
 * matrix.h
 *
 *  Created on: Apr 22, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#ifndef MATRIX_H_
#define MATRIX_H_

#include "InvTypes.h"
#include "objectfits.h"
#include "utilities.h"
#include <complex.h>


/**
 * Type of data kept in matrix: either Real (Float) or complex (FloatComplex)
 * @ingroup Matrix
 */
typedef enum {REAL, COMPLEX, BOOL} MatrixType;

/**
 * MatrixOrder designates the layout of data in a matrix.
 * @ingroup Matrix
 */
typedef enum  {ROW_MAJOR, COL_MAJOR} MatrixOrder;

/**
 * MATRIX is a main datatype and contains either Float or FloatComplex
 * data.
 * @ingroup Matrix
 */
typedef struct {
  void *data;			///< array of data values stored in row-major order
  int nRow;				///< number of rows in matrix
  int nCol;				///< number of columns in matrix
  MatrixType mtype;		///< type of data, either real or complex.
  MatrixOrder morder;	///< layout of data, either row-major or column-major.
} MATRIX;

/**
 * VECTOR is a main datatype and contains data of type Float only.
 * @ingroup Matrix
 */
typedef struct {
  Float *data;	///< array of data values
  int n;  		///< number of elements in vector
} VECTOR;



MATRIX newMatrix(int r, int c);
MATRIX newMatrixI(int r, int c, void *Idata, MatrixType mType);
MATRIX newComplexMatrix(int r, int c);
MATRIX newMatrixG(int r, int c, MatrixType mType);
MATRIX newMatrixGI(int r, int c, void *Idata, MatrixType mType);

void freeMatrix(MATRIX m);

VECTOR newVector(int n);
void freeVector(VECTOR v);

MATRIX matrix_copy(MATRIX m);

unsigned int setRandomSeedToTime();
unsigned int setRandomSeed(unsigned int seed);

MATRIX newRandomMatrix(int r, int c);

VECTOR multiplyMV(MATRIX m, VECTOR v);

MATRIX multiplyMM2(char TransA, char TransB, MATRIX A, MATRIX B);
MATRIX multiplyMM2NoForceC(char TransA, char TransB, MATRIX A, MATRIX B);
MATRIX multiplyMMFull(char TransA, char TransB, Float alpha, Float beta, MATRIX A, MATRIX B);

/**
 * Multiply a matrix, m, by a diagonal matrix represented by its diagonal elements in
 * vector, d
 */
MATRIX multiplyMD(MATRIX m, VECTOR d);

// multiply a diagonal matrix, whose diagonal elements are contained in the
// vector d, by a matrix m
MATRIX multiplyDM(VECTOR d, MATRIX m);

VECTOR multiplyDV(VECTOR d, VECTOR v);

MATRIX multiplyScalar(MATRIX m, Float scalar);
MATRIX multiplyScalarZ(MATRIX m, FloatComplex scalar);
MATRIX starMultiply(MATRIX P, MATRIX Q);

MATRIX multiplyMMplusC(MATRIX A, MATRIX B, Float beta, MATRIX C);
MATRIX multiplyMMplusCZ(MATRIX A, MATRIX B, Float beta, MATRIX C);
MATRIX multiplyABplusC(char TransA, char TransB, MATRIX A, MATRIX B, Float beta, MATRIX C);

MATRIX toMatrix(fits2d f);
VECTOR toVector(fits2d f);

fits2d toFitsObject(VECTOR v);
fits2d MatrixToFitsObject(MATRIX m);
fits3d MatrixToFits3dObject(MATRIX m);

Float* toRowMajor(Float* A, int R, int C);
Float* toColumnMajor(Float* A, int R, int C);

MATRIX toRowMajor2(MATRIX m);
MATRIX toColumnMajor2(MATRIX m);

MATRIX expandDiagonalMatrix(VECTOR v);

MATRIX transpose(MATRIX m);

MATRIX conjugate(MATRIX m);

MATRIX inverseD(MATRIX m);

MATRIX inverse(MATRIX m);
MATRIX inverseOld(MATRIX m);

MATRIX Cholesky(MATRIX m);

bool isIdentity(MATRIX m);

MATRIX ConvertToReal(MATRIX m);

MATRIX ConvertToComplex(MATRIX m);

VECTOR MatrixToVector(MATRIX m);

MATRIX VectorToMatrix_Column(VECTOR v);

MATRIX reduceMatrix(MATRIX from, VECTOR depths, int areRows) ;

void insertVector(Float* to, VECTOR from);

void APlusLambdaB(MATRIX A, Float lambda, MATRIX B, MATRIX C);

FloatComplex Norm2(MATRIX m);

MATRIX BItimesV(MATRIX B, MATRIX V);

int matrix_equal(MATRIX a, MATRIX b);

void setToNaN(VECTOR v);
void setToNaNMatrix(MATRIX m);

MATRIX pad(MATRIX a, int newRow, int newCol, Float value);
MATRIX embed(MATRIX a, int factor);
void splitMatrix(MATRIX m, MATRIX *real, MATRIX *imag);
void mergeMatrix(MATRIX *m, MATRIX real, MATRIX imag);

MATRIX zeroScatterer(MATRIX K, int scatterer, int Nz, int NBeta);
MATRIX scaleScatterer(MATRIX K, Float factor, int scatterer, int Nz, int NBeta);
MATRIX scaleNonScatterer(MATRIX K, Float factor, int scatterer, int Nz, int NBeta);

void addToDiagonal(MATRIX *m, Float epsilon);

Float calcCondNum(MATRIX ak);

/** read a vector from a text file (ASCII file)
 *
 */
VECTOR readText(char *filename);

void print_matrix( char* desc, int m, int n, Float* a, int lda, MatrixOrder order);
void print_matrix2(char* desc, MATRIX m, MatrixOrder order);
void print_matrixNoI(char* desc, MATRIX m, MatrixOrder order);
void print_matrixML(char* desc, MATRIX m, MatrixOrder order);
void print_matrix2toFile(char* desc, MATRIX m, MatrixOrder order);


int testTranspose();

#endif /* MATRIX_H_ */
