/*
 * writefits.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <complex.h>
#include <fitsio.h>
#include "objectfits.h"
#include "utilities.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "FITSParameters.h"

/**
 * Write complex data to a file using FITS format
 * @param[in] filename Name of output file
 * @param[in] data Array of complex data in a fits3d object
 * @return FITS file status.
 * @ingroup FITS
 */
int writefits(char *filename, fits3d data) {
    fitsfile *fptr;       /* pointer to the FITS file; defined in fitsio.h */
    int status;
    long  fpixel[3] = {1,1,1}, naxis = 3;
    long naxes[3];
    char overwriteFlag = '!';
    char *tempFilename;

    tempFilename = (char *) td_malloc((strlen(filename) + 2) * sizeof(char));
    strcpy(&tempFilename[1], filename);
    tempFilename[0] = overwriteFlag;

    status = 0;         /* initialize status before calling fitsio routines */
    fits_create_file(&fptr, tempFilename, &status);   /* create new file */
    td_free(tempFilename, (strlen(filename) + 2) * sizeof(char));

    naxes[0] = data.yDim;
    naxes[1] = data.xDim;
    naxes[2] = 2; // for complex data only real and imaginary

    /* Create the primary array image (64-bit Float precision floating point */
#ifdef SINGLE
	fits_create_img(fptr, FLOAT_IMG, naxis, naxes, &status);
#else
    fits_create_img(fptr, DOUBLE_IMG, naxis, naxes, &status);
#endif

      /* loop over all the rows in the image, top to bottom */
    for (fpixel[1] = naxes[1]; fpixel[1] >= 1; fpixel[1]--) {
#ifdef SINGLE
      fits_write_pix(fptr, TFLOAT, fpixel, naxes[0],
                     &(data.Real[fpixel[1]-1][0]), &status);
#else
      fits_write_pix(fptr, TDOUBLE, fpixel, naxes[0],
		     &(data.Real[fpixel[1]-1][0]), &status);
#endif
    }
    // move to next frame; set fpixel to next value in z dimension
    fpixel[2] = 2;
    for (fpixel[1] = naxes[1]; fpixel[1] >= 1; fpixel[1]--) {
#ifdef SINGLE
      fits_write_pix(fptr, TFLOAT, fpixel, naxes[0],
                     &(data.Imaginary[fpixel[1]-1][0]), &status);
#else
      fits_write_pix(fptr, TDOUBLE, fpixel, naxes[0],
		     &(data.Imaginary[fpixel[1]-1][0]), &status);
#endif
    }

    fits_close_file(fptr, &status);            /* close the file */

    fits_report_error(stderr, status);  /* print out any error messages */
    return( status );
}

/**
 * Write real data to a file using FITS format
 * @param[in] filename Name of output file
 * @param[in] data Array of real data in a fits2d object
 * @param[in] lambda Value of lambda parameter used during inversion calculation for this data.
 * @param[in] useLambda True, if writing lambda parameter in FITS header.
 * @return FITS file status.
 * @ingroup FITS
 */
int writefits2dWithLambdaSwitch(char *filename, fits2d data, Float lambda, int useLambda) {
    fitsfile *fptr;       /* pointer to the FITS file; defined in fitsio.h */
    int status;
    long  fpixel[2] = {1,1}, naxis = 2;
    long naxes[2];
    char overwriteFlag = '!';
    char *tempFilename;
    int tempFilenameLen;
    char card[MAX_CARD_LENGTH_FOR_FITS];

    tempFilenameLen = (strlen(filename) + 2) * sizeof(char);
    tempFilename = (char *) td_malloc(tempFilenameLen);
    strcpy(&tempFilename[1], filename);
    tempFilename[0] = overwriteFlag;

    status = 0; /* initialize status before calling fitsio routines */
	if (!fits_create_file(&fptr, tempFilename, &status)) { /* create new  or overwrite existing one file */

		naxes[0] = data.yDim;
		naxes[1] = data.xDim;

		/* Create the primary array image (64-bit Float precision floating point */
#ifdef SINGLE
		fits_create_img(fptr, FLOAT_IMG, naxis, naxes, &status);
#else
		fits_create_img(fptr, DOUBLE_IMG, naxis, naxes, &status);
#endif

		if (useLambda) {
			char lambdastr[] = "LAMBDA";
			char comment[] = "lambda value";

			// for numeric scientific notation
			snprintf(card, MAX_CARD_LENGTH_FOR_FITS, "%-7s = %20E / %s", lambdastr, lambda, comment);
			fits_write_record(fptr, card, &status);
			if (status != 0) {
				fits_report_error(stderr, status); // print out any error messages
				reportError(WriteFitsHeaderError, "writefits2dWithLambdaSwitch", FATAL);
				fits_close_file(fptr, &status); // close the file
			}
		}

		Float *thedata;
		thedata = data.data;
		/* Write the array of Floats to the image */
		fpixel[0] = 1; // this is the starting point for output

		int idx;
		/* loop over all the rows in the image, top to bottom */
		for (fpixel[1] = naxes[1]; fpixel[1] >= 1; fpixel[1]--) {
			idx = (fpixel[1] - 1) * naxes[0];
#ifdef SINGLE
            fits_write_pix(fptr, TFLOAT, fpixel, naxes[0], &thedata[idx],
                    &status);
#else
			fits_write_pix(fptr, TDOUBLE, fpixel, naxes[0], &thedata[idx],
					&status);
#endif
		}

		fits_close_file(fptr, &status); /* close the file */
	}
	td_free(tempFilename, tempFilenameLen);
	fits_report_error(stderr, status); /* print out any error messages */
	if (status != 0) reportError(FileOpenError, filename, FATAL);
	return (status);
}

/**
 * Write real data to a file using FITS format
 * @param[in] filename Name of output file
 * @param[in] data Array of real data in a fits2d object
 * @return FITS file status.
 * @ingroup FITS
 */
int writefits2d(char *filename, fits2d data) {
	return writefits2dWithLambdaSwitch(filename, data, 0.0, FALSE);
}

/**
 * Write real data to a file using FITS format
 * @param[in] filename Name of output file
 * @param[in] data Array of real data in a fits2d object
 * @param[in] lambda Value of lambda parameter used during inversion calculation for this data.
 * @return FITS file status.
 * @ingroup FITS
 */
int writefits2dWithLambda(char *filename, fits2d data, Float lambda) {
	return writefits2dWithLambdaSwitch(filename, data, lambda, TRUE);
}

