/*
 * lambda.c
 *
 *  Created on: Apr 30, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdlib.h>
#include <math.h>
#include "lambda.h"
#include "utilities.h"
#include "errorHandler.h"
#include "CoreErrorMessages.h"

#define LAMBDA_STEP(L) ((L.max - L.min)/((Float) (L.n - 1)))

/**
 * @defgroup LambdaSet LambdaSet - LambdaSet is a range of values.
 * Used for parameters lambda, mu, sigma, and nu.
 * @ingroup SupportingModules
 */

/**
 * Determines the value of lambda at nth step in range.
 * @param[in] lambdaSet LambdaSet defines a range of values for a parameter.
 * @param[in] n Fetch the nth value in the parameter range.
 * @return The nth value in the parameter range.
 */
Float lambdaAt(LambdaSet lambdaSet, int n) {
	Float value;

	if (lambdaSet.min > 0) {
		Float N;
		N = lambdaSet.base + n;
		value = lambdaSet.mantissa * pow(10.0, N);
	} else {
		// if the base exponent is positive, just use n
		if (lambdaSet.base >= 0)
			value = pow(10.0, (Float) n-1);
		else {
			// otherwise, we'll need to ADD the number of steps to the
			// negative base;  this is for values between 0 and 1 with
			// negative log values.
			Float N;
			N = lambdaSet.base + n;
			value = pow(10.0, N);
		}

	}
	return value;
}


/**
 * newLambdaSet initializes a range of values, which either linear or
 * logarithmic steps.
 * @param[in] min Minimum value for range.
 * @param[in] max Maximum value for range.
 * @param[in] linearSteps Number of linear steps including min and max values.
 * @param[in] logStep 1 if using log steps, 0 if using linear steps.
 * @returns Pointer to an initialized LambdaSet.
 */
LambdaSet *newLambdaSet(Float min, Float max, int linearSteps, int logStep) {

	LambdaSet *ls;

	// NOTE: n goes from 1 ... nsteps, inclusive
	ls = (LambdaSet *) td_malloc(sizeof(LambdaSet));
	ls->min = min;
	ls->max = max;
	if (logStep) {
	    ls->log = 1;

		if (ls->min > 0) {

			ls->base = (int) log10(ls->min);
			ls->mantissa = ls->min/pow(10,ls->base);
			//Float mantissaMax;
			Float baseMax;
			baseMax = (int) log10(ls->max);
			//mantissaMax = ls->max/pow(10,baseMax);
			//ls->n = (log10(ls->max/mantissaMax) - log10(ls->min/ls->mantissa)) + 1;
			int emax, emin;
			emax = log10(ls->max);
			emin = log10(ls->min);
			ls->n = emax - emin + 1;
		} else {
			ls->n = (int) log10(ls->max) + 2;
			ls->base = 0;
			ls->mantissa = 1;
			// if range is 0 ... 1 issue informative message and set the lower bound
			if ((ls->max < 10)) {
				reportError(LambdaSetSettingLowerBound, "newLambdaSet", INFO);
				ls->base = (int) log10(ls->max) - 4;
				ls->n = 5;
			}
		}
		// if we won't end up at the max, then add one more step
		Float nv = lambdaAt(*ls, ls->n - 1);
		if (nv < ls->max)
			ls->n += 1;

	} else {
		ls->n = linearSteps;
		ls->log = 0;
	}
	return ls;
}

void freeLambdaSet(LambdaSet *ls) {
	td_free(ls, sizeof(LambdaSet));
}

/*
 * nextLambdaValue returns the next value in a lambda set.
 * @param[in] lambdaSet A lambda set with initial values.
 * @param[in] n which step value.  n should be >= 1. The
 * zero step is the minimum value.
 * @returns The the nth value in the range specified by
 * the lambdaSet.
 */
Float nextLambdaValue(LambdaSet lambdaSet, int n) {
	Float value;

	if (lambdaSet.log) {

		value = lambdaAt(lambdaSet, n);

		if (value > lambdaSet.max)
			value = lambdaSet.max;
	} else {
		value = lambdaSet.min + LAMBDA_STEP(lambdaSet) * n;
	}
	return value;
}

