/*
 * utilities.c
 *
 *  Created on: May 24, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdbool.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <errno.h>
#include <err.h>
#include <unistd.h>
#include <time.h>
#include "utilities.h"
#include "errorHandler.h"
#include "CoreErrorMessages.h"

int MAX_DIRECTORY_VERSION = 9999;

bool endsWith(char *string, char *ending) {
	int stringLen, endingLen;

	stringLen = strlen(string);
	endingLen = strlen(ending);

	if (stringLen < endingLen)
		return false;

	return !strncmp(&string[stringLen - endingLen], ending, endingLen);
}

char *strip(char *aString) {
	size_t nonWhiteLen;
	char whiteSpace[] = " \t\n\0";
	// Finds the first sequence of characters in the string str1 that does not contain any character specified in str2.
	// Returns the length of this first sequence of characters found that do not match with str2.
    //size_t strcspn(const char *str1, const char *str2);
	nonWhiteLen = strcspn(aString, whiteSpace);

	char *newString;
	newString = td_malloc(nonWhiteLen + 1);
	strncpy(newString, aString, nonWhiteLen + 1);
	return newString;
}

int longestName(char *filenames[], int nFiles) {
	int i, max, len;

	// find the longest filename for memory allocation
	max = 0;
	len = 0;
	for (i=0; i<nFiles; i++) {
		if ((len = strlen(filenames[i])) > max) {
			max = len;
		}
	}
	return len;
}

/**
 * Compare two string up to n characters. Comparison is done
 * with lowercase version of each string.
 * @param[in] str1 A string to compare.
 * @param[in] str2 Another string to compare.
 * @param[in] n maximum number of characters to compare.
 * @return TRUE if strings are equal; FALSE otherwise.
 */
bool lowerStrNCmp(const char *str1, const char *str2, int n) {
	int i=0;
	bool Equal = true;
	while ((str1[i] != '\0') && (str2[i] != '\0') && (i<n) && (Equal)) {
		if (tolower(str1[i]) != tolower(str2[i])) {
			Equal = false;
		}
		i += 1;
	}
	if (i != n) Equal = false;
	return Equal;
}

static int parserIdx = -1;
static char *parserLine = NULL;
static int parserLineLength = -1;

/**
 *  stringTokenizer and nextToken work together to parse a character
 *  string into tokens, separated by spaces.
 *  stringTokenizer intializes variables for breaking a string into tokens.
 * @param[in] line The string to break into words.
 */
void stringTokenizer(char *line) {
	parserIdx = -1;
	parserLine = line;
	parserLineLength = strlen(line);
}

/**
 * stringTokenizer and nextToken work together to parse a character
 * string into tokens, separated by spaces.
 * nextToken finds the next set of non-white space characters, which is
 * considered a token.
 * @return The next token in the string or NULL if there are no more tokens.
 */
char *nextToken() {
	char *p = NULL;

	if (parserIdx < parserLineLength) {
		parserIdx += 1;

		while ((parserLine[parserIdx] == ' ') && (parserIdx < parserLineLength)) {
			parserIdx += 1;
		}
		if (parserIdx < parserLineLength) {
			p = &parserLine[parserIdx];
			// move the parserIdx past this token
			while ((parserLine[parserIdx] != ' ') && (parserIdx < parserLineLength)) {
				parserIdx += 1;
			}
		}
	}
	return p;
}

/*
 * Returns a pointer to a new string; remember to free the space.
 * free strlen() + 1 for end-of-string character
 */
char *constructFileName(char *baseDir, const char *fileName) {

	// get the total string length
	int l1 = strlen(baseDir);
	int l2 = strlen(fileName);

	// allocate the space
	bool needSlash = false;
	char *fullName;
	int fullNameLength;
	fullNameLength = l1 + l2 + 1; // +1 for '\0' end of string character
	if (!endsWith(baseDir, "/")) {
		fullNameLength += 1;
		needSlash = true;
	}
	fullName = (char *) td_malloc(fullNameLength * sizeof(char));
	// copy the strings
	strcpy(fullName, baseDir);
	if (needSlash && (strlen(fullName) > 0))
		strcat(fullName, "/");
	strcat(fullName, fileName);
	// return the new string
    return fullName;
}

/*
 * append the filename to the basedir
 * assumes there is enough room in basedir for filename
 */
char *constructFileNameOld(char *basedir, char *filename) {
	if (!endsWith(basedir, "/")) {
		strcat(basedir, "/");
	}
	return strcat(basedir, filename);
}



static long MemoryInUse = 0;

char *td_malloc(size_t size) {

	char *mem = NULL;
	MemoryInUse += size;

	mem = malloc(size);
	if (mem == NULL) reportError(OutOfMemory, "td_malloc", FATAL);
	return mem;
}



void td_free(void *mem, size_t size) {
	MemoryInUse -= size;
	free(mem);
}


long getMemoryReport() {
	return MemoryInUse;
}

/*
 * on error, NULL is returned and errno is set.
 */
DIR  *exists(const char *name) {
	DIR *fd;
    fd = opendir(name);
    if (fd == NULL) {
        int errsv = errno;
    	reportErrorNo(errsv, name, "exists", WARNING);
    }
    return fd;
}

/**
 * Determine if the directory is empty.
 * @param[in] pd Pointer to a DIR structure
 * @returns true if the directory is empty; false otherwise.
 */
bool isEmpty(DIR* pd) {
	struct dirent *cur;
	if (pd == NULL)
		return false;
	while (cur = readdir(pd)) {
		if (cur->d_name[0] != '.') {
			puts(cur->d_name);
			return false;
		}
	}
	return true;
}

/**
 * Determine if we have write permission to the directory.
 * @param[in] pathname Character string containing full path to directory.
 * @returns true is if write access; false otherwise.
 */
bool isWritable(char *pathname) {
    // maybe use lstat or fstat?
    // I could just copy the input parameter file there and check for success?
    // Nope, I'll use access()
    /*
	*        int access(const char *pathname, int mode);

DESCRIPTION
access()  checks  whether the calling process can access the file pathname.  If pathname is a symbolic link,
it is dereferenced.

The mode specifies the accessibility check(s) to be performed, and is either the value F_OK, or a mask  con‐
sisting of the bitwise OR of one or more of R_OK, W_OK, and X_OK.  F_OK tests for the existence of the file.
R_OK, W_OK, and X_OK test whether the file exists and grants read, write, and execute  permissions,  respec‐
tively.

The  check  is  done  using the calling process's real UID and GID, rather than the effective IDs as is done
when actually attempting an operation (e.g., open(2)) on the file.  This allows set-user-ID programs to eas‐
ily determine the invoking user's authority.

If  the  calling  process is privileged (i.e., its real UID is zero), then an X_OK check is successful for a
regular file if execute permission is enabled for any of the file owner, group, or other.

RETURN VALUE
On success (all requested permissions granted), zero is returned.  On error (at least one bit in mode  asked
for  a  permission that is denied, or some other error occurred), -1 is returned, and errno is set appropri‐
ately.

*/
	if (access(pathname, W_OK) == -1) {
		int errsv;
		errsv = errno;

    	reportErrorNo(errsv, pathname, "isWritable", ERROR);
	}
	printf("%s is writable\n", pathname);
	return true;
}

/**
 * Determine if the directory or file is accessible for reading.
 * @param[in] pathname Character string containing the name of the directory or file.
 * @returns true if the resource is accessible for reading.
 */
bool isReadable(char *pathname) {
    // maybe use lstat or fstat?
    // I could just copy the input parameter file there and check for success?
    // Nope, I'll use access()
    /*
	*        int access(const char *pathname, int mode);

	DESCRIPTION
	access()  checks  whether the calling process can access the file pathname.  If pathname is a symbolic link,
	it is dereferenced.

	The mode specifies the accessibility check(s) to be performed, and is either the value F_OK, or a mask  con‐
	sisting of the bitwise OR of one or more of R_OK, W_OK, and X_OK.  F_OK tests for the existence of the file.
	R_OK, W_OK, and X_OK test whether the file exists and grants read, write, and execute  permissions,  respec‐
	tively.

	The  check  is  done  using the calling process's real UID and GID, rather than the effective IDs as is done
	when actually attempting an operation (e.g., open(2)) on the file.  This allows set-user-ID programs to eas‐
	ily determine the invoking user's authority.

	If  the  calling  process is privileged (i.e., its real UID is zero), then an X_OK check is successful for a
	regular file if execute permission is enabled for any of the file owner, group, or other.

	RETURN VALUE
	On success (all requested permissions granted), zero is returned.  On error (at least one bit in mode  asked
	for  a  permission that is denied, or some other error occurred), -1 is returned, and errno is set appropri‐
	ately.
    */
	if (access(pathname, R_OK) == -1) {

		int errsv;
		errsv = errno;
		// also check for .gz version of file
		// Could also look for these as well:  .zip, or .Z
		char *gzipVersion;
		gzipVersion = td_malloc(strlen(pathname) + 4);
		strncpy(gzipVersion, pathname, strlen(pathname) + 1);
		strcat(gzipVersion, ".gz");
		printf("trying |%s|\n", gzipVersion);
		if (access(gzipVersion, R_OK) == -1) {
			reportErrorNo(errsv, pathname, "isReadable", WARNING);
			td_free(gzipVersion, strlen(pathname) + 4);
			return false;
		} else {
			td_free(gzipVersion, strlen(pathname) + 4);
			return true;
		}
	}
	return true;
}


/**
 * Copy file from source to destination. This function make a system call
 * to perform the copy.
 * @param[in] source Character string containing path to source file.
 * @param[in] dest Character string containing path to destination.
 */
void copyFiles(const char *source, const char *dest) {

#define COPY_STR "cp %s %s"  // some sort of OS-specific syntax

	char copy_str[NAME_MAX];
	sprintf (copy_str, COPY_STR, source, dest);
	int ret;
	ret = system (copy_str);
	if (ret == -1)
		reportError(CopyFailed, "copyFiles", ERROR);
}

/**
 * Get the directory structure (pointer) associated with a string name
 * If the directory exists, return a pointer to it.
 * If the directory doesn't exist, create it and return a pointer to it.
 * If the directory cannot be created, report an error and return NULL.
 * @param[in] dirName Character string containing full path for directory.
 * @return A pointer to a system DIR structure.
 */
DIR *getDirectory(char *dirName) {
	DIR *dirPtr = NULL;
	int errsv;

    // use mkdir() to create the directory.  If it exists, mkdir will fail (return -1).
	// If it succeeds, then it has write permission,
    // because we are creating it with write permission.  So there!

	if (mkdir(dirName, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0) {
		// there was an error, see if the directory exists.
        errsv = errno;
		if (errsv != EEXIST) {
			// some other kind of error
	    	reportErrorNo(errsv, dirName, "getDirectory", FATAL);
		}
	}
	dirPtr = opendir(dirName);
    if (dirPtr == NULL) {
    	errsv = errno;
    	reportErrorNo(errsv, dirName, "getDirectory", FATAL);
    }
    return dirPtr;
}

/**
 * Get directory recursive; start a the left side and create or get each
 * part of the directory.
 * @param[in] dirNameOrig Character string containing the full directory path
 * @return A pointer to a system DIR structure.
 */
DIR *getDirectoryR(char *dirNameOrig) {
	DIR *dirPtr = NULL;

    // use mkdir() to create the directory.  If it exists, mkdir will fail (return -1).  If it succeeds, then it has write permission,
    // because we are creating it with write permission.  So there!

	// make sure there is an ending slash!!! and no trailing spaces!
	char *dirName = strip(dirNameOrig);

	if (!endsWith(dirName, "/")) {
		//ADD '/' to the end of the string!
		char *newString;
		size_t dirNameLength;
		dirNameLength = strlen(dirName);
		// remember, strlen points to the end-of-string character
		newString = td_malloc(dirNameLength + 2); // one for the end-of-string character; one for the slash
		strncpy(newString, dirName, dirNameLength);
		newString[dirNameLength] = '/';
		newString[dirNameLength+1] = '\0';
		td_free(dirName, dirNameLength+1);
		dirName = newString;
	}

	//printf("correctly formatted dir |%s|\n", dirName);
	char *ptrToDelim, *ptrToDelim2;

	ptrToDelim = strchr(dirName,  '/');

	// what about root directory? "/skld" or "//sskds"
	if (ptrToDelim == dirName) { // it is pointing to the first character, which is root
		// find the next occurrance of '/'
		ptrToDelim = strchr(ptrToDelim+1,  '/');
	}

	// returns a pointer to the first instance of c in s.
	// Returns a NULL pointer if c is not encountered in the string.
	//char *strchr( const char *s,  int c);

	if (ptrToDelim != NULL)
		*ptrToDelim = '\0';

	bool done = false;
	do {
		DIR *mydir;
		mydir = getDirectory(dirName);
		if (mydir != NULL) {

			// move to the next part of the directory path
			// move past it to find next one
			ptrToDelim2 = strchr(ptrToDelim+1,  '/');
			// restore '/'

			if (ptrToDelim2 != NULL) {
				*ptrToDelim = '/';
				*ptrToDelim2 = '\0';
				ptrToDelim = ptrToDelim2;
			} else {
				done = true;
			}
		} else {
			done = true;
		}

	} 	while (!done);

	dirPtr = opendir(dirName);
    if (dirPtr == NULL) {
    	int errsv;
    	errsv = errno;

    	reportErrorNo(errsv, dirName, "getDirectoryR", WARNING);
    }
    td_free(dirName, strlen(dirName)+2); // one for the end-of-string character; one for slash we removed in the loop

    return dirPtr;
}


/**
 * Get the parent part of the path.
 * @param[in] path String containing the full path.
 * @return String containing the parent part of the path.
 */
char *getParent(char *path) {
	char *ptr;
	int pathLen;

	char *parent;

        pathLen = strlen(path);
	// if path length <=0 then, error
	if (pathLen <= 0) {
		reportError("Path length <= 0", "getParent", ERROR);
		return NULL;
	}

	if ((pathLen == 1) && (path[0] == '/')) {
		reportError("Cannot find parent of root directory", "getParent", ERROR);
		return NULL;
	}

	ptr = &(path[pathLen-1]);

	// if path ends with /, remove it
        if (*ptr == '/')
		ptr -= 1;
	// locate /
	// from /-1, while idx > 0, and not equal to dir separator (don't remove the first
	// character, is needed),
	// decrease index
	while ((*ptr != '/') && (ptr > path)) {
		ptr -= 1;
	}

	if (*ptr != '/') { // then this is a relative path and we need to use
			  // the current working directory
		parent = td_malloc(2);
		parent[0] = '.';
		parent[1] = '\0';
	} else {
		if (ptr > path) { // then we stopped because of a directory separator that is
			// not at the beginning
			ptr -= 1;
		} // else { // we stopped because we are pointing right at the beginning of the path
		// the parent string goes from path to ptr, let's copy it
		//  return copy of parent string; remember to free space after finished with string
		parent = td_malloc(ptr-path + 2);
		strncpy(parent, path, (ptr - path) + 1);
	}


	return parent;
}

/**
 * Split path into parent and child
 * @param[in] path String containing path to split.
 * @param[in,out] parentPtr Pointer to string containing parent name.
 * @param[in,out] childPtr Pointer to string containing child name.
 */
void getParentNChild(char *path, char **parentPtr, char **childPtr) {
	char *ptr;
	int pathLen;

	char *parent;
	char *child;
	parent = *parentPtr;
	child = *childPtr;

    pathLen = strlen(path);
	// if path length <=0 then, error
	if (pathLen <= 0) {
		reportError("Path length <= 0", "getParentNChild", ERROR);
		parent = NULL;
		child = NULL;
	}

	if ((pathLen == 1) && (path[0] == '/')) {
		reportError("Cannot find parent of root directory", "getParentNChild", ERROR);
		parent = NULL;
		child = NULL;
	}

	ptr = &(path[pathLen-1]);

	// if path ends with /, remove it
    if (*ptr == '/') {
		ptr -= 1;
		pathLen -= 1;
    }

    // we should have something like this
    // 1) /ppp/ccc or (parent = "/ppp", child = "ccc")
    // 2) ppp/ccc or  (parent = "ppp", child = "ccc")
    // 3) ccc  (parent = ".", child = "ccc")
    // 4) /ccc  (parent = "/", child = "ccc");
	// locate right-most /
	// from /-1, while idx > 0, and not equal to dir separator (don't remove the first
	// character, it is needed),
	// decrease index
	while ((*ptr != '/') && (ptr > path)) {
		ptr -= 1;
	}

	if (*ptr != '/') { // then this is a relative path and we need to use
			  // the current working directory, case 3)
		parent = td_malloc(2);
		parent[0] = '.';
		parent[1] = '\0';
		child = td_malloc(pathLen + 1);
		strncpy(child, path, pathLen);
	} else {
		if (ptr > path) { // then we stopped because of a directory separator that is
			// not at the beginning
			ptr -= 1;
		} // else { // we stopped because we are pointing right at the beginning of the path
		// the parent string goes from path to ptr, let's copy it
		//  return copy of parent string; remember to free space after finished with string
		int parentLen;
		parentLen = ptr-path + 2;
		parent = td_malloc(parentLen);
		strncpy(parent, path, parentLen-1);
		parent[parentLen-1] = '\0';
		int childLen;
		childLen = path + pathLen - ptr - 1;
		child = td_malloc(childLen);
		strncpy(child, ptr+2, childLen-1);
		child[childLen-1] = '\0';
	}

    *parentPtr = parent;
    *childPtr = child;
}

/**
 * function to parse base output directory line
 *  find a directory that is empty, create one if needed by appending _v2, _v3, etc.
 * up to _v9999, which is the max.
 * @param[in] baseDir Character string containing full path to a directory.
 * @return a pointer to a character string that contains an empty output directory;
 *  NULL if an appropriate output directory cannot be found.
 */
char *getNextEmptyDirectory(char *baseDir) {
	// remove ending white space and directory separator if present
	char *originalDir;
    size_t nonWhiteLen;
	char whiteSpace[] = " \t\n\0";
	// Finds the first sequence of characters in the string str1 that does not contain any character specified in str2.
	// Returns the length of this first sequence of characters found that do not match with str2.
    //size_t strcspn(const char *str1, const char *str2);
	nonWhiteLen = strcspn(baseDir, whiteSpace);

	int workingDirLength = nonWhiteLen + 7;
	originalDir = td_malloc(workingDirLength); // allocate enough room for "_v9999\0"
	strncpy(originalDir, baseDir, nonWhiteLen);

	// originalDir = originalDir.rstrip(os.path.sep)
	//if (endsWith(originalDir, "/") || endsWith(originalDir, "\\")) {
		originalDir[nonWhiteLen] = '\0';
	//}
	char *outBaseDir;
	outBaseDir =  td_malloc(workingDirLength);
	strcpy(outBaseDir, originalDir);

	// try to create directory; if it fails, then it already exists,

        bool done = false;
        int vNumber = 1;
		// condition of allOK: exists && (empty || overwrite) && writable
		// another attempt at logic using do while
		while (!done) {

			DIR *fd;
			fd = getDirectoryR(outBaseDir);
			if (isEmpty(fd)) {
				if (isWritable(outBaseDir)) done = true;
			}
			else {
				vNumber += 1;
				if (vNumber < MAX_DIRECTORY_VERSION) {
					//move to next directory
					sprintf(outBaseDir, "%s_v%1d", originalDir, vNumber);
					printf(" ... trying output dir %s|\n", outBaseDir);
				} else {
					done = true;
				}
			}
		}

		if (vNumber >= MAX_DIRECTORY_VERSION) {
		    // free working string
		    td_free(outBaseDir, workingDirLength);
			reportError("exceeded max number of output directory versions", "getNextEmptyDirectory", FATAL);
		}

	    // make a copy of the output base dir with the exact length
	    char *newString;
	    nonWhiteLen = strlen(outBaseDir) + 1;
		newString = td_malloc(nonWhiteLen);
		strncpy(newString, outBaseDir, nonWhiteLen);
	    // free working string
	    td_free(outBaseDir, workingDirLength);
	    td_free(originalDir,workingDirLength);
		return newString;

}
