/*
 * BaseDirectory.c
 *
 *  Created on: Sep 8, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdio.h>
#include <string.h>
#include "Constants.h"
#include "utilities.h"
#include "averagingKernel.h"
#include "BaseDirectory.h"
#include "errorHandler.h"

//							muStep(5)_sigmaStep(5)_k.x(5)_k.y(5)q.fits(10)
#define MAX_OUTPUT_FILENAME 5+5+5+5+10  ///> maximum length of output filename including mu and sigma steps


/**
 * Set the base directory for all output files.
 * @param[in] directoryName Output base directory.
 * @return BaseDirectory
 * @ingroup BaseDirectory
 */
 BaseDirectory *newBaseDirectory(char *directoryName) {
	static BaseDirectory bd;

	int length = strlen(directoryName);
	bd.maxLength = MAX_OUTPUT_FILENAME + length;
	bd.Name = (char *) td_malloc((bd.maxLength) * sizeof(char));
	bd.dirLength = length;
	strcpy(bd.Name, directoryName);

	// add ending slash if not present
	if (!endsWith(bd.Name, Constants.DirectorySeparator)) {
		bd.Name[length] = Constants.DirectorySeparator[0];
		bd.Name[length+1] = '\0';
		bd.dirLength += 1;
	}

	// set default values
	bd.nFz0 = 1;
	bd.nV = MAX_ALPHA;

		bd.filenameBank = (char **) td_malloc(bd.nV*bd.nFz0 * sizeof(char *));
		int n;
		for (n = 0; n < bd.nV*bd.nFz0; n++) {
			bd.filenameBank[n] = (char *) td_malloc((bd.maxLength) * sizeof(char));
			strncpy(bd.filenameBank[n], bd.Name, bd.dirLength);
		}

	return &bd;
}

void destroyBaseDirectory(BaseDirectory *bd) {

	td_free(bd->Name, bd->maxLength);

	int n;
	for (n = 0; n < bd->nV * bd->nFz0; n++) {
		td_free(bd->filenameBank[n], bd->maxLength);
	}
	td_free(bd->filenameBank, bd->nV*bd->nFz0 * sizeof(char *));
}

/**
 * Construct bank of file names for Target output files.
 * There can be many output files ... Hm, I'll need to do this
 * for all the output files, they can have multiple target
 * depths (fz0) .... so, a change in structure here is needed.
 * But, at initialization, I'll know how many files are
 * needed for the file name bank.  I can send Nfz0 as a parameter
 * to the newBaseDirectory. YES!
 * @param[in] directoryName Output base directory.
 * @return BaseDirectory
 * @ingroup BaseDirectory
 */
BaseDirectory *newBaseDirectoryNrhs(char *directoryName) {
	static BaseDirectory bd;

	int length = strlen(directoryName);
	bd.maxLength = MAX_OUTPUT_FILENAME + length;
	bd.Name = (char *) td_malloc((bd.maxLength) * sizeof(char));
	bd.dirLength = length;
	strcpy(bd.Name, directoryName);

	// add ending slash if not present
	if (!endsWith(bd.Name, Constants.DirectorySeparator)) {
		bd.Name[length] = Constants.DirectorySeparator[0];
		bd.Name[length+1] = '\0';
		bd.dirLength += 1;
	}
	bd.filenameBank = NULL;
	// set default values
	bd.nFz0 = 1;
	bd.nV = MAX_ALPHA;
	return &bd;
 }

/**
 * To be used by OLA inversions only, when there are multiple
 * right hand sides.
 * To have three filenames in bank with redundant info is ok.
 * To have 30 filenames in bank with redundant info is a waste.
 * How to generate filenames on the fly?
 *
 * Allocate space for filenames.  Copy basedirectory name into each
 * file name allocated.
 */
void SetupFileNameBank(BaseDirectory *bd, int nFz0, int nV) {

	if (bd->filenameBank == NULL) {
		bd->nFz0 = nFz0;
		bd->nV = nV;
		// Do a lazy allocation of the file name bank because we
		// don't know nFz0 until after reading the target file
		bd->filenameBank = (char **) td_malloc(nV*nFz0 * sizeof(char *));
		int n;
		for (n = 0; n < nV*nFz0; n++) {
			bd->filenameBank[n] = (char *) td_malloc((bd->maxLength) * sizeof(char));
			strncpy(bd->filenameBank[n], bd->Name, bd->dirLength);
		}
	}
}

// For add parameters (send list of parameters ) ...
void addParameters(BaseDirectory *baseDir, int *values, int nValues, bool debug) {
	char *filename;

	// get the correct string index
	int endOfName;
	if (debug) { // then adding kx_ky_
		endOfName = baseDir->baseLength;
	} else { // then adding mu_sigma_
		endOfName = baseDir->dirLength;
	}
	int maxCharToPrint = baseDir->maxLength - endOfName;
	int n;
	for (n = 0; n < nValues; n++) {
		filename = baseDir->Name;
		snprintf(&filename[endOfName], maxCharToPrint, "%1d_", values[n]);
		// keep the bank of filenames up-to-date
		int a;
		for (a = 0; a < baseDir->nFz0 * baseDir->nV; a++) {
			filename = baseDir->filenameBank[a];
			snprintf(&filename[endOfName], maxCharToPrint, "%1d_", values[n]);
		}

		endOfName = strlen(filename);
	}
	// update the correct string index
	if (debug) {
		baseDir->debugLength = strlen(filename);
	} else {
		if (nValues == 0)
			baseDir->baseLength = baseDir->dirLength;
		else
			baseDir->baseLength = strlen(filename);
	}
}

	// ---------------------------
	// here, Nfz0 is the last parameter before the file name,
	// and Nfz0 is implied, NOT sent as a parameter;
	// ok, just add Nfz0 before adding the file name.
	// Add the Nalpha after the file name.  So, tie Nfz0 and Nalpha to the file name.
	// When the file name is added, add Nfz0 and Nalpha.
	// 1. Call addParameters with mu, sigmah, nu, epsilon
	// 2. Call finishFilename to add Nfz0, filename, NAlpha and the fits extension
	// ---------------------------

// for finish filenames (just send the filename) ...
// NAlpha needs to be a parameter because not all output files have scatterer dependence.
// In fact, only the target files, and the averaging kernels have a scatterer dependence.
// q, w, and noise do not have a scatterer dependence.
char **finishFilenamesGeneral(BaseDirectory *baseDir, const char *UniqueFilename, int NAlpha,
		bool compress, bool debug, const char *extension) {
	char *filename;
	int endOfBaseDir;

	if (baseDir->filenameBank == NULL)
		reportError("File name bank not allocated; call SetupFileNameBank first", "finishFilenames", FATAL);

	// get the correct string index and extension
	if (debug) {
		endOfBaseDir = baseDir->debugLength;
	} else {
		endOfBaseDir = baseDir->baseLength;
		if (compress)
			extension = Constants.gzipExtension;
	}

	int maxCharToPrint = baseDir->maxLength - endOfBaseDir;

	if (baseDir->nFz0 == 1) {
		if (NAlpha > 1) {
			int n;
			for (n = 0; n < baseDir->nV; n++) {
				filename = baseDir->filenameBank[n];
				snprintf(&filename[endOfBaseDir], maxCharToPrint, "%s_%1d.%s",
						UniqueFilename, n+1, extension);
			}
		} else {
			filename = baseDir->filenameBank[0];
			snprintf(&filename[endOfBaseDir], maxCharToPrint, "%s.%s",
					UniqueFilename, extension);
		}
	} else { // multiple depths
		int nf, count;
		count = 0;
		for (nf=1; nf<=baseDir->nFz0; nf++) {
			if (NAlpha > 1) {
				int n;
				for (n = 0; n < baseDir->nV; n++) {
					filename = baseDir->filenameBank[count];
					snprintf(&filename[endOfBaseDir], maxCharToPrint, "%1d_%s_%1d.%s",
							nf, UniqueFilename, n+1, extension);
					printf("filename: %s\n", filename);
					count += 1;
				}
			} else {
				filename = baseDir->filenameBank[count];
				snprintf(&filename[endOfBaseDir], maxCharToPrint, "%1d_%s.%s",
						nf, UniqueFilename, extension);
				count += 1;
			}
		} // end for each depth
	} // end multiple depths

	return baseDir->filenameBank;
}

//------------



// >>>>>>>>>>>> REMEMBER ONLY OLA WILL HAVE MULTIPLE RIGHT HAND SIDES !!!!!!!!! <<<<<<<<<<<<<<

/**
 * Construct a complete file name, assuming no previous strings other than base directory.
 * Use this when X and Y, in pattern basedir/X_Y_M_A.fits, have different meaning than mu and sigma steps.
 * Note: UniqueFilename must be base name, like T, q, etc.  No extension.
 * @param[in,out] baseDir BaseDirectory object already initialized with base directory information.
 * @param[in] UniqueFilename Name of output matrix (M); UniqueFilename must be base name, like T, q, etc.  No extension.
 * @param[in] muStep Value used as X in pattern; integer indicating the nth-value of mu parameter.
 * @param[in] sigmaStep Value used as Y in pattern; integer indicating the nth-value of sigma parameter.
 * For 3d_scalar and 2d, this value is always 1;
 * @return Character string containing complete output filename.  This is a pointer to
 * a character string (which is reused) in BaseDirectory structure.

 * Need to send Nfz0 like NAlpha, because it is embedded in the matrix, like NAlpha
 * AND they need to be in the right order for the write function.
 * What is that order? nV changes faster than M.
 * @ingroup BaseDirectory
 */
char **constructOutputFilenames(BaseDirectory *baseDir, const char *UniqueFilename, int muStep, int sigmaStep) {

	if (baseDir->filenameBank == NULL)
		reportError("File name bank not allocated; call SetupFileNameBank first", "constructOutputFilenames", FATAL);

	int values[2];
	values[0] = muStep;
	values[1] = sigmaStep;
	int nValues = 2;
	bool debug = false;

	// 1.
	addParameters(baseDir, values, nValues, debug);
	// 2.
	char **filenames;
	bool compress = false;
	filenames = finishFilenamesGeneral(baseDir, UniqueFilename, baseDir->nV,
			compress, debug, Constants.extension);
	return filenames;
}

char **constructOutputFilenamesOriginal(BaseDirectory *baseDir, const char *UniqueFilename, int muStep, int sigmaStep, int NAlpha) {
	char *filename;
	int endOfBaseDir;

	endOfBaseDir = baseDir->dirLength;
	int maxCharToPrint = baseDir->maxLength - endOfBaseDir;
	if (NAlpha > 1) {
		int n;
		for (n = 0; n < NAlpha; n++) {
			filename = baseDir->filenameBank[n];
			snprintf(&filename[endOfBaseDir], maxCharToPrint, "%1d_%1d_%s_%1d.%s", muStep, sigmaStep,
					UniqueFilename, n+1, Constants.extension);
		}
	} else {
		filename = baseDir->filenameBank[0];
		snprintf(&filename[endOfBaseDir], maxCharToPrint, "%1d_%1d_%s.%s",
				muStep, sigmaStep, UniqueFilename, Constants.extension);
	}
	return baseDir->filenameBank;
}


// /directory/mu_sigma_kx_ky_A.fits
//                           ^ debugLength
//                     ^ baseLength
//            ^ dirLength
/**
 * Add two parameters to the ouptut filename.  Usually, they are the
 * input parameters (mu and sigma).  Since most output files contain the
 * input parameter steps, the output filename kept in a BaseDirectory object holds the information.
 * Value1 and value2 are appended to base name.  Base name = \<baseName\>value1_value2_
 * This function is also used to add debug parameters to the intermediate matrix filename.
 * @param[in,out] baseDir BaseDirectory object already initialized with base directory information.
 * @param[in] value1 value for first parameter.
 * @param[in] value2 value for second parameter.
 * @param[in] debug True if running with debug option; false otherwise.  If debug, then we are adding
 * the x and y components of the wave number to the output filename.
 * @ingroup BaseDirectory
 */
void addParmsToFilename(BaseDirectory *baseDir, int value1, int value2, bool debug) {

	if (baseDir->filenameBank == NULL)
		reportError("File name bank not allocated; call SetupFileNameBank first", "addParmsToFilename", FATAL);

	int values[2];
	values[0] = value1;
	values[1] = value2;
	int nValues = 2;

	addParameters(baseDir, values, nValues, debug);
}

/**
 * Add a variable number of parameters to the ouptut filename.  Usually, they are the
 * input parameters (mu and sigma).  Since most output files contain the
 * input parameter steps, the output filename kept in a BaseDirectory object holds the information.
 * Value1 and value2 are appended to base name.  Base name = \<baseName\>value1_value2_
 * This function is also used to add debug parameters to the intermediate matrix filename.
 * @param[in,out] baseDir BaseDirectory object already initialized with base directory information.
 * @param[in] values List of values for parameters.
 * @param[in] nValues Number of parameters.
 * @param[in] debug True if running with debug option; false otherwise.  If debug, then we are adding
 * the x and y components of the wave number to the output filename.
 * @ingroup BaseDirectory
 */
// take care to update all filenames
// Maybe have one routine to update all filenames
// One function to add parameters to all files
// One function to finish the filenames.
//
void addParmsToFilename2(BaseDirectory *baseDir, int *values, int nValues, bool debug) {

	if (baseDir->filenameBank == NULL)
		reportError("File name bank not allocated; call SetupFileNameBank first", "addParmsToFilename2", FATAL);

	addParameters(baseDir, values, nValues, debug);
}

/** Insert the matrix name plus any alpha values
 * Assumes the mu and sigma parameters are already inserted.
 * @param[in] baseDir BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name.
 * @param[in] NAlpha number of alpha parameters, usually three.
 * @param[in] compress True if output file should be compressed using gzip.
 * @param[in] debug True if using debug file naming convention.
 * @param[in] extension File name extension.
 * @return fully constructed filenames. Don't free these filenames; they are part of a
 * constant bank of strings.
 * @ingroup BaseDirectory
 */
char **finishFilenamesWExt(BaseDirectory *baseDir, const char *UniqueFilename, int NAlpha,
		bool compress, bool debug, const char *extension) {

	char **filenames;
	filenames = finishFilenamesGeneral(baseDir, UniqueFilename, NAlpha, compress, debug, extension);
	return filenames;
}

/** Insert the matrix name plus any alpha values
 * Assumes the mu and sigma parameters are already inserted.
 * @param[in] baseDir BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name.
 * @param[in] NAlpha number of alpha parameters, usually three.
 * @param[in] compress True if output file should be compressed using gzip.
 * @param[in] debug True if using debug file naming convention.
 * @return fully constructed filenames. Don't free these filenames; they are part of a
 * constant bank of strings.
 * @ingroup BaseDirectory
 */
char **finishFilenames(BaseDirectory *baseDir, const char *UniqueFilename, int NAlpha,
		bool compress, bool debug) {

	return finishFilenamesWExt(baseDir, UniqueFilename, NAlpha,
			compress, debug, Constants.extension);
}

/**
 * Add the matrix name to construct a single output file.
 * @param[in] baseName BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name
 * @param[in] compress True if output file should be compressed using gzip.
 * @param[in] debug True if using debug file naming convention.
 * @return fully constructed filename.
 * @ingroup BaseDirectory
 */
char *finishFilename(BaseDirectory *baseName, const char *UniqueFilename, bool compress, bool debug) {
	char **filenames;
	int NAlpha = 1;

	filenames = finishFilenames(baseName, UniqueFilename, NAlpha, compress, debug);

	return filenames[0];
}

/**
 * no underscore \<Matrix\>\<suffix\> where suffix is 'v_x', 'v_y', or 'v_z' if RLS 3D Vector problem.
 * @param[in] baseDir BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name
 * @param[in] nVectors Number of vectors in problem, usually 1 for RLS 1D and 3D scalar;
 * 3 for RLS 3D vector.
 * @return fully constructed filenames. Don't free these filenames; they are part of a
 * constant bank of strings.
 * @ingroup BaseDirectory
 */
char **constructBFilenames(BaseDirectory *baseDir, const char *UniqueFilename, int nVectors) {
	char *filename;
	const char *extension;
	int endOfName;

	// get the correct string index and extension
		endOfName = baseDir->baseLength;
		extension = Constants.extension;

	int maxCharToPrint = baseDir->maxLength - endOfName;
	if ((nVectors > 1) && (nVectors <= MAX_ALPHA)) {
		int n;
		for (n = 0; n < nVectors; n++) {
			filename = baseDir->filenameBank[n];
			snprintf(&filename[endOfName], maxCharToPrint, "%s%s.%s",
					UniqueFilename, Constants.UNDERV[n], extension);
		}
	} else {
		filename = baseDir->filenameBank[0];
		snprintf(&filename[endOfName], maxCharToPrint, "%s.%s", UniqueFilename, extension);
	}
	return baseDir->filenameBank;
}

/**
 * Construct list of filenames with this format
 * \<Matrix\>\<N\>.fits where N is an integer. Notice, there is no underscore between Matrix name and N.
 * Always compress the output files.
 * @param[in] baseDir BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name
 * @param[in] nFiles number of filenames
 * @param[in,out] fnameLength length of each filename; used when freeing the filenames
 * @return fully constructed filenames. Remember to free these filenames.
 * @ingroup BaseDirectory
 */
char **constructDTildaFilenames(const BaseDirectory *baseDir, const char *UniqueFilename, int nFiles, int *fnameLength) {
	char **filenames;
	int i;
	int baseLength;

	baseLength = baseDir->baseLength;

	*fnameLength = (baseLength +
			MAX_DIGITS_IN_MAP +
			strlen(Constants.gzipExtension) + 1) * sizeof(char);
	filenames = (char **) td_malloc(nFiles * sizeof(char *));

	for (i=0; i<nFiles; i++) {
		filenames[i] = (char *) td_malloc(*fnameLength);
		strncpy(filenames[i], baseDir->Name, baseLength);
		int maxCharToPrint = baseDir->maxLength - baseLength;
		snprintf(&(filenames[i][baseLength]), maxCharToPrint, "%s%1d.%s", UniqueFilename, i+1, Constants.gzipExtension);
	}
	return filenames;
}

/**
 * Construct list of filenames for averaging kernel or covariance matrices for RLS 3d.
 * They have a special format: \<parm1Step\>_\<parm2Step\>_\<matrix name\>_\<depth\>_\<vector\>\<vector\>.fits
 * \<vector\>\<vector\> is xx, xy, xz, yx, yy, etc. and is only for RLS 3D vector problems.
 * \<Matrix\>\<N\>.fits where N is an integer. Notice, there is no underscore between Matrix name and N.
 * Always compress the output files.
 * @param[in] baseDir BaseDirectory structure containing output base directory information.
 * @param[in] UniqueFilename Matrix name
 * @param[in] nFiles number of filenames
 * @param[in] nAvgKernels number of averaging kernels
 * @param[in] nVectors number of vectors, usually 1 or 3.
 * @param[in,out] fnameLength length of each filename; keep and use when freeing the filenames.
 * @return fully constructed filenames. Remember to free these filenames.
 * @ingroup BaseDirectory
 */
char **constructAvgCovFilenames(const BaseDirectory *baseDir, const char *UniqueFilename, int nFiles,
		int nAvgKernels, int nVectors, int *fnameLength) {
	char **filenames;
	int whichKernel;
	int baseLength;

	baseLength = baseDir->baseLength;

	int d, vt, vd;
	// plus 5 for space
	*fnameLength = (baseLength + strlen(UniqueFilename) + MAX_DIGITS_IN_MAP + strlen(Constants.gzipExtension) + 5)
			* sizeof(char);
	filenames = (char **) td_malloc(nFiles * sizeof(char *));

	int maxCharToPrint = *fnameLength - baseLength;

	for (vt=0; vt<nVectors; vt++) {
		for (d=0; d<nAvgKernels; d++) {
			for (vd=0; vd<nVectors; vd++) {
				int idx = vt*nVectors*nAvgKernels + d*nVectors + vd;
				filenames[idx] = (char *) td_malloc(*fnameLength);
				strncpy(filenames[idx], baseDir->Name, baseLength);
				if (subsetKernels()) {
					whichKernel = getAvgKernelDepth(d);
				} else {
					whichKernel = d+1;
				}
				if (nVectors == 1) {
					// 3d scalar problem
					snprintf(&(filenames[idx][baseLength]), maxCharToPrint, "%s_%1d.%s", UniqueFilename, whichKernel, Constants.gzipExtension);
				} else {
					// 3d vector problem
					snprintf(&(filenames[idx][baseLength]), maxCharToPrint, "%s_%1d_%s%s.%s", UniqueFilename, whichKernel, Constants.XYZ[vt], Constants.XYZ[vd], Constants.gzipExtension);
				}
			}
		}
	}
	return filenames;
}

/**
 * Free list of filenames dynamically allocated (i.e. not those from file bank).
 * @param[in] filenames Pointer to array of filenames
 * @param[in] nFiles number of filenames
 * @param[in] fnameLength maximum length of each filename
 * @ingroup BaseDirectory
 */
void freeNFilenames(char **filenames, int nFiles, int fnameLength) {
	int i;

	for (i=0; i<nFiles; i++) {
		td_free(filenames[i], fnameLength);
	}
	td_free(filenames, nFiles * sizeof(char *));
}

