/*
 * Inversion.h
 *
 *  Created on: Nov 1, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#ifndef INVERSION_H_
#define INVERSION_H_

#include "InversionData.h"
#include "FileMapAndParameters.h"

/**
 * \mainpage
 *
 * The software performs a time-distance inversion of travel-time maps
 * and sensitivity kernels and puts all the results in a specified directory.
 * The results include inversion velocities, averaging
 * kernels and error covariances.  Velocities are also called models throughout the software.
 * All input and output data are in FITS file format.
 * An input file specifies regularization parameters, input, and output file locations.
 * The inversion code loops through the regularization parameters and performs an inversion at each
 * combination of values.
 * The software is written in Python and C.  Python scripts call C-library functions
 * to parse the input file (libpars.so) and calculate inversions.
 * The parser takes an input file and returns an internal data structure (InputParms)
 * which is passed to an inversion function.  The internal data structure is
 * passed to the forward calculation, as well.
 *
 * \section InvTypes Functions
 * \ref InvTypesPage
 *
 * \ref Forward
 *
 * \section Structure Software Structure
 * \ref page2.
 *
 * \page page2 Software Structure
 * The main data types in the inversion code are MATRIX, MAP_DEPTH_MATRIX, Model, and BaseDirectory.
 * Only three dimensional problems use MAP_DEPTH_MATRIX. All inversions
 * use MATRIX, Model, and BaseDirectory.
 *
 * \section MATRIX
 * MATRIX is a structure that contains 2D data as either double or double complex
 * numbers.
 *
 * \section MAP_DEPTH_MATRIX
 * MAP_DEPTH_MATRIX is a four-dimensional matrix and it keeps input and output data.
 * This is a main data type for the 2D and 3D problems.  The input kernel data,
 * input map data, solution data, averaging kernel data and covariance
 * data are all kept in a MAP_DEPTH_MATRIX object.  During the inversion,
 * pointers to 2D sub-matrices are passed around and
 * accumulated to make full solutions.  Map-depth data are kept in
 * contiguous memory in order to make the inversion run faster.  This is done at the
 * expense of data in the xy-plane.  Since input and output to FITS
 * files use the xy-plane, the expense is incurred during I/O which
 * is inherently slow.  Input files are read only once while the
 * inversion calculations are performed many times.
 *
 *\section Model
 * The Model data type represents a solution to an inversion.
 * Solutions at each (kx,ky) accumulate into full solutions.
 * If the problem is 1D, then there is only one kx and one ky.
 * If the problem is 2D or 3D, then the code aggregates solutions at each (kx,ky)
 * into
 * a full model.  When a model is complete, all the results are
 * written to FITS files.
 *
 *\section BaseDirectory
 * The BaseDirectory structure is a single instance and it
 * manages the string manipulations
 * needed to generate the various output file names.
 */


/**
 * Inversion structure contains all input and output data,
 * inversion parameters, and pointers to functions
 * which perform specific inversion algorithms.
 */
struct Inversion {
	InversionData *data;								///< Pointer to input and output data for inversion
	// define function pointers
	InversionData *(*init) (InputParms, char *, bool);	///< Pointer to function to initialize inversion; read data from input file.
	void (*solve) (InversionData *);					///< Pointer to inversion function.
	void (*complete) (InversionData *);					///< Pointer to function to complete inversion; accumulate intermediate results.
	void (*destroy) (InversionData *);					///< Pointer to function to destroy inversion data.
	bool (*moreParameters) (InversionData *);		///< Pointer to function that moves parameters to next step.
	void (*nextParameter) (InversionData *);			///< Pointer to function that determines end of parameter loop.
};

#endif /* INVERSION_H_ */
