/*
 * Model.c
 *
 *  Created on: Aug 18, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <math.h>
#include "Constants.h"
#include "Model.h"
#include "BaseDirectory.h"
#include "InversionErrorMessages.h"
#include "errorHandler.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "FITSParameters.h"
#include "utilities.h"
#include "averagingKernel.h"
#include "ProblemDimension.h"
#include "IgnoreK0.h"

/**
 * @defgroup Model Model - Inversion Solution and accompanying data.
 * @ingroup InversionData
 */

/** Initialize Model object, allocate space for data and constant values.
 * Use for OLA inversions.
 * @param[in,out] model Initialize this Model.
 * @param[in] X Number of grid points, Nx
 * @param[in] Y Number of grid points, Ny
 * @param[in] M Number of measurements
 * @param[in] Z Number of grid points, Nz
 */
void initializeModel(Model *model, int X, int Y, int M, int Z) {

	bool shift = false;
	// initialize w, but, it can change when we get a call to saveW;
	// change Nfz0 if needed when saving the first w.
	model->w = newMapDepthMatrix(X, Y, 1, M, shift);
	model->parms.n = MAX_FITS_PARMS;
	model->parms.names[0] = Constants.ModelMuName;
	model->parms.names[1] = Constants.ModelSigmaName;
	model->parms.names[2] = Constants.ModelNuName;
	model->parms.names[3] = Constants.ModelEpsilonName;
	model->parms.comments[0] = Constants.ModelComment;
	model->parms.comments[1] = Constants.ModelComment;
	model->parms.comments[2] = Constants.ModelComment;
	model->parms.comments[3] = Constants.ModelComment;
	model->q.data = NULL;
	model->AvgKern.data = NULL;
	model->noise.data = NULL;
	model->noMaps = false;
}

/**
 * Initialize output model for use by RLS inversion
 * @param[in,out] model Pointer to Model structure
 * @param[in] kernelTdm Matrix containing 3D RLS Kernel.
 */
void initializeOutputModel(Model *model, MAP_DEPTH_MATRIX kernelTdm) {
	int x, y, z, t, v;
	x = kernelTdm.nX;
	y = kernelTdm.nY;
	z = kernelTdm.nZ;
	int useShift = true;

	v = 1;
	if (getProblemDimension() == PDC.D5) {
		v = 3;
	}
	t = 1;
	// output matrices are based on the kernel dimensions
	// allocate space
	model->bFull = newMapDepthMatrixV(x, y, z, t, v, useShift); // t = 1
	t = kernelTdm.nT;
	z = 1;
	model->dTilde = newMapDepthMatrix(x, y, z, t, useShift); // z = 1
	z = kernelTdm.nZ;
	if (calcAvgKernelCov()) {
		t = getNumberOfAveragingKernels();
		model->AvgKern = newMapDepthMatrixV(x, y, z, t * v, v, useShift); // t = # of averaging kernels to calculate
		model->noise = newMapDepthMatrixV(x, y, z, t * v, v, useShift);
	}
	model->kxGridValues = newMatrix(x, y);
	model->kyGridValues = newMatrix(x, y);
	model->lambdaGridValues = newMatrix(x, y);
	model->kxkyDone = false;
	model->parms.n = MAX_FITS_PARMS - 1;  // there are only two parameters for RLS inversions
	model->parms.names[0] = Constants.ModelLambda1Name;
	model->parms.names[1] = Constants.ModelLambda2Name;
	model->parms.comments[0] = Constants.ModelLambdaComment;
	model->parms.comments[1] = Constants.ModelLambdaComment;
}

/**
 * Initialize output model for use by forward calculation
 * @param[in,out] model Pointer to Model structure
 * @param[in] kernelTdm Matrix containing 3D Kernel.
 */
void initializeForwardOutputModel(Model *model, MAP_DEPTH_MATRIX kernelTdm) {
	int x, y, z, t;
	x = kernelTdm.nX;
	y = kernelTdm.nY;
	z = kernelTdm.nZ;
	int useShift = true;

	t = 1;
	// output matrices are based on the kernel dimensions
	// allocate space
	t = kernelTdm.nT;
	z = 1;
	// dTilde is in real space, but kernels are in wave space, so
	// make the conversions in the dimensions
	y = x; // assume the xy-plane is square; don't use this: 2*(y-1);
	model->dTilde = newMapDepthMatrix(x, y, z, t, useShift); // z = 1
	z = kernelTdm.nZ;

	model->kxkyDone = false;
	model->parms.n = 1;
	model->parms.names[0] = Constants.ModelNoiseScaleName;
	model->parms.comments[0] = Constants.NoiseScaleComment;
}

void initializeCalcQOutputModel(Model *model, MAP_DEPTH_MATRIX w) {

	clearParameters(model);

	// Should I copy the parameters from the weights header file
	// and then repeat them in the q file? YES!
	//model->parms.n = 0; //  MAX_FITS_PARMS;
	model->parms.n = MAX_FITS_PARMS;
	model->parms.names[0] = Constants.ModelMuName;
	model->parms.names[1] = Constants.ModelSigmaName;
	model->parms.names[2] = Constants.ModelNuName;
	model->parms.comments[0] = Constants.ModelComment;
	model->parms.comments[1] = Constants.ModelComment;
	model->parms.comments[2] = Constants.ModelComment;

	model->q.data = NULL;
	model->AvgKern.data = NULL;
	model->noise.data = NULL;
	model->w = w;
}

/**
 * Free all memory associated with outputModel.
 * @ingroup OutputModel
 */
void freeModel(Model *model) {
    if (!isEmptyMapDepthMatrix(&(model->AvgKern)))
		destroyMapDepthMatrix(model->AvgKern);
	if (!isEmptyMapDepthMatrix(&(model->noise)))
		destroyMapDepthMatrix(model->noise);
	destroyMapDepthMatrix(model->q);
	destroyMapDepthMatrix(model->w);

}

/**
 *
 */
void useNoMaps(Model *model) {
	model->noMaps = true;
}

/**
 * Record parameter values in the model.
 * @param[in,out] model A Model object.
 * @param[in] mu Value for mu parameter.
 * @param[in] sigma Value for sigma parameter.
 */
void setParameters(Model *model, Float mu, Float sigma) {
	snprintf(&(model->parms.scientificNotation[0][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, mu);
	snprintf(&(model->parms.scientificNotation[1][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, sigma);
	model->mu = mu;
	model->sigma = sigma;
	model->parms.n = 2;
}

/**
 * Record multiple parameter values in the model.
 * @param[in,out] model A Model object.
 * @param[in] values Pointer to array of parameter values. NOTE: there is a special ordering
 * to the parameters, mu, sigma, then nu (which is optional).
 * @param[in] nValues How many parameter values.
 */
void setParameters2(Model *model, Float *values, int nValues) {
	int i;
	for (i=0; i<nValues; i++) {
		snprintf(&(model->parms.scientificNotation[i][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, values[i]);
	}
	model->mu = values[0];
	model->sigma = values[1];
	if (nValues > 2) {
		model->nu = values[2];
	} else {
		model->nu = 0.0;  // if no crosstalk, then nValues = 2, and
						  // set default value for nu (equal to zero), when there is no crosstalk
		                  // This is related to Parameter.initLoopParameters() when there is no
		                  // crosstalk, we need a default value for nu.
	}
	model->parms.n = nValues;
}

/**
 * Clear parameter values (mu and sigma) in the model.
 * Output FITS files will not have sigma and mu parameters in the header.
 * @param[in,out] model A Model object.
 */
void clearParameters(Model *model) {
	model->parms.n = 0;
}

/**
 * Save inversion weights, w, for wave number, kx, ky.
 * @param[in,out] model A Model object.
 * @param[in] w Inversion weights.
 * @param[in] kx X-index in Fourier space.
 * @param[in] ky Y-index in Fourier space.
 */
void saveW(Model *model, MATRIX w, int kx, int ky) {
	// ok, now we really know the true dimension of the weights.
	// The number of target depths wasn't known when
	// w was initialized, so check the dimensions and increase
	// w as needed for multiple target depths.

	//if ((w.nRow != model->w.nT) || (w.nCol != model->w.nZ)) {
	if (w.nRow != model->w.nT) {
		reportError(ModelBadDimForW, "Model.saveW", FATAL);
	}
	if (w.nCol != model->w.nZ) {
	//	reportError(ModelBadDimForW, "Model.saveW", FATAL);
	//  free w, and allocate new one based on the size of w coming in.
		// save dimensions:
		int Nx, Ny;
		Nx = model->w.nX;
		Ny = model->w.nY;
		bool shift;
		shift = model->w.shift;
		MAP_DEPTH_MATRIX wc;
		wc = model->w;
		printf("w was (x,y,z,m,v)=(%d,%d,%d,%d,%d)\n", wc.nX, wc.nY, wc.nZ, wc.nT, wc.nV);
		destroyMapDepthMatrix(model->w);
		//  w.nCol = nZ * nV, and nV is always 1 for weights.
		model->w = newMapDepthMatrix(Nx, Ny, w.nCol, w.nRow, shift);
		printf("Increasing size of w to accommodate multiple target depths\n");
		wc = model->w;
		printf("w now (x,y,z,m,v)=(%d,%d,%d,%d,%d)\n", wc.nX, wc.nY, wc.nZ, wc.nT, wc.nV);
	}

	insertMapDepth(kx, ky, w, &(model->w));
}

/**
 * Save inversion solution, b, for RLS 1D.
 * @param[in,out] model A Model object.
 * @param[in] b Solution matrix.
 * @param[in] nStep Parameter step number.
 */
void saveB(Model *model, MATRIX b, int nStep) {

	Float *p;
	p = model->bVECTORS.data;

	// insert solution vectors as rows, then just transpose to output as column vectors
	VECTOR bv;
	bv = MatrixToVector(b);
	int idx;
	idx = (nStep - 1) * bv.n;
	insertVector(&p[idx], bv);
}

/**
 * Calculate the model, q, at a given kx, ky in wave space.
 * @param[in] wk A MATRIX containing the solution weights at a specific kx,ky in wave space.
 * @param[in] tau Travel-map matrix
 * @param[in] kx The x-coordinate in wave space.
 * @param[in] ky The y-coordinate in wave space.
 * @returns a Matrix containing the model, q, at a particular kx,ky.
 */
MATRIX calculateQK(MATRIX wk, MAP_DEPTH_MATRIX tau, int kx, int ky) {

	// extract t(k)
	MATRIX tk = extractMapDepth(kx, ky, tau);

	// calculate q(k)
	// model.q(k) = w*(k) * t(k);
	MATRIX qk = multiplyMMFull('C', 'N', TWO_PI_SQUARED, 0.0, wk, tk);

	return qk;
}

/**
 * Calculate the averaging kernel at a given kx, ky in wave space.
 * @param[in] wk A MATRIX containing the solution weights at a specific kx,ky in wave space.
 * @param[in] K A MAP_DEPTH_MATRIX containing the kernels.
 * @param[in] kx The x-coordinate in wave space.
 * @param[in] ky The y-coordinate in wave space.
 * @returns a Matrix containing the averaging kernel at a particular kx,ky.
 */
MATRIX calculateAvgK(MATRIX wk, MAP_DEPTH_MATRIX K, int kx, int ky) {
	// extract tK(k)
	MATRIX Kk = extractMapDepth(kx, ky, K);
	MATRIX AvgKernk = multiplyMMFull('T', 'N', TWO_PI_SQUARED, 0.0, wk, Kk);

	return AvgKernk;
}

/**
 * Calculate the noise at a given kx, ky in wave space.
 * @param[in] wk A MATRIX containing the solution weights at a specific kx,ky in wave space.
 * @param[in] Lambda A MAP_DEPTH_MATRIX containing the input covariance.
 * @param[in] kx The x-coordinate in wave space.
 * @param[in] ky The y-coordinate in wave space.
 * @returns a Matrix containing the averaging kernel at a particular kx,ky.
 */
MATRIX calculateNoiseK(MATRIX wk, MAP_DEPTH_MATRIX Lambda, int kx, int ky) {
	MATRIX Lambdak = extractMapDepth(kx, ky, Lambda);
	MATRIX temp = multiplyMMFull('C', 'N', TWO_PI_SQUARED, 0.0, wk, Lambdak);
	// for each fz0, multiply row of temp by column of Lambdak
	int nfz0;

	MATRIX temp2 = multiplyMMFull('N', 'N', TWO_PI_SQUARED, 0.0, temp, wk);

	MATRIX noisek;
	noisek = newMatrixG(temp.nRow, 1, COMPLEX);
	// Extract the diagonal elements if Nfz0 > 1
	FloatComplex *nkPtr, *tmpPtr;
	nkPtr = noisek.data;
	tmpPtr = temp2.data;
	for (nfz0=0; nfz0<temp.nRow; nfz0++) {
		*nkPtr = *tmpPtr;
		nkPtr+=1;
		tmpPtr += temp2.nCol + 1;
	}

	freeMatrix(temp);
	freeMatrix(temp2);
	return noisek;
}

/**
 * Calculate q
 * @param[in] model A Model object.
 * @param[in] tau Travel-map matrix
 * @param[in] DValueShift Boolean indicating whether to shift output matrix q.
 */
void calculateQValues(Model *model, MAP_DEPTH_MATRIX tau, bool DValueShift) {

	int kx, ky;
	int Nx, Ny;
	MATRIX wk;
	MATRIX qk;

	Nx = model->w.nX;
	Ny = model->w.nY;

	bool shift = DValueShift;
	if (model->q.data == NULL) {
		model->q = newMapDepthMatrix(Nx, Ny, tau.nZ, 1, shift);
	}

	//for each wavenumber, k = (kx, ky)
	for (kx = 0; kx < Nx; kx++) {
		for (ky = 0; ky < Ny; ky++) {
			// extract w(k)
			wk = extractMapDepth(kx, ky, model->w);

			qk = calculateQK(wk, tau, kx, ky);
			// insert q(k)
			insertMapDepth(kx, ky, qk, &(model->q));

			// free some memory
			freeMatrix(qk);
		}
	}
}


/**
 * Calculate q, Averaging Kernel and noise squared.
 * @param[in] model A Model object.
 * @param[in] K Kernel matrix
 * @param[in] tau Travel-map matrix
 * @param[in] Lambda Travel-map noise covariance matrix
 * @param[in] DValueShift Boolean indicating whether to shift output matrices q, and averaging kernel, and noise.
 */
void calculateDerivedValues(Model *model, MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX tau, MAP_DEPTH_MATRIX Lambda,
		bool DValueShift) {

	if (model->noMaps)
		calculateDerivedValuesNoQ(model, K, Lambda, DValueShift);
	else {

	// calculate averaging kernel first, because it doesn't need complex conjugate of w
	// and calculating conjugate of w changes it

		int kx, ky;
		int Nx, Ny;
		MATRIX wk, noisek, AvgKernk;
		MATRIX qk;

		Nx = model->w.nX;
		Ny = model->w.nY;

		int nFz0;
		nFz0 = model->w.nZ;

		bool shift = DValueShift;
		if (model->q.data == NULL) {
			// with multiple target depths, q has dimension:
			// Nx, Ny, tau.nZ=1, M=Nfz_0, nV=1
			model->q = newMapDepthMatrix(Nx, Ny, tau.nZ, nFz0, shift);
		}

		// AvgKern and noise are allocated in LSQRInit; check for null values here before creating new ones.
		if (model->AvgKern.data == NULL) {
			// For multiple target depths, use these dimensions: Nx, Ny, K.nZ, Nfz_0, K.nV
			// otherwise,                  use these dimensions: Nx, Ny, K.nZ, 1, K.nV
			model->AvgKern = newMapDepthMatrixV(Nx, Ny, K.nZ, nFz0, K.nV, shift);
		}

		if (model->noise.data != NULL) destroyMapDepthMatrix(model->noise);
		shift = true;
		// For multiple target depths, how does the dimension of the noise change?
		// It becomes nFz0 x nFz0, but what does this mean?
		// Only the diagonal elements make sense
		model->noise = newMapDepthMatrix(Nx, Ny, 1, nFz0, shift);

		//for each wavenumber, k = (kx, ky)
		for (kx = 0; kx < Nx; kx++) {
			for (ky = 0; ky < Ny; ky++) {

				// extract w(k)
				wk = extractMapDepth(kx, ky, model->w);

				// calculate q(k)
				qk = calculateQK(wk, tau, kx, ky);
				// insert q(k)
				insertMapDepth(kx, ky, qk, &(model->q));

				// calculate AvgKern
				AvgKernk = calculateAvgK(wk, K, kx, ky);
				// insert AvgKernk(k)
				insertMapDepth(kx, ky, AvgKernk, &(model->AvgKern));

				// calculate Lambda(k)
				noisek = calculateNoiseK(wk, Lambda, kx, ky);

				// insert noise(k)
				insertMapDepth(kx, ky, noisek, &(model->noise));

				freeMatrix(AvgKernk);
				freeMatrix(noisek);

				// free some memory
				freeMatrix(qk);
			}
		}
	}
}


void calculateDerivedValuesNoQ(Model *model, MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX Lambda,
		bool DValueShift) {

// calculate averaging kernel first, because it doesn't need complex conjugate of w
// and calculating conjugate of w changes it

	int kx, ky;
	int Nx, Ny;
	MATRIX wk, noisek, AvgKernk;

	Nx = model->w.nX;
	Ny = model->w.nY;

	bool shift = DValueShift;

	// AvgKern and noise are allocated in LSQRInit; check for null values here before creating new ones.
	if (model->AvgKern.data == NULL) {
		model->AvgKern = newMapDepthMatrixV(Nx, Ny, K.nZ, 1, K.nV, shift);
	}

	if (model->noise.data != NULL) destroyMapDepthMatrix(model->noise);
	shift = true;
	model->noise = newMapDepthMatrix(Nx, Ny, 1, 1, shift);

	//for each wavenumber, k = (kx, ky)
	for (kx = 0; kx < Nx; kx++) {
		for (ky = 0; ky < Ny; ky++) {
			// extract w(k)
			wk = extractMapDepth(kx, ky, model->w);

			// calculate AvgKern
			AvgKernk = calculateAvgK(wk, K, kx, ky);
			// insert AvgKernk(k)
			insertMapDepth(kx, ky, AvgKernk, &(model->AvgKern));

			// calculate noise
			noisek = calculateNoiseK(wk, Lambda, kx, ky);

			// insert noise(k)
			insertMapDepth(kx, ky, noisek, &(model->noise));

			freeMatrix(AvgKernk);
			freeMatrix(noisek);
		}
	}
}

/**
 * Write solution, q, for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeQ(Model *model, BaseDirectory *baseName) {
	bool useFT = true;
	char **filenames, *filenameBase;

	if (!model->noMaps) {
		bool debugPosition = false;
		bool compress = false;
		if (IgnoreK0()) {
			filenameBase = Constants.QK0Filename;
		} else {
			filenameBase = Constants.QFilename;
		}

		int nFiles;
		nFiles = baseName->nFz0;
		int NAlpha = 1;  // because q is not dependent on number of scatterers
		filenames = finishFilenames(baseName, filenameBase, NAlpha, compress, debugPosition);
		writeMapDepthMatrix(filenames, nFiles, model->q, useFT, model->parms);
	}
}

/**
 * Write OLA inversion weights, w, to file.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeW(Model *model, BaseDirectory *baseName) {
	bool useFT = true;
	char *filenameBase;

	bool debugPosition = false;
	bool compress = false;
	if (IgnoreK0()) {
		filenameBase = Constants.WK0Filename;
	} else {
		filenameBase = Constants.WFilename;
	}
	if (model->w.nZ == 1) {
		char *filename;
		filename = finishFilename(baseName, filenameBase, compress, debugPosition);
		printf("writing w to %s\n", filename);
		writeMapDepthMatrixToOneFile(filename, model->w, useFT, model->parms);
	} else {
		// add NAlpha to the averaging kernel filenames
		char **filenames;
		int NAlpha = 1; // because the weights don't have a scatterer dependence
		filenames = finishFilenames(baseName, filenameBase, NAlpha, compress, debugPosition);
		// There will be Nfz0 filenames;
		int nFiles;
		nFiles = baseName->nFz0;
		writeDepthMapMatrix(filenames, nFiles, model->w, useFT, model->parms);
	}
}

/**
 * Write averaging kernel to file for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeAvgKern(Model *model, BaseDirectory *baseName) {
	bool useFT = true;
	bool debugPosition = false;
	bool compress = true;

	int NAlpha;
	NAlpha = model->AvgKern.nV;
	int NFz0 = model->AvgKern.nT;
	char **filenames, *filenameBase;
	if (IgnoreK0()) {
		filenameBase = Constants.AvgKernK0Filename;
	} else {
		filenameBase = Constants.AvgKernFilename;
	}
	// add NAlpha to the averaging kernel filenames
	filenames = finishFilenames(baseName, filenameBase, NAlpha, compress, debugPosition);
	writeMapDepthMatrix(filenames, NAlpha*NFz0, model->AvgKern, useFT, model->parms);

}

/**
 * Write the  Averaging Kernel data for 1d problem.
 * @param[in] model The OutputModel containing the Averaging Kernel.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 * @param[in] lambda Which lambda value was used for this solution.
 * @ingroup OutputModel
 */
void writeAVectors1d(Model *model, BaseDirectory *baseName, Float lambda) {

		//output A
		if (model->A1d.data == NULL) {
			reportError(NullMatrix,
					"Averaging Kernel (Matrix A) in Model", ERROR);
		} else {
			// write fits file
			fits2d result;
			result = MatrixToFitsObject(model->A1d);
			bool debugPosition = false;
			bool compress = false;
			char *filename;
			filename = finishFilename(baseName, Constants.AvgKernFilename1d, compress, debugPosition);
			writefits2dWithLambda(filename, result, lambda);
			td_free(result.data, result.xDim * result.yDim * sizeof(Float));
		}
}

/**
 * Write noise estimate to file for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeNoise(Model *model, BaseDirectory *baseName) {
	bool useFT = true;
	char **filenames, *filenameBase;

	bool debugPosition = false;
	bool compress = true;

	printf("writing noise\n");
	// Only printing the diagonal elements from each sigma^2(k)

	if (IgnoreK0()) {
		filenameBase = Constants.NoiseK0Filename;
	} else {
		filenameBase = Constants.NoiseFilename;
	}

	int nFiles;
	nFiles = baseName->nFz0;
	int NAlpha = 1;  // because noise is not dependent on number of scatterers
	filenames = finishFilenames(baseName, filenameBase, NAlpha, compress, debugPosition);
	writeMapDepthMatrix(filenames, nFiles, model->noise, useFT, model->parms);
}

/**
 * Write the Error Covariance data for 1d problem.
 * @param[in] model Model structure containing the Error Covariance.
 * @param[in] baseName Which lambda step was used for this solution.
 * @param[in] lambda Which lambda value was used for this solution.
 * @ingroup OutputModel
 */
void writeCovVectors1d(Model *model, BaseDirectory *baseName, Float lambda) {

		//output A
		if (model->Cov1d.data == NULL) {
			reportError(NullMatrix, "Covariance Matrix in Model", ERROR);
		} else {
			// write fits file
			fits2d result;
			result = MatrixToFitsObject(model->Cov1d);

			bool debugPosition = false;
			bool compress = false;
			char *filename;
			filename = finishFilename(baseName, Constants.CovFilename, compress, debugPosition);
			writefits2dWithLambda(filename, result, lambda);
			td_free(result.data, result.xDim * result.yDim * sizeof(Float));
		}
}

/**
 * Write solution matrix for RLS 1D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeBVectors1d(Model *model, BaseDirectory *baseName) {

	printf("writing solution files ... \n");
	MATRIX bVECTORST = transpose(model->bVECTORS);
	freeMatrix(model->bVECTORS);
	// write B matrix to a fits file

	fits2d result;
	result = MatrixToFitsObject(bVECTORST);

	bool debugPosition = false;
	bool compress = false;
	char *filename;
	addParmsToFilename2(baseName, NULL, 0, debugPosition);
	filename = finishFilename(baseName, Constants.BFilename, compress, debugPosition);
	writefits2d(filename, result);
	td_free(result.data, result.xDim * result.yDim * sizeof(Float));

}

/**
 * Write solution B vectors to FITS files
 * matrix for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeBVectors3d(Model *model, BaseDirectory *baseName) {
	bool useFT = true;
	char **filenames;
	filenames = constructBFilenames(baseName, Constants.BFilename, model->bFull.nV);

	writeMapDepthMatrix(filenames, model->bFull.nT * model->bFull.nV, model->bFull, useFT, model->parms);
}

/**
 * Save lambda values and kx, ky values for each wave number  (RLS 3D inversion).
 * @param[in,out] m A Model object.
 * @param[in] r Row (index into matrix).
 * @param[in] c Column (index into matrix).
 * @param[in] kx Value of kx at (x,y)
 * @param[in] ky Value of ky at (x,y)
 * @param[in] lambda Value of lambda (regularization) parameter.
 */
void saveGridValues(Model *m, int r, int c, Float kx, Float ky, Float lambda) {
	Float *p;
	int idx;

	idx = ROW_MAJOR_IDX(r, c, m->kxGridValues.nCol);
	if (!m->kxkyDone) {
		p = m->kxGridValues.data;
		p[idx] = kx;
		p = m->kyGridValues.data;
		p[idx] = ky;
	}
	p = m->lambdaGridValues.data;
	p[idx] = lambda;
}

/**
 * Write lambda values and kx, ky values for each wave number  (RLS 3D inversion).
 * Write kxGridValues, kyGridValues and lambdaGridValues at each grid point to FITS files
 * matrix for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeGridValues3d(Model *model, BaseDirectory *baseName) {
	// write kxGridValues, kyGridValues and lambdaGridValues at each grid point to FITS files
	char* filename;
	bool debugPosition = false;
	bool compress = false;

	filename = finishFilename(baseName, Constants.LAMBDA_FILENAME, compress, debugPosition);
	writefits2d(filename, MatrixToFitsObject(model->lambdaGridValues));

	if (!model->kxkyDone) {
		addParmsToFilename2(baseName, NULL, 0, debugPosition);
		filename = finishFilename(baseName, Constants.KX_FILENAME, compress, debugPosition);
		writefits2d(filename, MatrixToFitsObject(model->kxGridValues));
		filename = finishFilename(baseName, Constants.KY_FILENAME, compress, debugPosition);
		writefits2d(filename, MatrixToFitsObject(model->kyGridValues));
		model->kxkyDone = true;
	}
}

/**
 * Write d~ matrices for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 * @param[in] useFT Boolean indicates if d~ matrices should be Fourier transformed.
 */
void writeDTildaValues3d(Model *model, BaseDirectory *baseName, bool useFT) {

	char **filenames;
	int nFiles;
	int fnameLength;

	printf("writing d~ Files ... \n");
	nFiles = model->dTilde.nT;
	filenames = constructDTildaFilenames(baseName, Constants.DTILDA_FILENAME, nFiles, &fnameLength);
	printf("writing to %s\n", filenames[0]);
	writeMapDepthMatrix(filenames, nFiles, model->dTilde, useFT, model->parms);
	freeNFilenames(filenames, nFiles, fnameLength);
}

/**
 * Write averaging kernels and error covariance matrices for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeAvgKernValues3d(Model *model, BaseDirectory *baseName) {

	if (calcAvgKernelCov()) {
		bool useFT = true;
		char **filenames;
		int nAvgKernels;
		int nMaps;
		int nVectors;
		int nFiles;
		int fnameLength;

		printf("writing Averaging Kernels  ... \n");
		// multiply by (1/(2pi)^2 to correct for ??; bug detected in superg test data
		TdmMultiplyScalar(&(model->AvgKern), 1.0/TWO_PI_SQUARED);

		nMaps = model->AvgKern.nT;
		nVectors = model->AvgKern.nV;

		if (subsetKernels()) {
			nAvgKernels = getNumberOfAveragingKernels();
		} else {
			nAvgKernels = nMaps;
		}

		nFiles = nAvgKernels * nVectors * nVectors;
		filenames = constructAvgCovFilenames(baseName,
				Constants.AvgKernFilename, nFiles, nAvgKernels, nVectors, &fnameLength);

		writeMapDepthMatrix(filenames, nFiles, model->AvgKern, useFT, model->parms);
		freeNFilenames(filenames, nFiles, fnameLength);

		printf("writing Error Covariance Files ... \n");
		// multiply by ((2pi)/N)^2 to correct for two Fourier transforms instead of one
		Float factorForCI;
		Float x;
		x = model->noise.nX;
		factorForCI = ((2.0*M_PI)*(2.0*M_PI)) / (x * x);
		TdmMultiplyScalar(&(model->noise), factorForCI);

		filenames = constructAvgCovFilenames(baseName,
				Constants.CovFilename, nFiles, nAvgKernels, nVectors, &fnameLength);

		writeMapDepthMatrix(filenames, nFiles, model->noise, useFT, model->parms);
		freeNFilenames(filenames, nFiles, fnameLength);
	}
}


