/*
 * averagingKernel.c
 *
 *  Created on: Apr 30, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include "averagingKernel.h"
#include "utilities.h"
#include "List.h"
#include "ProblemDimension.h"
#include "errorHandler.h"
#include "InversionErrorMessages.h"

/**
 * @defgroup AveragingKernel AveragingKernel - Store the depths for averaging kernel and error covariance calculations.
 * @ingroup SupportingModules
 */
VECTOR depthsForReducing; 	///< Vector that contains offsets into Avg Kern & Cov matrices for 5d and 3d problem. Used when only calculating a subset of Averaging Kernels. @ingroup AveragingKernel
VECTOR depths;				///< Vector contains depths for Averaging Kernel calculations. @ingroup AveragingKernel
bool calculateSubset = false;

const static char *MINUS1 = "-1";
const static char *KERNEL_STR = "Kernel";
const static char *AVG_STR = "Avg";

/** Send a string, parse it and store the depths for averaging kernel calculation
 * @param[in] AvgKernInputLine Text line from input file specifying averaging kernel
 * 				calculations
 * @param[in] maxDepth maximum value for a depth; 1 is the minimum, the size
 * 				of the z-dimension in the input kernel files is the maximum.
 * @ingroup AveragingKernel
 */
void initializeAvgKernelCalculation(char *AvgKernInputLine, int maxDepth) {
	int i;
	List l;
	char *token;

	// parser avg kernel line of input
	stringTokenizer(AvgKernInputLine);

	l = newList();

	// throw away first two tokens, they should be
	// Avg Kernel
	token = nextToken();
	// compare these to lower version, send error message or warning.
	if (lowerStrNCmp(token, AVG_STR, strlen(AVG_STR)) != true) {
		reportError(AvgStrExpected, token, WARNING);
	}
	token = nextToken();
	if (lowerStrNCmp(token, KERNEL_STR, strlen(KERNEL_STR)) != true) {
		reportError(KernelStrExpected, token, WARNING);
	}
	token = nextToken();
	if (token == NULL) {
		// calculate averaging kernels at every level
		calculateSubset = false;
		depths.n = maxDepth;
	} else if (strncmp(token, MINUS1, strlen(MINUS1)) == 0) {
		// check for -1 which indicates no averaging kernel calculations
		calculateSubset = true;
		depths.n = 0;
	} else {
		while (token != NULL) {
			insertList(token, l);
			token = nextToken();
		}
		// convert token list into list of depths for averaging kernel calculation
		int ndepths = getLengthList(l);
		if (ndepths > maxDepth) {
			reportErrorInt(AvgKernTooManyValues, ndepths, "", FATAL);
		}
		depths = newVector(ndepths);

		ListIterator li = newListIterator(l);
		int intVal;
		int nconversions;
		for (i = ndepths - 1; i >= 0; i--) {
			token = listIteratorNext(li);
			nconversions = sscanf(token, "%d", &intVal);
			if (nconversions != 1) {
				reportError(AvgKernBadDepthValue, token,
						FATAL);
			}
			// the numbers do  NOT need to be in sorted order.
			if ((intVal > maxDepth) || (intVal < 0)) {
				reportError(AvgKernInvalidDepthValue, token, FATAL);
			} else {
				depths.data[i] = (double) intVal;
			}
		}
		//freeList(l);
		calculateSubset = true;
		// prepare a vector for reducing a matrix
		if (getProblemDimension() == 5) {
		// if 5d problem, then calculate offsets into each v-dim section
			int vdim = 3;
			depthsForReducing = newVector(ndepths*vdim);
			int v;
			for (v = 0; v < vdim; v++) {
				for (i = 0; i < depths.n; i++) {
					int idx = v*ndepths + i;
					int offset = v*maxDepth + depths.data[i];
					depthsForReducing.data[idx] = offset;
				}
			}
		} else {
			// 3d problem, just use the depths vector;
			depthsForReducing = depths;
		}
	}
	freeList(l);
}

/**
 * Return true if only a subset of the averaging kernels need to be calculated.
 * @ingroup AveragingKernel
 */
bool subsetKernels() {
	 // then only calculate needed columns for the averaging kernels
	return calculateSubset;
}

/**
 * Return true if calculating ANY averaging kernels
 * @ingroup AveragingKernel
 */
bool calcAvgKernelCov() {
	//						calculateSubset	depths.n
	// calculate all			false		 	??
	// calculate some			true			> 0
	// calculate none			true			0
	if (calculateSubset && (depths.n == 0)) {
		return false;
	} else {
		return true;
	}
}

/**
 * Return the number of averaging kernels to calculate.
 * @ingroup AveragingKernel
 */
int getNumberOfAveragingKernels() {
	return depths.n;
}

/**
 * Return the Averaging Kernel depth at index i
 * @ingroup AveragingKernel
 */
int getAvgKernelDepth(int i) {
	return depths.data[i];
}

/**
 * Reduce the matrix by extracting only the rows corresponding to the depth
 * at which averaging kernels are calculated.
 * @param[in] m A matrix to reduce.
 * @return The reduced matrix.
 * @ingroup AveragingKernel
 */
MATRIX selectKernels(MATRIX m) {

	bool areRows;
	areRows = true;
	return reduceMatrix(m, depthsForReducing, areRows);
}

/**
 * Free memory associated with averaging kernel calculations
 */
void freeAvgKernelCalculation() {
	if (calculateSubset) {
		if (depths.data != depthsForReducing.data)
			freeVector(depthsForReducing);
		freeVector(depths);
	}
}
