/*
 * main.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include <math.h>
#include <errno.h>
#include "matrix.h"

#include "utilities.h"
#include "Constants.h"
#include "errorHandler.h"
#include "Inversion.h"
#include "ProblemDimension.h"
#include "coramfVersion.h"
#include "inversionVersion.h"
#include "parserVersion.h"


#ifdef OLA_INSTALLED
#include "OLAInit.h"
#include "OLAInvert.h"
#include "ParameterLoop.h"
#include "obaseVersion.h"
#include "olaVersion.h"
#include "OLAMain.h"
#endif

#ifdef RLS_INSTALLED
#include "RLSInit.h"
#include "RLSInvert.h"
#include "RLSMain.c"
#include "rlsVersion.h"
#endif

#ifdef LSQR_INSTALLED
#include "LSQRInit.h"
#include "LSQRInvert.h"
#include "CLSQRInvert.h"
#include "LSQRMain.h"
#include "CLSQRMain.h"
#include "olsqrVersion.h"
#include "olsqrcVersion.h"
#include "Checkpoint.h"
#endif

#ifdef FWD_INSTALLED
#include "Forward.h"
#include "forwardMain.h"
#include "fwdVersion.h"
#endif

#ifdef CALCQ_INSTALLED
#include "CalcQ.h"
#include "calcQMain.c"
#include "calcqVersion.h"
#endif

#include <signal.h>

// local structures and functions
typedef struct {
	bool rls;
	bool ola;
	bool lsqr;
	bool copyInput;
	bool debug;
	char *inputFile;
	bool parseOnly;
	bool noMaps;
	long  ckptIter;
} CommandLineArgs;

void reportUsage();
void processCommandLineArgs(int argc, char *argv[], CommandLineArgs *cla);
void copyInputParamFile(const InputParms *inputParms, const char *inParmFileName);

/**
 * \mainpage
 *
 * The main flow of the code loops over regularization parameters, calculating
 * inversions at each value.  Along the way,
 * full or partial solutions are aggregated, averaging kernels and
 * error covariances are calculated and accumulated, and output
 * data is written to FITS files when it is complete.  There is
 * a distinction in the datatypes between the 1D and 3D problems.
 * The 1d problem uses 2D matrices and the solution is complete
 * after each inversion.  The 3d problem uses 4D matrices and
 * partial solutions are produced by each inversion.  The partial
 * solutions are aggregrated into full 4D solutions and written
 * when they are complete.  There is one main loop inside 'main.c'
 * which extracts data, sends it to the inversion function,
 * then aggregates and outputs results when ready. Each inversion
 * structure contains the data, and function pointers to the unique OLA and RLS
 * algorithms.
 *
 * The main datatypes are MATRIX, MAP_DEPTH_MATRIX, Model, and BaseDirectory.
 * Only three dimensional problems use MAP_DEPTH_MATRIX. All inversions
 * use MATRIX, Model, and BaseDirectory.
 *
 * \section MATRIX
 * MATRIX is a main datatype and contains either double or double complex
 * data.
 *
 * \section MAP_DEPTH_MATRIX
 * MAP_DEPTH_MATRIX is a four-dimensional matrix to keep input and output data.
 * This is a main datatype for the 3D problem.  The input kernel data,
 * input map data, solution data, averaging kernel data and covariance
 * data are all kept in a MAP_DEPTH_MATRIX object.  During the inversion,
 * pointers to 2D sub-matrices are passed around and
 * inserted to accumulate full solutions.  Map-depth data is kept
 * contiguous to make the inversion run faster.  This is done at the
 * expense of data in the xy-plane.  Since input and output to FITS
 * files uses the xy-plane, the expense is incurred during I/O which
 * is inherently slow anyway. Input files are read only once while
 * inversion calculations are performed many times.
 *
 *\section Model
 * Model is an individual solution to an inversion.
 * Individual solutions for each kx, ky accumulate into full solutions.
 * If the problem is 1D, then an individual solution is the
 * full solution, since there is only one kx and one ky.
 * If the problem is 3D, then an individual
 * solution is the solution at one kx,ky.  The code passes
 * around individual solutions, aggregates them into
 * a full model, then writes the full model to FITS files.
 *
 *\section BaseDirectory
 * The BaseDirectory single instance manages the string manipulations
 * to generated the various output file names.
 */

int main(int argc, char *argv[]) {

    //signal (SIGINT, termination_handler);

    //signal (SIGHUP, decrease_handler);


#ifdef OLA_INSTALLED
	static struct Inversion Ola = {
			NULL, // inversion Data
			// define function pointers and initialize to NULL
			&OLAInit,				// int (*init)(char *, char *, bool, InversionData *);
			&OLAInvert2,			// int (*solve)(InversionData *inversionData);
			&OLAComplete, 			// int (*complete)
			&OLADestroy,			// 	int (*destroy)
			&OLAMoreParameterSteps,	// bool (*moreParameters) ();
			&OLANextParameter 		// void (*nextParameter) ();
	};
#endif

#ifdef LSQR_INSTALLED
	static struct Inversion Lsqr = {
			NULL, // inversion Data
			// define function pointers and initialize to NULL
			&LSQRInit,				// int (*init)(char *, char *, bool, InversionData *);
			&LSQRInvert2,			// int (*solve)(InversionData *inversionData);
			// May need to change or parameterize FFT use in complete
			&LSQRComplete, 			// int (*complete)
			&LSQRDestroy,			// 	int (*destroy)
			&LSQRMoreParameterSteps,	// bool (*moreParameters) ();
			&LSQRNextParameter 		// void (*nextParameter) ();
	};

	static struct Inversion CLsqr = {
			NULL, // inversion Data
			// define function pointers and initialize to NULL
			&LSQRInit,
			&CLSQRInvert2,
			&CLSQRComplete,
			&CLSQRDestroy,
			&LSQRMoreParameterSteps,
			&LSQRNextParameter
	};
#endif

#ifdef RLS_INSTALLED
	static struct Inversion Rls3d = {
			NULL, // InputMatrices
			// define function pointers and initialize to NULL
			&RLSInit3d,
			&RLSInvert3d,
			&RLSComplete3d,
			&RLSDestroy3d,
			&RLSMoreParameterSteps,
			RLSNextParameter3d
	};

	static struct Inversion Rls1d = {
			NULL, // InputMatrices
			// define function pointers and initialize to NULL
			&RLSInit1d,
			&RLSInvert1d,
			&RLSComplete1d,
			&RLSDestroy1d,
			&RLSMoreParameterSteps,
			&RLSNextParameter1d
	};
#endif

#ifdef FWD_INSTALLED
	static struct Inversion Forward = {
			NULL, // inversion Data
			// define function pointers
			&ForwardInit,
			&ForwardAddNoise,
			&ForwardComplete,
			&ForwardDestroy,
			&ForwardMoreParameterSteps,
			&ForwardNextParameter
	};
#endif

#ifdef CALCQ_INSTALLED
	static struct Inversion CalcQ = {
			NULL, // inversion Data
			// define function pointers
			&CalcQInit,
			&CalcQSolve,
			&CalcQComplete,
			&CalcQDestroy,
			&CalcQMoreParameters,
			&CalcQNextParameter
	};
#endif


#ifdef CALCQ_INSTALLED
//	printf("invrs Version %s\n", invrsVersion);
#endif

	// check library and inversion for compatible versions
    //  which libraries depends on the type of calculation!
	printf("Expecting version %f\n", CODE_VERSION);
	printf("core Version %f\n", coreVersion);
	printf("coramf Version %f\n", coramfVersion);
	printf("pars Version %f\n", parsVersion);
	printf("invrs Version %f\n", invrsVersion);
	//printf("obase Version %f\n", obaseVersion);

	if ((CODE_VERSION != coreVersion)
		|| (CODE_VERSION != parsVersion)

#if defined(LSQR_INSTALLED) || defined(OLA_INSTALLED)
			|| (CODE_VERSION != obaseVersion)
#endif

#ifdef RLS_INSTALLED
			|| (CODE_VERSION != rlsVersion)
#endif

#ifdef LSQR_INSTALLED
			|| (CODE_VERSION != olsqrVersion)
			|| (CODE_VERSION != olsqrcVersion)
#endif

#ifdef OLA_INSTALLED
			|| (CODE_VERSION != olaVersion)
#endif

#ifdef FWD_INSTALLED
			|| (CODE_VERSION != fwdVersion)
#endif

#ifdef CALCQ_INSTALLED
			|| (CODE_VERSION != calcqVersion)
#endif
			)
		reportError(BadCodeVersions, "main", FATAL);

	CommandLineArgs cla;
	cla.rls = false;
	processCommandLineArgs(argc, argv, &cla);

	InputParms inputParms;
	// sets the value of inputParms.lsqr
	int error = getInputFromFile(cla.inputFile, &inputParms, cla.copyInput);
	if (error) exit(EXIT_FAILURE);
	if (cla.parseOnly) {
		freeInputParms(inputParms);
		printf("Memory Usage  %ld \tFinal\n", getMemoryReport());
		printf("parsed OK\n");
		exit(EXIT_SUCCESS);
	}
	// Otherwise, let's do the calculation ...

	if (inputParms.xt)
		printf("crosstalk is true\n");
	else
		printf("crosstalk is false\n");

	inputParms.noMaps = cla.noMaps;


	struct Inversion op;
	switch(inputParms.whichCalc) {

#ifdef LSQR_INSTALLED
	case LSQR_INV:
		if (inputParms.constrained) {
			op = CLsqr;
			printf("using clsqr\n");
		}
		else {
			op = Lsqr;
			printf("using lsqr\n");
		}
		break;
#endif

#ifdef RLS_INSTALLED
	case RLS_INV:
		if (getProblemDimension() == PDC.D1) {
			op = Rls1d;
		} else {
			op = Rls3d;
		}
		printf("using rls\n");
		break;
#endif

#ifdef OLA_INSTALLED
	case OLA_INV:
		op = Ola;
		printf("using ola\n");
		break;
#endif

#ifdef FWD_INSTALLED
	case FWD_CALC:
		op = Forward;
		printf("using forward calc\n");
		break;
#endif

#ifdef CALCQ_INSTALLED
	case CALCQ_CALC:
		op = CalcQ;
		printf("calculating q\n");
		break;
#endif
	default:
		reportError("Unrecognized type of inversion", "main", FATAL);
		break;
	}

#ifdef LSQR_INSTALLED
	if (cla.ckptIter > 0) {
		if (inputParms.whichCalc != LSQR_INV)
			reportError("Checkpoints used only with LSQR and CLSQR inversions; checkpoints being ignored", "main", WARNING);
		else {
			setCheckpoint(cla.ckptIter, inputParms.outputBaseDirNameOrig, inputParms.outputBaseDirName);
			// decided on a checkpoint object
			// check point is needed by lsqr, passed through callback
			//                          LSQRInvert
			//                          CLSQRInvert
			// let's go with separate object in oBase!!!
			// pass checkpoints through inputParms? yuck;
			// just create object here and set it's value; also have a default value if no check point is specified.

		}
		// make sure there is only one iteration for each parameter if checkpoints are used
		if ((inputParms.lambdaSet1.n > 1) || (inputParms.lambdaSet1.log) ||
				(inputParms.lambdaSet2.n > 1) || (inputParms.lambdaSet2.log) ||
				(inputParms.xt && ((inputParms.lambdaSet3.n > 1) || (inputParms.lambdaSet3.log))) ||
				(inputParms.spread && ((inputParms.lambdaSet4.n > 1) || (inputParms.lambdaSet4.log))) )
				{
			printf("nsteps(mu, sigma_h, nu, epsilon): %d %d %d %d\n", inputParms.lambdaSet1.n,
					inputParms.lambdaSet2.n, inputParms.lambdaSet3.n, inputParms.lambdaSet4.n);
			reportError("Checkpoints allowed only when each parameter (mu, sigma_h, nu) has one value. Exiting", "main", FATAL);
		}
	}
#endif

	// copy the input parameter file to the output directory
	copyInputParamFile(&inputParms, cla.inputFile);

	int result;

	op.data = op.init(inputParms, NULL, cla.debug);

	// loop over parameters
	while (op.moreParameters(op.data)) {
		op.solve(op.data);
		op.complete(op.data);
		op.nextParameter(op.data);
	}
	// end loop
	op.destroy(op.data);

	return(EXIT_SUCCESS);

	return result;
}

/**
 * Parse the command-line arguments.
 * @param[in] argc The number of arguments.
 * @param[in] argv Pointer to an array of strings containing the arguments
 * @param[out] cla Pointer to CommandLineArgs structure containing needed information from
 * command line.
 */

void processCommandLineArgs(int argc, char *argv[], CommandLineArgs *cla) {

	cla->debug = false;
	cla->inputFile = NULL;
	cla->rls = false;
	cla->ola = false;
	cla->lsqr = false;
	cla->parseOnly = false;
	cla->copyInput = false;
	cla->noMaps = false;
	cla->ckptIter = 0;

	if (argc == 1) reportUsage();
	int argIdx = 1;
	while (argIdx < argc) {
		// assume any argument that doesn't begin with a '-' is the input file name
		// Then,  the -f flag is optional
		if (argv[argIdx][0] != '-') {
			// try to read input from file
			cla->inputFile = argv[argIdx];
			argIdx += 1;
		} else if (strcmp(argv[argIdx], Constants.FileFlag) == 0) {
			// try to read input from file
			cla->inputFile = argv[argIdx + 1];
			argIdx += 2;
		} else if (strcmp(argv[argIdx],  Constants.CopyFlag) == 0) {
			cla->copyInput = true;
			argIdx += 1;
		} else if ((strcmp(argv[argIdx], Constants.DebugFlag) == 0)) {
			cla->debug = true;
			argIdx += 1;
		} else if ((strcmp(argv[argIdx], Constants.ParseOnlyFlag) == 0)) {
			cla->parseOnly = true;
			argIdx += 1;
		} else if ((strcmp(argv[argIdx], Constants.NoMapsFlag) == 0)) {
			cla->noMaps = true;
			argIdx += 1;
		} else if ((strcmp(argv[argIdx], Constants.CkptFlag) == 0)) {
			int value;
			int nscans = sscanf(argv[argIdx + 1], "%d", &value);
		    if (nscans == 1) {
		    	cla->ckptIter = value;
		    	argIdx += 2;
		    } else {
		    	reportErrorStr("Cannot read number for check points", argv[argIdx + 1], "main", ERROR);
		    }
		} else {
			reportError(BadCommandLineOption, argv[argIdx], ERROR);
			reportUsage();
		}
	} // end while argIdx < argc

	// make sure all expected command line args are present.
	if (cla->inputFile == NULL)
		reportError(NoInputData, "main", FATAL);
}

/**
 * Print the expected command-line arguments and their syntax.
 */

void reportUsage() {
	printf("Usage (1):  inversion [-f] <input-filename> [-d] [-c] [-p] [-m] [-k N]\n");
	printf("            where input-filename is the name of a file containing input parameters such as M.fits, d.fits, etc.\n");
	printf("                  f is an optional flag to indicate the input file name immediately follows.\n");
	printf("				  d is a debug flag which outputs intermediate values during the inversion calculation. \n");
	printf("				  c is a flag to copy all input files to the output directory. \n");
	printf("				  k is a flag to dump checkpoint files every N iterations for OLSQR and CLSQR inversions. \n");
	printf("				  p is a flag to parse the input file only, no further calculations. \n");
	printf("				  m is a flag to exclude the maps from the calculation.  The map files must still \n");
    printf("                    exist, but they are not held in memory, nor is the model, q, calculated. \n");
    printf("                    This option is only available with OLA, LSQR, or CLSQR inversions. \n");
	printf("\n");
	exit(EXIT_FAILURE);
}

void copyInputParamFile(const InputParms *inputParms, const char *inParmFileName) {


	//copyFiles(const char *source, const char *dest);
	copyFiles(inParmFileName, inputParms->outputBaseDirName);
}
